package abbot;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.WeakHashMap;

import javax.swing.JComponent;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.MenuElement;
import javax.swing.SwingUtilities;

import abbot.tester.awt.Comparator;
import abbot.tester.awt.ComponentTester;
import abbot.tester.awt.Robot;
import abbot.tester.awt.WindowTracker;
import abbot.script.ComponentReference;

/** Basic component lookup implementation.
 */
// FIXME maybe move some of the component methods to component tester and
// derive from component tester?  or just move them to component utilities?
// FIXME could use some merging of getComponents() functions -- there are two
// places that combine children, owned windows, and menuelements
// FIXME match weights is really not quite the right approach.  what you
// really want is to have a few types of lookup; one needs to be exact, one
// needs to be slightly flexible (allow slight changes in hierarchy/layout),
// and maybe another, though it doesn't come to mind.  then the
// comparisons/tradeoffs can be made explicit instead of being obscured by the
// weight checking.
public class DefaultComponentFinder implements ComponentFinder {
    /** Match weight for a strong match. */
    public static final int MW_STRONG_MATCH = 50;
    /** Match weight for a weak match. */
    public static final int MW_WEAK_MATCH = 1;

    // Matching weights for various attributes
    private static final int MW_NAME = 100;
    private static final int MW_TAG = 50;
    private static final int MW_PARENT = 25;
    private static final int MW_INVOKER = 25;
    private static final int MW_INDEX = 25;
    private static final int MW_WINDOW = 10;
    private static final int MW_TITLE = 10;
    private static final int MW_CLASS = 1;
    /** Match weight corresponding to no possible match. */
    public static final int MW_FAILURE = 0;
    /** Timeout for waiting on a popup menu. */
    public static int POPUP_TIMEOUT = 5000;

    private WeakHashMap filteredComponents = new WeakHashMap();
    /** Allow chaining to existing filter sets. */
    private ComponentFinder parent = null;
    private boolean filter = true;
    private WindowTracker tracker = WindowTracker.getInstance();

    static {
        String to = System.getProperty("abbot.finder.popup_timeout");
        if (to != null) {
            try {
                POPUP_TIMEOUT = Integer.parseInt(to);
            }
            catch(Exception e) {
            }
        }
    }

    private static DefaultComponentFinder defaultFinder = null;

    /** This is the factory method to use unless you already know otherwise :)
     */ 
    public static ComponentFinder getFinder() {
        return getFinder(null);
    }

    /** This factory method allows the finder to be used where a finder
     * already exists and is filtering components.  This should only be
     * required in very rare cases like when the script editor needs to test
     * itself. 
     */
    public static synchronized ComponentFinder getFinder(ComponentFinder context) {
        if (defaultFinder == null) {
            defaultFinder = new DefaultComponentFinder(context);
        }
        return defaultFinder;
    }

    /** Only factory creation is allowed. */
    private DefaultComponentFinder(ComponentFinder parent) {
        this.parent = parent;
    }

    /** 
     *  Attempt to find the component corresponding to the given reference
     * among existing, visible components.
     *
     *  @param ref A Component reference to use to identify a component
     *  @throws ComponentNotFoundException if a Component cannot be located via
     *  the passed in ComponentReference
     *  @throws MultipleComponentsFoundException if more then one component is 
     *  found via the passed in reference       
     *  @return The component which matches the passed in ComponentReference
     */
    public Component findComponent(ComponentReference ref)
        throws ComponentNotFoundException, MultipleComponentsFoundException {

        HashSet comps = findMatches(null, ref, MATCH_WEAK);
        Iterator iter = null;
        Component comp = null;
        int max = -99999;
        ArrayList multiples = new ArrayList();

        if (comps.isEmpty()) {
            throw new ComponentNotFoundException("Component '" + ref
                                                 + "' not found");
        }

        if (comps.size() == 1) {
            iter = comps.iterator();
            return (Component)iter.next();
        }

        iter = comps.iterator();
        comp = (Component)iter.next();
        max = getComponentMatchWeight(comp, ref);
        while(iter.hasNext()) {
            Component next = (Component)iter.next();
            int val = getComponentMatchWeight(next, ref);
            if (val > max) {
                max = val;
                comp = next;
                multiples.clear();
            }
            else if (val == max) {
                if (multiples.size() == 0) {
                    multiples.add(comp);
                }
                multiples.add(next);
            }
        }

        if (multiples.size() > 0) {
            String msg = "More than one match for '" + ref + "' found";
            Component[] list = (Component[])
                multiples.toArray(new Component[multiples.size()]);
            throw new MultipleComponentsFoundException(msg, list);
        }

        return comp;
    }


    /**
     * Locates a window using the name as the match critia.
     * @param match The string to match with a window name
     * @return A window with a name equal to the passed in string, null
     * otherwise 
     */
    public Window findWindowByName(String match) {
        Window[] windows = getWindows();
        for (int i=0;i < windows.length;i++) {
            Window w = windows[i];
            if (w.isShowing()) {
                String name = getComponentName(w);
                if ((match != null && match.equals(name))
                    || name == match)
                    return w;
            }
        }
        return null;
    }

    /** Return the first visible window whose title matches the given
     * pattern.
     */
    public Window findWindowByTitle(String title) {
        Window[] windows = getWindows();
        for (int i=0;i < windows.length;i++) {
            Window w = windows[i];
            if (w.isShowing()) {
                if (w instanceof Dialog) {
                    if (titlesMatch(title, ((Dialog)w).getTitle())) {
                        return w;
                    }
                }
                else if (w instanceof Frame) {
                    if (titlesMatch(title, ((Frame)w).getTitle())) {
                        return w;
                    }
                }
            }
        }
        return null;
    }

    /** Return the window with the given name or title.  Attempts to find a
     * named window first, since that search is more restrictive. */
    public Window findWindow(String nameOrTitle) {
        Window w = findWindowByName(nameOrTitle);
        if (w == null)
            w = findWindowByTitle(nameOrTitle);
        return w;
    }

    /** Returns the currently active popup menu with the given invoker. */
    public Component findActivePopupMenu(Component invoker) {
        Component popup = findActivePopupMenu(null, invoker);
        if (popup == null
            && !SwingUtilities.isEventDispatchThread()) {
            long now = System.currentTimeMillis();
            while ((popup = findActivePopupMenu(null, invoker)) == null) {
                if (System.currentTimeMillis() - now > POPUP_TIMEOUT) {
                    break;
                }
                Robot.sleep();
            }
        }
        return popup;
    }

    // FIXME there has got to be a better way of finding the currently active
    // popup menu
    private Component findActivePopupMenu(Component root, Component invoker) {
        if (root == null) {
            Window[] wins = getWindows();
            for (int i=0;i < wins.length;i++) {
                Component comp = (Component)wins[i];
                Component menu = findActivePopupMenu(comp, invoker);
                if (menu != null)
                    return menu;
            }
        }
        else if (root instanceof javax.swing.JPopupMenu
            && ((JPopupMenu)root).isShowing()) {
            return root;
        }
        else if (root instanceof Container) {
            Component[] subs = ((Container)root).getComponents();
            for (int i=0;i < subs.length;i++) {
                Component menu = findActivePopupMenu(subs[i], invoker);
                if (menu != null)
                    return menu;
            }
        }
        return null;
    }

    public Component findMenuItemByName(Component root, String name) {
        if (root instanceof JMenuItem
            && ((JMenuItem)root).getText().equals(name))
            return root;
        if (root instanceof Container) {
            Component[] subs = ((Container)root).getComponents();
            for (int i=0;i < subs.length;i++) {
                Component comp = findMenuItemByName(subs[i], name);
                if (comp != null)
                    return comp;
            }
        }
        if (root instanceof MenuElement) {
            MenuElement[] els = ((MenuElement)root).getSubElements();
            for (int i=0;i < els.length;i++) {
                Component comp = findMenuItemByName((Component)els[i], name);
                if (comp != null)
                    return comp;
            }
        }
        return null;
    }

    /** Return the the set of all components under the given component's
     * hierarchy (inclusive) which match the given reference.
     */
    private HashSet findMatches(Component ancestor, ComponentReference ref, 
                                int type) {
        Log.debug("Scanning " + Robot.toString(ancestor));
        HashSet set = new HashSet();
        if (isComponentFiltered(ancestor))
            return set;

        if (ancestor == null) {
            // Examine all top-level components and their owned windows.
            Window[] frames = getRootWindows();
            for (int f=0;f < frames.length;f++) {
                set.addAll(findMatches(frames[f], ref, type));
            }
            return set;
        }

        if (componentsMatch(ancestor, ref, type)) {
            Log.debug("Found match");
            set.add(ancestor);
        }

        if (ancestor instanceof Container) {
            Component[] children = getComponents((Container)ancestor);
            String title = ref.getTitle();
            for (int i=0;i < children.length;i++) {
                Component child = children[i];
                if (child instanceof Window) {
                    // Optimization:
                    // Skip this hierarchy if the frame titles don't match,
                    if (!((child instanceof Frame)
                          || (child instanceof Dialog))
                        && ((title != null
                             && !titlesMatch(title, getComponentWindowTitle(child))))) {
                        Log.debug("Skipping window with wrong title "
                                  + Robot.toString(child));
                        continue;
                    }

                    // Skip this hierarchy if the window is not showing
                    if (!((Window)child).isShowing()) {
                        Log.debug("Skipping hidden window "
                                  + Robot.toString(child));
                        continue;
                    }
                }
                set.addAll(findMatches(child, ref, type));
            }
        }

        return set;
    }

    /** 
     * Return whether the the given title matches the given pattern. 
     */
    private boolean titlesMatch(String pattern, String actual) {
        boolean match = Comparator.stringsMatch(pattern, actual);
        return match;
    }

    /** Return the component's name, ensuring that null is returned if the
     * name appears to be auto-generated.
     */
    public String getComponentName(Component comp) {
        // Component.getName() can cause deadlocks if invoked off the event
        // dispatch thread (since it locks both the component and then the
        // component's class; the event thread is likely to want to lock the
        // component and may have already locked the class.
        String name = comp.getName();
        // Pretend auto-generated names don't exist.
        if (name != null 
            && (Regexp.stringMatch("win[0-9]+", name)
                || Regexp.stringMatch("dialog[0-9]+", name)
                || Regexp.stringMatch("frame[0-9]+", name)
                || Regexp.stringMatch("canvas[0-9]+", name)
                // AWT components always get a default name; ignore it if it
                // looks like it's the default
                || (!(comp instanceof JComponent)
                    && Regexp.stringMatch("[a-z]+[0-9]+", name)))) {
            name = null;
        }
        return name;
    }

    /** Return the total weight required for an exact match. */
    private int getExactMatchWeight(ComponentReference ref) {
        int weight = MW_CLASS;
        if (ref.getName() != null)
            weight += MW_NAME;
        if (ref.getTag() != null)
            weight += MW_TAG;
        if (ref.getInvokerID() != null)
            weight += MW_INVOKER;
        if (ref.getParentID() != null)
            weight += MW_PARENT;
        if (ref.getWindowID() != null)
            weight += MW_WINDOW;
        if (ref.getTitle() != null)
            weight += MW_TITLE;
        if (ref.getIndex() != -1)
            weight += MW_INDEX;
        return weight;
    }

    // FIXME should probably throw if more than one match is found. 
    public ComponentReference matchComponent(Component comp,
                                             Iterator iter, int matchType) {
        //Log.debug("Looking for " + Robot.toString(comp) + " in a group");
        ComponentReference match = null;
        int matchWeight = -1000;
        while (iter.hasNext()) {
            ComponentReference ref = (ComponentReference)iter.next();
            if (componentsMatch(comp, ref, matchType)) {
                int wt = getComponentMatchWeight(comp, ref);
                if (wt > matchWeight) {
                    match = ref;
                    matchWeight = wt;
                }
                else if (wt == matchWeight) {
                    // FIXME work out how to throw and handle this properly
                    Log.warn("Multiple references match");
                    throw new RuntimeException("Multiple references match");
                }
            }
        }
        //Log.debug(match != null ? "Found" : "Not found");
        return match;
    }

    /** Return a measure of how well the given component matches the given
     * component reference.
     */
    int getComponentMatchWeight(Component comp, ComponentReference ref) {
        int weight = MW_FAILURE;

        if (null == comp || null == ref){
            return MW_FAILURE;
        }

        if (!ref.isAssignableFrom(comp.getClass())) {
            return MW_FAILURE;        
        } 
        else {
            weight += MW_CLASS;
            // Exact class matches are better than non-exact matches
            if (ref.getRefClassName().equals(comp.getClass().getName()))
                weight += MW_CLASS;
        }
       
        String compTag = ComponentTester.getTag(comp);
        String refTag = ref.getTag();
        if (null != compTag && null != refTag) {
            if (compTag.equals(refTag)) {
                weight += MW_TAG;
            }
            else {
                return MW_FAILURE;
            }
        }

        String compName = getComponentName(comp);
        String refName = ref.getName();
        if (null != compName && null != refName) {
            if (compName.equals(refName)) {
                weight += MW_NAME;
            }
            else {
                weight -= MW_NAME;
            }
        }

        if (null != ref.getInvokerID()) {
            if ((comp instanceof JPopupMenu)
                && componentsMatch(((JPopupMenu)comp).getInvoker(),
                                   ref.getInvokerReference(), MATCH_EXACT)) {
                weight += MW_INVOKER;
            }
            else {
                // Invoking components aren't likely to change
                weight -= MW_INVOKER;
            }
        }

        if (null != ref.getParentID()) {
            if (componentsMatch(getComponentParent(comp),
                                ref.getParentReference(), MATCH_EXACT)) {
                weight += MW_PARENT;
            }
            // Don't detract on parent mismatch, since changing a parent is
            // not that big a change (e.g. adding a scroll pane)
        }

        if (null != ref.getWindowID()) {
            if (componentsMatch(getComponentWindow(comp),
                                ref.getWindowReference(), MATCH_EXACT)) {
                weight += MW_WINDOW;
            } 
            else {
                // Changing windows is a big change and not very likely
                weight -= MW_WINDOW;
            }
        }

        if (ref.getTitle() != null) {
            if (titlesMatch(ref.getTitle(), getComponentWindowTitle(comp))) {
                weight += MW_TITLE;
            }
            // Don't subtract on mismatch, since title changes are common
        }

        if (ref.getIndex() != -1) {
            Container parent = comp.getParent();
            if (null != parent) {
                Component[] children = parent.getComponents(); 
                for (int i = 0; i < children.length; ++i) {
                    if (children[i] == comp && i == ref.getIndex()) {
                        weight += MW_INDEX;
                        break;
                    }
                }
            }
        }
        /*
        Log.debug("Comparing " + Robot.toString(comp)
                  + " to '" + ref + "', weight is " + weight);
        */
        return weight;
    }

    /** Determine the best we can whether the component is the one referred to
     * by the reference.  The match may be a strong or a weak one.
     */
    public boolean componentsMatch(Component comp, ComponentReference ref,
                                   int type){
        int weight = getComponentMatchWeight(comp, ref);
        int min = 0;
        if (type == MATCH_EXACT) {
            min = getExactMatchWeight(ref);
            if (weight >= min && min < MW_STRONG_MATCH) {
                Log.warn("Exact match weight comparing "
                         + ref + " to " + Robot.toString(comp)
                         + " is less than a strong match");
            }
        }
        else if (type == MATCH_STRONG) {
            min = Math.min(MW_STRONG_MATCH, getExactMatchWeight(ref));
        }
        else if (type == MATCH_WEAK) {
            min = Math.min(MW_WEAK_MATCH, getExactMatchWeight(ref));
        }
        else
            throw new IllegalArgumentException("Type must be one of MATCH_EXACT, MATCH_STRONG, or MATCH_WEAK");
        return weight >= min;
    }

    /** Return all windows owned by the given window (filtered). */
    private List getWindowList(Window parent) {
        return getWindowList(parent, filter, false);
    }

    /** Return all windows owned by the given window, optionally filtered. */
    private List getWindowList(Window parent, boolean filter, boolean recurse){
        Window[] windows = parent.getOwnedWindows();
        ArrayList list = new ArrayList();
        for (int i=0;i < windows.length;i++) {
            if (!filter || !isComponentFiltered(windows[i])) {
                list.add(windows[i]);
                if (recurse) {
                    list.addAll(getWindowList(windows[i], filter, recurse));
                }
            }
        }
        return list;
    }

    /** Return an array of all available Frames, across all known contexts.
    */ 
    public Window[] getRootWindows() {
        ArrayList unf = new ArrayList();
        Iterator iter = tracker.getRootWindows().iterator();
        while (iter.hasNext()) {
            Window win = (Window)iter.next();
            if (!isComponentFiltered(win))
                unf.add(win);
        }
        return (Window[])unf.toArray(new Window[unf.size()]);
    }

    /** Return all windows owned by the given window that have not been
     * filtered.
     */ 
    public Window[] getWindows(Window parent) {
        List list = getWindowList(parent, filter, false);
        return (Window[])list.toArray(new Window[list.size()]);
    }

    private List getWindowList() {
        List frames = tracker.getRootWindows();
        Iterator iter = frames.iterator();
        ArrayList list = new ArrayList();
        while(iter.hasNext()) {
            Window root = (Window)iter.next();
            if (!filter || !isComponentFiltered(root)) {
                list.add(root);
                list.addAll(getWindowList(root, filter, true));
            }
        }
        return list;
    }

    /** Returns the set of all available windows that have not been
     * filtered.  This includes frames and windows.
     */ 
    public Window[] getWindows() {
        List list = getWindowList();
        return (Window[])list.toArray(new Window[list.size()]);
    }

    /** Returns all components below the GUI hierarchy of the given Container,
     * including Windows and MenuElements.
     */ 
    public Component[] getComponents(Container c) {
        Component[] children = c.getComponents();
        ArrayList unf = new ArrayList();
        for (int i=0;i < children.length;i++) {
            if (!isComponentFiltered(children[i]))
                unf.add(children[i]);
        }
        if (c instanceof javax.swing.JMenu) {
            MenuElement[] els = ((javax.swing.JMenu)c).getSubElements();
            for (int i=0;i < els.length;i++) {
                if (!isComponentFiltered(els[i].getComponent()))
                    unf.add(els[i].getComponent());
            }
        }
        if (c instanceof Window) {
            unf.addAll(getWindowList((Window)c));
        }
        children = (Component[])unf.toArray(new Component[unf.size()]);
        return children;
    }

    /** Look up the apparent parent of a component.  A
     * popup menu's parent is the menu or component that spawned it.
     */
    public Component getComponentParent(Component comp) {
        Component parent = comp.getParent();
        // A popup menu's "parent" is its invoker
        if (comp instanceof JPopupMenu) {
            parent = ((JPopupMenu)comp).getInvoker();
        }
        return parent;
    }

    /** Return the component's owning frame.   There will
     * <b>always</b> one of these; even a frameless window will have a
     * temporary frame generated for it.
     */
    public Frame getComponentFrame(Component comp) {
        Component parent = comp;
        while (!(parent instanceof Frame) 
               && (comp = getComponentParent(parent)) != null) {
            parent = comp;
        }
        return (Frame)comp;
    }

    /** Return the nearest Window ancestor of the given Component.
        Note that for popups, this means the window of the invoker.
     */
    public Window getComponentWindow(Component comp) {
        Component parent = comp;
        while (!(parent instanceof Window) 
               && (comp = getComponentParent(parent)) != null) {
            parent = comp;
        }
        return (Window)comp;
    }

    /** 
     * Return the title of the nearest ancestor window with a title.  
     * If no title is found, return null.
     */
    public String getComponentWindowTitle(Component comp) {
        Window w = getComponentTitledWindow(comp);
        String title = null;
        if (w instanceof Frame) {
            title = ((Frame)w).getTitle();
        }
        else if (w instanceof Dialog) {
            title = ((Dialog)w).getTitle();
        }
        return title;
    }

    /** Return the component's owning frame or dialog.   There will
     * <b>always</b> one of these; even a frameless window will have a
     * temporary frame generated for it.
     */
    public Window getComponentTitledWindow(Component comp) {
        Component parent = comp;
        while (!(comp instanceof Frame || comp instanceof Dialog)
               && (comp = getComponentParent(parent)) != null) {
            parent = comp;
        }
        return (comp instanceof Frame || comp instanceof Dialog) 
            ? (Window)comp : null;
    }

    private boolean isComponentFiltered(Component comp) {
        if (comp == null)
            return false;
        return filter 
            && (filteredComponents.containsKey(comp)
                || (parent != null && parent.isFiltered(comp)));
    }

    /** Returns whether any window or frame ancestor is filtered. */
    private boolean isAncestorFiltered(Component comp) {
        Window w = getComponentWindow(comp);
        if (w == comp) {
            w = (Window)getComponentParent(w);
        }
        else if (filter && comp.getParent() == null) {
            // Ignore null-parented components
            return true;
        }
        return isFiltered(w);
    }

    /** Returns true if the component or its Window ancestor is filtered. */
    public boolean isFiltered(Component comp) {
        return (comp != null)
            && (isComponentFiltered(comp) || isAncestorFiltered(comp));
    }

    private boolean isSharedInvisibleFrame(Component comp) {
        return comp.getClass().getName().startsWith("javax.swing.SwingUtilities$1");
    }

    public void filterComponent(Component comp) {
        if (isSharedInvisibleFrame(comp)) {
            // If the component is Swing's parent for windows and dialog with
            // a null parent, don't filter it, but do filter its children.  
            Component[] children = getComponents((Container)comp);
            for (int i=0;i < children.length;i++) {
                filterComponent(children[i]);
            }
        }
        else {
            //Log.debug("Now filtering " + Robot.toString(comp));
            filteredComponents.put(comp, comp);
        }
    }

    /** Discard and no longer reference the given component. */
    public void discardComponent(Component comp) {
        Log.debug("Discarding " + Robot.toString(comp));
        filterComponent(comp);
        if (parent != null) {
            parent.discardComponent(comp);
        }
    }

    public void discardAllComponents() {
        Iterator iter = getWindowList().iterator();
        while(iter.hasNext()) {
            discardComponent((Window)iter.next());
        }
    }

    public void setFilterEnabled(boolean enable) {
        this.filter = enable;
    }

    /** Send an explicit window close event to all showing windows.  Note
        that this is not guaranteed to actually make the window go away.  */
    public void closeWindows() {
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                Iterator iter = getWindowList().iterator();
                while (iter.hasNext()) {
                    Window win = (Window)iter.next();
                    if (win.isShowing()) {
                        WindowEvent ev = 
                            new WindowEvent(win, WindowEvent.WINDOW_CLOSING);
                        win.getToolkit().getSystemEventQueue().postEvent(ev);
                    }
                }
            }
        });
    }

    /** Dispose of all available windows, and does not return until they have
        been disposed of.  */
    public void disposeWindows() {
        // This must be done on the swing thread, since disposing of a window
        // may trigger more gui events and/or locking the tree, and we want to
        // wait until it all gets done.
        Log.debug("Disposing all windows");
        Window[] windows = getWindows();
        for (int i=0;i < windows.length;i++) {
            final Window win = windows[i];
            discardComponent(win);
            // HACK to distinguish between the abbot framework
            // disposing a window and anyone else (AppletViewer, for
            // instance). 
            // FIXME maybe use ThreadLocal instead of a property
            System.setProperty("abbot.finder.disposal", "true");
            try {
                Log.debug("Disposing of " + Robot.toString(win));
                if (!isSharedInvisibleFrame(win))
                    win.dispose();
            }
            catch(NullPointerException npe) {
                // Catch bug in AWT 1.3.1 when generating hierarchy
                // events 
                Log.log(npe);
            }
            System.setProperty("abbot.finder.disposal", "false");
        }
    }

    public boolean isWindowShowing(Window w) {
        return tracker.isWindowShowing(w);
    }
}
