package abbot;

import java.util.StringTokenizer;

/** Simple utility to figure out what platform we're on. */

public class Platform {

    public static final int JAVA_1_0 = 0x1000;
    public static final int JAVA_1_1 = 0x1100;
    public static final int JAVA_1_2 = 0x1200;
    public static final int JAVA_1_3 = 0x1300;
    public static final int JAVA_1_4 = 0x1400;

    private static String osName = System.getProperty("os.name");
    private static String osVersion = System.getProperty("os.version");
    private static String osArch = System.getProperty("os.arch");

    private static String javaVersion = System.getProperty("java.version");
    private static int javaVersionNumber = JAVA_1_0;

    private static boolean isWindows = osName.startsWith("Windows");
    private static boolean isWindowsXP = isWindows && osName.indexOf("XP") != -1;
    private static boolean isMac = System.getProperty("mrj.version") != null;
    private static boolean isOSX = isMac && osName.indexOf("OS X") != -1;
    private static boolean isSunOS = (osName.startsWith("SunOS")
                                      || osName.startsWith("Solaris"));
    private static boolean isHPUX = osName.equals("HP-UX");
    private static boolean isLinux = osName.equals("Linux");

    /** No instantiations. */
    private Platform() {
    }

    static boolean parsed = false;
    private static void parse() {
        if (!parsed) {
            javaVersionNumber = parse(javaVersion);
            parsed = true;
        }
    }

    private static String strip(String number) {
        while (number.startsWith("0") && number.length() > 1)
            number = number.substring(1);
        return number;
    }

    static int parse(String vs) {
        int version = 0;
        try {
            StringTokenizer st = new StringTokenizer(vs, "._");
            version = Integer.parseInt(strip(st.nextToken())) * 0x1000;
            version += Integer.parseInt(strip(st.nextToken())) * 0x100;
            version += Integer.parseInt(strip(st.nextToken())) * 0x10;
            version += Integer.parseInt(strip(st.nextToken()));
        }
        catch(NumberFormatException nfe) {
        }
        catch(java.util.NoSuchElementException nse) {
        }
        return version;
    }

    // FIXME this isn't entirely correct, maybe should look for a motif class
    // instead. 
    public static boolean isX11() { return !isOSX && !isWindows; }
    public static boolean isWindows() { return isWindows; }
    public static boolean isWindowsXP() { return isWindowsXP; }
    public static boolean isMacintosh() { return isMac; }
    public static boolean isOSX() { return isOSX; }
    public static boolean isSolaris() { return isSunOS; }
    public static boolean isHPUX() { return isHPUX; }
    public static boolean isLinux() { return isLinux; }
    public static String getJavaVersion() { return javaVersion; }
    public static int getJavaVersionNumber() {
        parse();
        return javaVersionNumber;
    }

    /** Prior to 1.4.1, hierarchy events are only sent if listeners are added
        to a given component.
     */
    public static boolean hasHierarchyEventGenerationBug() {
        return getJavaVersionNumber() < 0x1410;
    }

    /** OSX prior to 1.4 has really crappy key input handling. */
    public static boolean hasKeyStrokeGenerationBug() {
        return isOSX() && getJavaVersionNumber() < 0x1400;
    }

    /** Will we get incorrectly get multiple clicks on rapid robot clicks?
        frame1->click,frame2->click 
        frame1->click,frame1->hide/show,frame1->click
    */
    public static boolean hasMultiClickFrameBug() {
        // w32 (1.3.02) will count it even if it was on a different component!
        // w32 (1.3.1_06) same
        // w32 (1.4.1_02) same (sporadic)
        // OSX (1.3.x) has the same problem
        // Not sure if it's fixed in 1.4
        // Haven't seen it on linux
        return ((isWindows() && getJavaVersionNumber() < 0x1500)
                || (isOSX() && getJavaVersionNumber() < 0x1410)
                || getJavaVersionNumber() < 0x1300);
    }

    /** OS X (as of 1.3.1, v10.1.5), will sometimes send a click to the wrong
        component after a mouse move.<p>
        Linux x86 (1.3.1) has a similar problem, although it manifests it at
        different times (need a bug test case for this one).<p>
        Solaris and HPUX probably share code with the linux VM implementation,
        so the bug there is probably identical.<p>
        Add some jitter to work around the bug.<p>
        FIXME add tests to determine presence of bug.<p>
    */
    public static boolean hasRobotMotionBug() {
        return ((Platform.isOSX() || Platform.isLinux()
                 || Platform.isSolaris() || Platform.isHPUX())
                && getJavaVersionNumber() < 0x1410)
            || Boolean.getBoolean("abbot.robot.need_jitter");
    }

    public static void main(String[] args) {
        System.out.println("Java version is " + getJavaVersion());
        System.out.println("Version number is " + Integer.toHexString(getJavaVersionNumber()));
    }
}
