package abbot.script;

import java.awt.EventQueue;
import java.awt.Toolkit;
import java.util.EmptyStackException;

import javax.swing.SwingUtilities;

import abbot.Log;
import abbot.util.AWT;
import abbot.util.PreloadingClassLoader;

/**
 * A custom class loader which installs itself as if it were the application
 * class loader.  A classpath of null is equivalent to the system property
 * java.class.path.<p> 
 * The class loader may optionally load a class <i>before</i> the parent class
 * loader gets a chance to look for the class (instead of the default
 * behavior, which always delegates to the parent class loader first).  This
 * behavior enables the class to be reloaded simply by using a new instance of
 * this class loader with each launch of the app.<p>  
 * This class mimics the behavior of sun.misc.Launcher$AppClassLoader as much
 * as possible.<p>
 * Bootstrap classes are always delegated to the bootstrap loader, with the
 * exception of the sun.applet package, which should never be delegated, since
 * it does not work properly unless it is reloaded.<p>  
 * The parent of this class loader will be the normal, default AppClassLoader
 * (specifically, the class loader which loaded this class will be used).
 */
public class AppClassLoader extends PreloadingClassLoader {

    /** A new event queue installed for the lifetime of this class loader. */
    private AppEventQueue eventQueue;
    /** This class loader is used to load bootstrap classes that must be
     * preloaded.
     */
    private BootstrapClassLoader bootstrapLoader;
    /** For use in checking if a class is in the framework class path. */
    private PreloadingClassLoader extensionsLoader;
    /** Whether this is a delegating class loader. */
    private boolean delegate = false;

    /* Class path searched by this loader. */
    private String classPath = null;
    /** Old class loader context for the thread where this loader was
     * installed.
     */
    private ClassLoader oldClassLoader = null;
    private Thread installedThread = null;
    private String oldClassPath = null;
	 
    /**
     * Constructs a AppClassLoader with a custom classpath, indicating
     * whether the class loader should delegate to its parent class loader
     * prior to searching for a given resource.<p>
     * The class path argument may use either a colon or semicolon to separate
     * its elements.<p>
     */
    public AppClassLoader(String classPath, boolean delegate) {
        super(classPath, AppClassLoader.class.getClassLoader());
        this.delegate = delegate;
        bootstrapLoader = new BootstrapClassLoader(getBootstrapPath());
        // Use this one to look up extensions; we want to reload them, but may
        // need to look in the framework class path for them.
        extensionsLoader =
            new PreloadingClassLoader(null, 
                                      AppClassLoader.class.getClassLoader()) {
                /** Make sure we <i>only</i> load extensions with this
                 * loader.
                 */ 
                protected boolean shouldDelegate(String name) {
                    return !isExtension(name);
                }
            };
    }

    /** Should the parent class loader try to load this class first? */
    protected boolean shouldDelegate(String name) {
        return delegate || bootstrapLoader.shouldDelegate(name);
    }

    private boolean isExtension(String name) {
        return name.startsWith("abbot.tester.extensions.")
            || name.startsWith("abbot.converter.extensions.");
    }

    /** Returns the path to the primary JRE classes, not including any
     * extensions.  This is primarily needed for loading
     * sun.applet.AppletViewer, since most other classes in the bootstrap path
     * should <i>only</i> be loaded by the bootstrap loader.
     */
    protected String getBootstrapPath() {
        return System.getProperty("sun.boot.class.path");
    }

    /**
     * Finds and loads the class with the specified name from the search
     * path.  If the class is a bootstrap class and must be preloaded, use our
     * own bootstrap loader. 
     *
     * @param name the name of the class
     * @return the resulting class
     * @exception ClassNotFoundException if the class could not be found
     */
    public Class findClass(String name) throws ClassNotFoundException {
        try {
            if (mustPreload(name)) 
                return bootstrapLoader.findClass(name);
        }
        catch(ClassNotFoundException cnf) {
            Log.warn(cnf);
        }
        try {
            return super.findClass(name);
        }
        catch(ClassNotFoundException cnf) {
            // Look for extensions in the app class path; if found there
            // (as opposed to the primordial class path) then enable
            // reloading of the extension.  
            if (isExtension(name))
                return extensionsLoader.findClass(name);
            throw cnf;
        }
    }

    boolean isBootstrapClass(String name) {
        return bootstrapLoader.isBootstrapClass(name);
    }

    /** Ensure that everything else subsequently loaded on the same thread or
     * any subsequently spawned threads uses the given class loader.  Also
     * ensure that classes loaded by the event dispatch thread and threads it
     * spawns use the given class loader.
     */
    public synchronized void install() {
        if (SwingUtilities.isEventDispatchThread()) {
            throw new IllegalStateException("Class loader may not be installed directly on the event dispatch thread");
        }

        if (installedThread != null) {
            String msg = "Class loader may not be installed on multiple "
                + "independent threads.  Already installed on "
                + installedThread;
            throw new IllegalStateException(msg);
        }

        // Change the effective classpath
        oldClassPath = System.getProperty("java.class.path");
        System.setProperty("java.class.path", getClassPath());
        Log.debug("java.class.path set to "
                  + System.getProperty("java.class.path"));

        // Ensure the event exception handler is set prior to setting
        // the context class loader.  Must install from the event queue itself
        // to avoid deadlock.
        eventQueue = new AppEventQueue();
        eventQueue.install();

        Thread current = Thread.currentThread();
        installedThread = current;
        oldClassLoader = installedThread.getContextClassLoader();
        installedThread.setContextClassLoader(this);
    }
    
    /** Reverse the effects of install.   Has no effect if the class loader
     * has not been installed on any thread.
     */
    public synchronized void uninstall() {
        if (eventQueue != null) {
            eventQueue.uninstall();
            eventQueue = null;
        }
        if (installedThread != null) {
            installedThread.setContextClassLoader(oldClassLoader);
            oldClassLoader = null;
            installedThread = null;
            System.setProperty("java.class.path", oldClassPath);
            oldClassPath = null;
        }
    }

    private class AppEventQueue extends EventQueue {
        /** Ensure the class loader for the event dispatch thread is the right
            one.
        */
        public void install() {
            // Avoid deadlock with the event queue, in case it has the tree
            // lock (pickens). 
            AWT.invokeAndWait(new Runnable() {
                public void run() {
                    Toolkit.getDefaultToolkit().
                        getSystemEventQueue().push(AppEventQueue.this);
                }
            });
            // Install our own handler for catching exceptions on the event
            // dispatch thread.
            try {
                EventExceptionHandler.install(EventExceptionHandler.class);
            }
            catch(Exception e) {
                // ignore it
            }
            AWT.invokeAndWait(new Runnable() {
                public void run() {
                    Thread.currentThread().
                        setContextClassLoader(AppClassLoader.this);
                }
            });
        }
        /** This restores the original event queue. */
        public void uninstall() {
            try {
                pop();
            }
            catch(EmptyStackException ese) {
                Log.warn(ese);
            }
        }
    }
    /** List of bootstrap classes we most definitely want to be loaded by this
     * class loader, rather than any parent, or the bootstrap loader.
     */
    private String[] mustPreloadPrefixes = {
        "sun.applet.",  // need the whole package, not just AppletViewer
    };
    /** Does the given class absolutely need to be preloaded? */
    private boolean mustPreload(String name) {
        for (int i=0;i < mustPreloadPrefixes.length;i++) {
            if (name.startsWith(mustPreloadPrefixes[i]))
                return true;
        }
        return false;
    }

    /** Classes we know *must* be loaded by the primordial bootstrap loader,
     * not ours.
     */ 
    private static final String[] bootstrapPrefixes = {
        "java.",
    };

    /** Provide access to bootstrap classes that we need to be able to
     * reload.
     */ 
    private class BootstrapClassLoader extends PreloadingClassLoader {
        public BootstrapClassLoader(String path) {
            super(path, null);
        }
        protected boolean shouldDelegate(String name) {
            return mustDelegate(name);
        }

        /** Must the given class be delegated to the parent loader?
         */
        private boolean mustDelegate(String name) {
            // Excluded all bootstrap classes, except for those we know we
            // *must* reload in order to have function properly (e.g. applet) 
            if (mustPreload(name))
                return false;
            // Otherwise, we must delegate all bootstrap classes
            return isBootstrapClass(name);
        }

        /** Return whether the class is in the bootstrap path. */
        private boolean isBootstrapClass(String className) {
            for (int i=0;i < bootstrapPrefixes.length;i++) {
                if (className.startsWith(bootstrapPrefixes[i]))
                    return true;
            }
            try {
                BootstrapClassLoader.this.findClass(className);
                return true;
            }
            catch(ClassNotFoundException cnf) {
                return false;
            }
            catch(LinkageError le) {
                // Bootstrap loader already has it
                return true;
            }
        }
    }
}
