package abbot.script;

import java.awt.Component;
import java.awt.Container;
import java.awt.Point;
import java.awt.Window;
import java.io.StringReader;
import java.util.Iterator;

import javax.swing.JMenu;
import javax.swing.JPopupMenu;

import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import abbot.ComponentFinder;
import abbot.ComponentNotFoundException;
import abbot.DefaultComponentFinder;
import abbot.Log;
import abbot.MultipleComponentsFoundException;
import abbot.Resolver;
import abbot.tester.awt.ComponentTester;


/** Encapsulate as much information as is available to identify a GUI
 * component. Usage:<br>
 * <blockquote><code>
 * &lt;component id="..." class="..." [...]&gt;<br>
 * </code></blockquote>
 * A number of optional tags are supported to provide an increasingly precise
 * specification of the desired component:<br>
 * <ul>
 * <li><b><code>tag</code></b> a class-specific identifier, e.g. the label on a
 * JButton.<br>
 * <li><b><code>name</code></b> the Component name, as set by
 * <code>setName()</code>.  Note that auto-generated names for windows and
 * frames are ignored and are considered to be null.
 * <li><b><code>parent</code></b> the reference id of this component's
 * parent.<br> 
 * </ul>
 * (This list subject to change as required by the
 * <a href="{@docRoot}/abbot/DefaultComponentFinder.html">DefaultComponentFinder</a>).  
 */
// TODO: extract resolver stuff
// - parent, window, invoker would have to be saved as components, then
//   optionally added when added to the resolver
public class ComponentReference implements Tags, XMLifiable {

    private static final String CR_USAGE = 
        "<component id=\"<id>\" class=\"...\" .../>";

    private Resolver resolver;
    private ComponentFinder finder;
    // unique id for this reference
    private String id = null;
    // component name, if any
    private String name = null;
    // component class name
    private String refClassName = null;
    // parent's ref id
    private String parentID = null;
    private ComponentReference parentReference = null;
    // index among parent's children
    private int index = -1;
    // parent window ref id
    private String windowID = null;
    private ComponentReference windowReference = null;
    // parent window title 
    private String title = null;
    // tag specific to this component
    private String tag = null;
    // for popup menus only, the invoking component
    private String invokerID = null;
    private ComponentReference invokerReference = null;
    // for popup menus only, the location of invocation
    private Point point = new Point(-1,-1);

    /** For general component lookup by class name.
     * @param id Desired ID for the reference.  Only used if this reference is
     * to be passed as the parent, window, or invoker of another.
     * @param compClass Class of the component (required)
     */
    public ComponentReference(String id, Class compClass) {
        this(id, compClass, null, null, null, null, -1, null);
    }

    /** For general component lookup.
     * @param id Desired ID for the reference.  Only used if this reference is
     * to be passed as the parent, window, or invoker of another.
     * @param compClass Class of the component (required)
     * @param name Name of the component, or null
     * @param tag Tag as returned by ComponentTester.getTag(Component)
     */
    public ComponentReference(String id, Class compClass,
                              String name, String tag) {
        this(id, compClass, name, tag, null, null, -1, null);
    }

    /** For general component lookup.
     * @param id Desired ID for the reference.  Only used if this reference is
     * to be passed as the parent, window, or invoker of another.
     * @param compClass Class of the component (required)
     * @param name Name of the component, or null
     * @param tag Tag as returned by ComponentTester.getTag(Component)
     * @param title Owning Frame/Dialog title, or null
     */
    public ComponentReference(String id, Class compClass,
                              String name, String tag, String title) {
        this(id, compClass, name, tag, title, null, -1, null);
    }

    /** For general component lookup.
     * @param id Desired ID for the reference.  Only used if this reference is
     * to be passed as the parent, window, or invoker of another.
     * @param compClass Class of the component (required)
     * @param name Name of the component, or null
     * @param tag Tag as returned by ComponentTester.getTag(Component)
     * @param title Owning Frame/Dialog title, or null
     * @param parent Reference to parent, or null
     * @param index Index within parent, or -1
     */
    public ComponentReference(String id, Class compClass,
                              String name, String tag, String title,
                              ComponentReference parent, int index) {
        this(id, compClass, name, tag, title, parent, index, null);
    }

    /** For general component lookup.
     * @param id Desired ID for the reference.  Only used if this reference is
     * to be passed as the parent, window, or invoker of another.
     * @param compClass Class of the component (required)
     * @param name Name of the component, or null
     * @param tag Tag as returned by ComponentTester.getTag(Component)
     * @param title Owning Frame/Dialog title, or null
     * @param parent Reference to parent, or null
     * @param index Index within parent, or -1
     * @param invokerOrWindow Window reference, invoker, or null
     */
    public ComponentReference(String id, Class compClass,
                              String name, String tag, String title,
                              ComponentReference parent, int index,
                              ComponentReference invokerOrWindow) {
        // hack, kinda
        resolver = parent != null ? parent.resolver : new Script();
        this.id = id;
        this.refClassName = compClass.getName();
        this.name = name;
        this.tag = tag;
        this.parentID = parent != null ? parent.getID() : null;
        this.index = index;
        if (JPopupMenu.class.isAssignableFrom(compClass)) {
            invokerID = invokerOrWindow != null
                ? invokerOrWindow.getID() : null;
        }
        else {
            windowID = invokerOrWindow != null
                ? invokerOrWindow.getID() : null;
        }
        this.title = title;
    }

    /** For creation from XML. */
    public ComponentReference(Resolver resolver, Element el) 
        throws InvalidScriptException {
        this.resolver = resolver;
        this.finder = DefaultComponentFinder.getFinder();
        fromXML(el, true);
    }

    /** Create a reference based on the given component.  May recursively add
     * other components required to identify this one.
     */
    public ComponentReference(Resolver resolver, Component comp) {
        this.resolver = resolver;
        this.finder = DefaultComponentFinder.getFinder();

        Class refClass = comp.getClass();
        Container parent = comp.getParent();
        Component win = null;

        refClass = ComponentTester.getCanonicalClass(refClass);
        refClassName = refClass.getName();
        name = finder.getComponentName(comp);
        tag = ComponentTester.getTag(comp);        
        // If there's no tag, we need to store the parent in order to get
        // enough identifying information.
        
        if (comp instanceof JPopupMenu) {
            ComponentReference ref =
                resolver.addComponent(((JPopupMenu)comp).getInvoker());
            invokerID = ref.getID();
        }

        if (null != parent) {
            Component[] children = parent.getComponents();
            for (int i = 0; i < children.length; ++i) {
                if (children[i] == comp) {
                    index = i;
                    break;
                }
            }
            if (tag == null) {
                ComponentReference ref = resolver.addComponent(parent);
                parentID = ref.getID();
            }
        }

        // Find a frame title, if one is available
        // Popup menus with a non-JMenu invoker shouldn't store a frame title.
        // They provide a tag which will store the invoking component instead.
        if (!(JPopupMenu.class.isAssignableFrom(refClass)
              && JMenu.class.isAssignableFrom(((JPopupMenu)comp).
                                              getInvoker().getClass()))) {
            title = finder.getComponentWindowTitle(comp);
        }

        // If no title is available, save the parent window id instead
        if (title == null) {
            win = finder.getComponentWindow(comp);
            if (win != null && !(comp instanceof Window)) {
                ComponentReference wref = resolver.addComponent(win);
                this.windowID = wref != null ? wref.id : null;
            }
        }
        
        // Ensure that what we just created can be used to look up the
        // component.  If the component is within a popup, don't do a finder
        // lookup since the popup may already have gone away.
        if (ComponentTester.isOnPopup(comp)) {
            if (!finder.componentsMatch(comp, this,
                                        ComponentFinder.MATCH_EXACT)) {
                throw new Error("Generated reference does not match original "
                                + "component (" + toString() + " vs. "
                                + ComponentTester.toString(comp));
            }
        }
        else try {
            // FIXME if the component has just become not visible, this lookup
            // will fail!  Need a lookup flag for visible components?
            finder.findComponent(this);
        }
        catch (MultipleComponentsFoundException e) {
            // More than one match found, so add more information
            if (parent != null && parentID == null) {
                ComponentReference ref = resolver.addComponent(parent);
                parentID = ref.getID();
                // Try the lookup again to make sure it works this time
                try { 
                    finder.findComponent(this);
                }
                catch(Exception exc) {
                    Log.warn(exc);
                    throw new Error("Reverse lookup failed for " + toString()
                                    + " trying to match "
                                    + ComponentTester.toString(comp));
                }
            }
        }
        catch (ComponentNotFoundException cnf) {
            // This indicates a failure in the reference recording mechanism,
            // and requires a fix.
            throw new Error("Reverse lookup failed for " + toString() 
                            + " trying to match "
                            + ComponentTester.toString(comp));
        }

        // Finally, get a unique ID for this reference
        id = resolver.getUniqueID(this);
        Log.debug("Unique ID is " + id);
    }

    /** Unique identifier for this component. */
    public String getID() { return id; }

    /** This component's name, null if no name was set. */
    public String getName() { return name; }

    /** This component's class name. */
    public String getRefClassName() { return refClassName; }

    /** Return whether this reference has the same class or is a superclass of
     * the given component's class.  Simply compare class names to avoid class
     * loader conflicts.  Note that this does not take into account interfaces
     * (which is okay, since with GUI components we're only concerned with
     * class inheritance). 
     */
    public boolean isAssignableFrom(Class cls) {
        return cls != null
            && (refClassName.equals(cls.getName())
                || isAssignableFrom(cls.getSuperclass()));
    }

    /** Reference ID of this component's parent (optional). */
    public String getParentID() { return parentID; }
    public ComponentReference getParentReference() {
        if (parentID != null && parentReference == null) {
            parentReference = resolver.getComponentReference(parentID);
        }
        return parentReference;
    }

    /** Index among parent's children (optional). */
    public int getIndex() { return index; }

    /** Reference ID of this component's parent window (optional). */
    public String getWindowID() { return windowID; }
    public ComponentReference getWindowReference() {
        if (windowID != null && windowReference == null) {
            windowReference = resolver.getComponentReference(windowID);
        }
        return windowReference;
    }

    /** Title string of this component's parent frame (optional). */
    public String getTitle() { return title; }

    /** Invoker of a JPopupMenu. */
    public String getInvokerID() { return invokerID; }
    public ComponentReference getInvokerReference() {
        if (invokerID != null && invokerReference == null) {
            invokerReference = resolver.getComponentReference(invokerID);
        }
        return invokerReference;
    }

    public Point getInvocationLocation() { return point; }

    /** Returns a component class-specific tag used to match this
        reference to a real component. */
    public String getTag() { return tag; }

    /** Set all options based on the given XML. */
    public void fromXML(String xml) throws InvalidScriptException {
        StringReader reader = new StringReader(xml);
        try {
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(reader);
            Element el = doc.getRootElement();
            if (el == null)
                throw new InvalidScriptException("Invalid ComponentReference"
                                                 + " XML '" + xml + "'");
            fromXML(el, false);
        }
        catch(JDOMException e) {
            throw new InvalidScriptException(e.getMessage()
                                             + " (when parsing " + xml + ")");
        }
    }

    /** Parse settings from the given XML.   Only overwrite the ID if
        useGivenID is set.
        @throws InvalidScriptException if the given Element is not valid XML
        for a ComponentReference.
    */
    private void fromXML(Element el, boolean useGivenID)
        throws InvalidScriptException {
        String id = this.id;
        String tag = null;
        String name = null;
        String parentID = null;
        int index = -1;
        String windowID = null;
        String title = null;
        String invokerID = null;
        Point point = new Point(-1,-1);

        Iterator iter = el.getAttributes().iterator();
        while (iter.hasNext()) {
            Attribute att = (Attribute)iter.next();
            String nodeName = att.getName();
            String value = att.getValue();
            if (nodeName.equals(TAG_ID)) {
                if (useGivenID) {
                    id = value;
                }
            }
            else if (nodeName.equals(TAG_TAG)) {
                tag = value;
            }
            else if (nodeName.equals(TAG_NAME)) {
                name = value;
            }
            else if (nodeName.equals(TAG_CLASS)) {
                refClassName = value;
            }
            else if (nodeName.equals(TAG_PARENT)) {
                parentID = value;
            }
            else if (nodeName.equals(TAG_INDEX)) {
                try { index = Integer.parseInt(value); }
                catch(NumberFormatException nfe) {
                    throw new InvalidScriptException(nfe.getMessage(), el);
                }
            }
            else if (nodeName.equals(TAG_WINDOW)) {
                windowID = value;
            }
            else if (nodeName.equals(TAG_TITLE)) {
                title = value;
            }
            else if (nodeName.equals(TAG_INVOKER)) {
                invokerID = value;
            }
            else if (nodeName.equals(TAG_X)) {
                try { point.x = Integer.parseInt(value); }
                catch(NumberFormatException nfe) {
                    throw new InvalidScriptException(nfe.getMessage(), el);
                }
            }
            else if (nodeName.equals(TAG_Y)) {
                try { point.y = Integer.parseInt(value); }
                catch(NumberFormatException nfe) {
                    throw new InvalidScriptException(nfe.getMessage(), el);
                }
            }
        }
        if (refClassName == null) {
            throw new InvalidScriptException("Class must be specified", el);
        }
        if (useGivenID) {
            // Make sure the ID we read in is not already in use by the manager
            if (id != null) {
                if (resolver.getComponentReference(id) != null) {
                    String msg = "Persistent ID '" 
                        + id + "' is already in use";
                    throw new InvalidScriptException(msg, el);
                }
            }
        }
        this.tag = tag;
        this.name = name;
        this.parentID = parentID;
        this.index = index;
        this.windowID = windowID;
        this.title = title;
        this.invokerID = invokerID;
        this.point = point;

        if (id == null)
            id = resolver.getUniqueID(this);
        this.id = id;
    }

    public Element toXML() {
        Element el = new Element(TAG_COMPONENT);
        el.setAttribute(TAG_ID, id);
        el.setAttribute(TAG_CLASS, refClassName);
        if (name != null)
            el.setAttribute(TAG_NAME, name);
        if (windowID != null) 
            el.setAttribute(TAG_WINDOW, windowID);
        if (parentID != null) 
            el.setAttribute(TAG_PARENT, parentID);
        if (index != -1) 
            el.setAttribute(TAG_INDEX, String.valueOf(index));
        if (title != null) 
            el.setAttribute(TAG_TITLE, title);
        if (tag != null) 
            el.setAttribute(TAG_TAG, tag);
        if (invokerID != null) {
            el.setAttribute(TAG_INVOKER, invokerID);
            if (point.x != -1 && point.y != -1) {
                el.setAttribute(TAG_X, String.valueOf(point.x));
                el.setAttribute(TAG_Y, String.valueOf(point.y));
            }
        }
        return el;
    }

    public String toEditableString() {
        return Step.toXMLString(this);
    }

    /** Two ComponentReferences with identical XML representations should have
        the same hash code. */
    public int hashCode() {
        return toEditableString().hashCode();
    }

    /** Two ComponentReferences with identical XML representations should 
        be equal. */
    public boolean equals(Object obj) {
        return this == obj
            || (obj != null && (obj instanceof ComponentReference)
                && toEditableString().equals(((ComponentReference)obj).toEditableString()));
    }

    public String toString() {
        String str = id != null ? id : (refClassName + " (no id yet)");
        if (str.indexOf("Instance") == -1)
            str += " (" + refClassName + ")";
        return str;
    }
}
