package abbot.script;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Stack;
import java.util.StringTokenizer;

import abbot.AssertionFailedError;
import abbot.Log;
import abbot.i18n.Strings;
import abbot.util.ProcessOutputHandler;
import abbot.util.Properties;

/**
 * A StepRunner that runs the step in a separate VM.  Behavior should be
 * indistinguishable from the base StepRunner.
 */
public class ForkedStepRunner extends StepRunner {

    int LAUNCH_TIMEOUT = Properties.getProperty("abbot.runner.launch_delay",
                                                0, 300000, 60000);
    int TERMINATE_TIMEOUT =
        Properties.getProperty("abbot.runner.terminate_delay",
                               0, 300000, 30000);

    private static ServerSocket serverSocket = null;
    private Process process = null;
    private Socket connection = null;

    /** When actually within the separate VM, this is what gets run. */
    protected static class SlaveProcess extends StepRunner {
        private Socket connection = null;
        private Script script = null;
        /** Translate the given event into something we can send back to the
         * master.
         */
        private void forwardEvent(StepEvent event) {
            Step step = event.getStep();
            final StringBuffer sb = new StringBuffer(encodeStep(script, step));
            sb.append("\n");
            sb.append(event.getType()); sb.append("\n");
            sb.append(event.getID()); 
            Throwable thr = event.getError();
            if (thr != null) {
                sb.append("\nMSG:"); sb.append(thr.getMessage()); 
                sb.append("\nSTR:"); sb.append(thr.toString()); 
                sb.append("\nTRC:"); 
                StringWriter writer = new StringWriter();
                thr.printStackTrace(new PrintWriter(writer));
                sb.append(writer.toString());
            }
            try {
                writeMessage(connection, sb.toString());
            }
            catch(IOException io) {
                // nothing we can do
            }
        }
        /** Handle running a script as a forked process. */
        public void launchSlave(int port) {
            // make connection back to originating port
            try {
                InetAddress local = InetAddress.getLocalHost();
                connection = new Socket(local, port);
            }
            catch(Throwable thr) {
                // Can't communicate so the only option is to quit
                Log.warn(thr);
                System.exit(1);
            }
            script = new Script();
            try {
                String contents = readMessage(connection);
                script.load(new StringReader(contents));
                Log.debug("Successfully loaded script");
                // Make sure we only fork once!
                script.setForked(false);
            }
            catch(IOException io) {
                Log.warn(io);
                System.exit(2);
            }
            catch(Throwable thr) {
                Log.debug(thr);
                StepEvent event = new StepEvent(script, StepEvent.STEP_ERROR,
                                                0, thr);
                forwardEvent(event);
            }
            
            // add listener to send messages back to the master
            addStepListener(new StepListener() {
                public void stateChanged(StepEvent ev) {
                    forwardEvent(ev);
                }
            });
            
            // Run the script like we normally would.  The listener handles
            // all events and communication back to the launching process
            try {
                // Explicit qualification to avoid javac error (bug in javac)
                SlaveProcess.this.runStep(script);
            }
            catch(Throwable thr) {
                // Listener catches all events and forwards them, so no one
                // else is interested.  just quit.
                Log.debug(thr);
            }
            
            try {
                connection.close();
            }
            catch(IOException io) {
                // not much we can do
                Log.warn(io);
            }
            // Return zero even on failure/error, since the script run itself
            // worked, regardless of test results.
            System.exit(0);
        }
    }

    Process fork(String vmargs, String[] cmdArgs) throws IOException {
        String java = System.getProperty("java.home")
            + File.separator + "bin"
            + File.separator + "java";
        ArrayList args = new ArrayList();
        args.add(java);
        args.add("-cp");
        // FIXME make sure this isn't the modified one!
        args.add(System.getProperty("java.class.path"));
        if (vmargs != null) {
            StringTokenizer st = new StringTokenizer(vmargs);
            while (st.hasMoreTokens()) {
                args.add(st.nextToken());
            }
        }
        args.addAll(Arrays.asList(cmdArgs));
        if (Log.isClassDebugEnabled(getClass())) {
            args.add("--debug");
            args.add(getClass().getName());
        }
        cmdArgs = (String[])args.toArray(new String[args.size()]);
        Process p = Runtime.getRuntime().exec(cmdArgs);
        return p;
    }

    /** Launch a new process, using this class as the main class. */
    Process fork(String vmargs) throws UnknownHostException, IOException {
        if (serverSocket == null) {
            serverSocket = new ServerSocket(0);
        }
        int localPort = serverSocket.getLocalPort();
        String[] args = {
            getClass().getName(), String.valueOf(localPort),
        };
        Process p = fork(vmargs, args);
        new ProcessOutputHandler(Strings.get("ForkSlave"), p).start();
        return p;
    }

    /** Running the step in a separate VM should be indistinguishable from
     * running a regular script.   When running as master, nothing actually
     * runs locally.  We just fork a subprocess and run the script in that,
     * reporting back its progress as if it were running locally.
     */
    public void runStep(Step step) {
        // Fire the start event prior to forking, then ignore the subsequent
        // forked script start event when we get it.
        fireStepStart(step);
        process = null;
        try {
            Script script = (Script)step;
            process = forkProcess(script.getVMArgs());
            sendScript(script);
            try {
                trackScript(script);
                int code = process.exitValue();
            }
            catch(IOException io) {
                fireStepError(script, io);
            }
            catch(IllegalThreadStateException its) {
                try { Thread.sleep(TERMINATE_TIMEOUT); }
                catch(InterruptedException ie) { }
                // check again?
            }
        }
        catch(Throwable thr) {
            fireStepError(step, thr);
        }

        // Destroy it whether it's terminated or not.
        if (process != null)
            process.destroy();
        fireStepEnd(step);
    }

    private Process forkProcess(String vmargs) throws Throwable {
        try {
            // fork new VM
            // wait for connection
            Process p = fork(vmargs);
            serverSocket.setSoTimeout(LAUNCH_TIMEOUT);
            connection = serverSocket.accept();
            Log.debug("Got slave connection on " + connection);
            return p;
        }
        catch(InterruptedIOException ie) {
            Log.warn(ie);
            throw new RuntimeException(Strings.get("SlaveTimedOut"));
        }
    }

    private void sendScript(Script script) throws IOException {
        // send script data
        StringWriter writer = new StringWriter();
        script.save(writer);
        writeMessage(connection, writer.toString());
        Log.debug("Successfully sent script");
    }

    private void trackScript(Script script) throws IOException {
        StepEvent ev;
        Stack stack = new Stack();
        stack.push(script);
        while ((ev = receiveEvent(script)) != null && !stopped()) {
            // If it's the script start event, ignore it since we
            // already sent one prior to launching the process
            if (ev.getStep() == script
                && (ev.STEP_START.equals(ev.getType())
                    || ev.STEP_END.equals(ev.getType()))) {
                continue;
            }
            else if (ev.getStep() instanceof Sequence) {
                Step step = ev.getStep();
                if (ev.STEP_START.equals(ev.getType()))
                    stack.push(step);
                else if (ev.STEP_END.equals(ev.getType())) {
                    stack.pop();
                    // At the end of a sequence, propagate its errors
                    if (getError(step) != null) {
                        setError((Step)stack.peek(), getError(step));
                    }
                }
            }

            Log.debug("Replaying forked event " + ev);
            Throwable err = ev.getError();
            if (err != null) {
                setError(ev.getStep(), err);
                setError((Step)stack.peek(), getError(ev.getStep()));
                if (err instanceof AssertionFailedError) {
                    if (getStopOnFailure())
                        stop();
                }
                else {
                    if (getStopOnError())
                        stop();
                }
            }
            fireStepEvent(ev);
        }
    }

    static Step decodeStep(Sequence root, String code) {
        if (code.equals("-1"))
            return root;
        int comma = code.indexOf(",");
        if (comma == -1) {
            // Let number format exceptions propagate up, since it's a fatal
            // script error.
            int index = Integer.parseInt(code);
            return root.getStep(index);
        }
        String ind = code.substring(0, comma);
        code = code.substring(comma + 1);
        return decodeStep((Sequence)root.getStep(Integer.parseInt(ind)), code);
    }

    /** Encode the given step into a set of indices. */
    static String encodeStep(Sequence root, Step step) {
        if (root.equals(step))
            return "-1";
        synchronized(root.steps()) {
            int index = root.indexOf(step);
            if (index != -1)
                return String.valueOf(index);
            index = 0;
            Iterator iter = root.steps().iterator();
            while (iter.hasNext()) {
                Step seq = (Step)iter.next();
                if (seq instanceof Sequence) {
                    String encoding = encodeStep((Sequence)seq, step);
                    if (encoding != null) {
                        return index + "," + encoding;
                    }
                }
                ++index;
            }
            return null;
        }
    }

    /** Receive a serialized event on the connection and convert it back into
     * a real event, setting the given step's exception/error if necessary.
     */
    private StepEvent receiveEvent(Script script) throws IOException {
        String buf = readMessage(connection);
        if (buf == null)
            return null; // end of stream
        StringTokenizer st = new StringTokenizer(buf, "\n");
        String code = st.nextToken();
        String type = st.nextToken();
        String id = st.nextToken();
        Step step = decodeStep(script, code);
        Throwable thr = null;
        if (st.hasMoreTokens()) {
            String msg = st.nextToken();
            String string;
            String trace;
            msg = msg.substring(4);
            String next = st.nextToken();
            while (!next.startsWith("STR:")) {
                msg += next;
                next = st.nextToken();
            }
            string = next.substring(4);
            next = st.nextToken();
            while (!next.startsWith("TRC:")) {
                string += next;
                next = st.nextToken();
            }
            trace = next.substring(4);

            if (type.equals(StepEvent.STEP_FAILURE)) {
                Log.debug("Setting step failure");
                thr = new ForkedFailure(msg, string, trace);
            }
            else {
                Log.debug("Setting step error");
                thr = new ForkedError(msg, string, trace);
            }
        }
        StepEvent event =
            new StepEvent(step, type, Integer.parseInt(id), thr);
        return event;
    }

    private static void writeMessage(Socket connection, String msg)
        throws IOException {
        OutputStream os = connection.getOutputStream();
        byte[] buf = msg.getBytes();
        int len = buf.length;
        for (int i=0;i < 4;i++) {
            byte val = (byte)(len >> 24);
            os.write(val);
            len <<= 8;
        }
        os.write(buf, 0, buf.length);
        os.flush();
    }

    private static String readMessage(Socket connection) throws IOException {
        InputStream is = connection.getInputStream();
        // FIXME probably want a socket timeout, in case the slave script
        // hangs 
        int len = 0;
        for (int i=0;i < 4;i++) {
            int data = is.read();
            if (data == -1) {
                return null;
            }
            len = (len << 8) | data;
        }
        byte[] buf = new byte[len];
        int offset = 0;
        while (offset < len) {
            int count = is.read(buf, offset, buf.length - offset);
            if (count == -1) {
                return null;
            }
            offset += count;
        }
        String msg = new String(buf, 0, len);
        return msg;
    }

    /** An exception that for all purposes looks like another exception. */
    class ForkedFailure extends AssertionFailedError {
        private String msg;
        private String str;
        private String trace;
        public ForkedFailure(String msg, String str, String trace) {
            this.msg = msg + " (forked)";
            this.str = str + " (forked)";
            this.trace = trace + " (forked)";
        }
        public String getMessage() { return msg; }
        public String toString() { return str; }
        public void printStackTrace(PrintWriter writer) {
            synchronized(writer) {
                writer.print(trace);
            }
        }
        public void printStackTrace(PrintStream s) {
            synchronized(s) {
                s.print(trace);
            }
        }
        public void printStackTrace() {
            printStackTrace(System.err);
        }
    }

    /** An exception that for all purposes looks like another exception. */
    class ForkedError extends RuntimeException {
        private String msg;
        private String str;
        private String trace;
        public ForkedError(String msg, String str, String trace) {
            this.msg = msg + " (forked)";
            this.str = str + " (forked)";
            this.trace = trace + " (forked)";
        }
        public String getMessage() { return msg; }
        public String toString() { return str; }
        public void printStackTrace(PrintWriter writer) {
            synchronized(writer) {
                writer.print(trace);
            }
        }
        public void printStackTrace(PrintStream s) {
            synchronized(s) {
                s.print(trace);
            }
        }
        public void printStackTrace() {
            printStackTrace(System.err);
        }
    }

    /** Provide means to control execution and feedback of a script in a
        separate process.
    */ 
    public static void main(String[] args) {
        args = Log.init(args);
        try {
            final int port = Integer.parseInt(args[0]);
            new Thread(new Runnable() {
                public void run() {
                    new SlaveProcess().launchSlave(port);
                }
            }, "Forked script").start();
        }
        catch(Throwable e) {
            System.err.println("usage: abbot.script.ForkedStepRunner <port>");
            System.exit(1);
        }
    }
}
