package abbot.script;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.HashMap;

import abbot.Log;
import abbot.NoExitSecurityManager;
import abbot.Platform;
import abbot.Resolver;
import abbot.i18n.Strings;

/** 
 * Provides scripted static method invocation.  Usage:<br>
 * <blockquote><code>
 *  &lt;launch class="package.class" method="methodName" args="..."
 * [classpath="..."] [delegate=false] [threaded=true]&gt;<br>
 * </code></blockquote><p> 
 * The args attribute is a comma-separated list of arguments to pass to the
 * class method, and may use square brackets to denote an array,
 * e.g. "[one,two,three]" will be interpreted as an array length 3 
 * of String.  The square brackets may be escapted ('\[' or '\]') to include
 * them literally in an argument.<p>
 * The class path attribute may use either colon or semicolon as a path
 * separator, but should preferably use relative paths to avoid making the
 * containing script platform- and location-dependent.<p>
 * In most cases, the classes under test will <i>only</i> be found under the
 * custom class path, and so the parent class loader will fail to find them.
 * If the class under test is in the JRE (appletviewer, for example), or if
 * the testing class path contains duplicates of classes under test (different
 * versions of xerces or gnu-regexp, for example), then the delegate flag
 * needs to explicitly be set false to ensure the dynamic class loader
 * <i>always</i> tries to load the classes under test before the parent class
 * loader.<p>
 * The 'threaded' attribute is provided in case your code under test requires
 * GUI event processing prior to returning from its invoked method.  An
 * example might be a main method which invokes dialog and waits for the
 * response before continuing.  In general, it's better to refactor the code
 * if possible so that the main method turns over control to the event
 * dispatch thread as soon as possible.  Otherwise, if the application under
 * test is background threaded by the Launch step, any runtime exceptions
 * thrown from the launch code will cause errors in the launch step out of
 * sequence with the other script steps.  While this won't cause any problems
 * for the Abbot framework, it can be very confusing for the user.<p>
 * Note that if the "reload" attribute is set true (i.e. Abbot's class loader
 * is used to reload code under test), ComponentTester extensions must also be
 * loaded by that class loader, so the path to extensions should be included
 * in the Launch class path.<p> 
 */

public class Launch extends Call {

    private String classpath = null;
    private boolean delegate = true;
    private boolean threaded = false;
    private ClassLoader classLoader = null;

    private static final String USAGE = 
        "<launch class=\"...\" method=\"...\" args=\"...\" "
        + "[delegate=false] [threaded=true]>";

    public Launch(Resolver resolver, HashMap attributes) {
        super(resolver, attributes);
        classpath = (String)attributes.get(TAG_CLASSPATH);
        String rl = (String)attributes.get(TAG_DELEGATE);
        if (rl != null)
            delegate = Boolean.valueOf(rl).booleanValue();
        String thr = (String)attributes.get(TAG_THREADED);
        if (thr != null)
            threaded = Boolean.valueOf(thr).booleanValue();
    }

    public Launch(Resolver resolver, 
                  String description, String className,
                  String methodName, String[] args) {
        this(resolver, description, className, methodName, args,
             null, false, false);
    }

    public Launch(Resolver resolver, 
                  String description, String className, String methodName,
                  String[] args, String classpath,
                  boolean delegate, boolean threaded) {
        super(resolver, description, className, methodName, args);
        this.classpath = classpath;
        this.delegate = delegate;
        this.threaded = threaded;
    }

    public String getClasspath() {
        return classpath;
    }

    public void setClasspath(String cp) {
        classpath = cp;
    }

    public boolean isThreaded() {
        return threaded;
    }

    public void setThreaded(boolean thread) {
        threaded = thread;
    }

    public boolean delegates() {
        return delegate;
    }

    public void setDelegates(boolean delegate) {
        this.delegate = delegate;
    }

    /** Install a security manager if there is none; this is a workaround
     * to prevent sun's applet viewer's security manager from preventing
     * any classes from being loaded.
     */
    private void installAppletSecurityManager(ClassLoader cl) {
        SecurityManager sm = System.getSecurityManager();
        if (sm == null || (sm instanceof NoExitSecurityManager)) {
            // NOTE: the security manager *must* be loaded with the same class
            // loader as the appletviewer.
            try {
                Class cls = Class.forName("abbot.script.AppletSecurityManager",
                                          true, cl);
                Constructor ctor = cls.getConstructor(new Class[] {
                    SecurityManager.class
                });
                sm = (SecurityManager)ctor.newInstance(new Object[] { sm });
                System.setSecurityManager(sm);
            }
            catch(Exception exc) {
                Log.warn(exc);
            }
        }
    }

    /** Augment the default run step. */
    public void runStep() throws Throwable {

        boolean isAppletViewer =
            "sun.applet.AppletViewer".equals(getTargetClassName());

        // AppletViewer will only run properly once unless it's reloaded each
        // time. 
        if (isAppletViewer && delegate) {
            // FIXME use a dialog
            Log.warn("appletviewer must have class reloading enabled");
            delegate = false;
        }

        // Use a custom class loader so that we can provide additional
        // classpath and also optionally reload the class on each run.
        // FIXME maybe classpath should be relative to the script? 
        AppClassLoader cl = new AppClassLoader(classpath, delegate);

        if (isAppletViewer) {
            // need to make sure applets get the security manager they expect
            installAppletSecurityManager(cl);
        }

        // Everything else loaded on the same thread as this 
        // launch should be loaded by this custom loader.  
        cl.install();
        classLoader = cl;

        // A bug in pre-1.4 VMs locks the toolkit prior to notifying AWT event
        // listeners.  This causes a deadlock when the main method invokes
        // "show" on a component which triggers AWT events for which there are
        // listeners.  To avoid this, grab the toolkit lock first so that the
        // locks are acquired in the same order by either sequence.
        // (Unfortunately, some swing code locks the tree prior to
        // grabbing the toolkit lock, so there's still opportunity for
        // deadlock).  One alternative (although very heavyweight) is to
        // always fork a separate VM.
        //
        // If threaded, take the danger of deadlock over the possibility that
        // the main method will never return and leave the lock forever held.
        // NOTE: this is guaranteed to deadlock if "main" calls
        // EventQueue.invokeAndWait. 
        if (Platform.getJavaVersionNumber() < Platform.JAVA_1_4
            && !isThreaded()) {
            synchronized(java.awt.Toolkit.getDefaultToolkit()) {
                super.runStep();
            }
        }
        else {
            super.runStep();
        }
    }

    public ClassLoader getClassLoader() {
        return classLoader;
    }

    public Class getTargetClass() throws InvalidScriptException {
        Class cls = resolveClass(getTargetClassName());
        Log.debug("Target class is " + cls.getName());
        return cls;
    }

    /** Return the target for the method invocation.  All launch invocations
     * must be static, so this always returns null.
     */ 
    protected Object getTarget() {
        return null;
    }

    /** Return the method to be used for invocation. */
    protected Method getMethod() throws InvalidScriptException {
        return resolveMethod(getMethodName(), getTargetClass(), null);
    }

    public HashMap getAttributes() {
        HashMap map = super.getAttributes();
        if (classpath != null) {
            map.put(TAG_CLASSPATH, classpath);
        }
        if (!delegate) {
            map.put(TAG_DELEGATE, "false");
        }
        if (threaded) {
            map.put(TAG_THREADED, "true");
        }
        return map;
    }

    protected String getDefaultDescription() {
        String desc = Strings.get("LaunchDesc",
                                  new Object[] { getTargetClassName()
                                                 + "." + getMethodName()
                                                 + "(" + getEncodedArguments()
                                                 + ")"});
        return desc;
    }

    public String getUsage() { return USAGE; }

    public String getXMLTag() { return TAG_LAUNCH; }
}
