package abbot.script;

import java.util.HashMap;

import abbot.Resolver;

/** Encapsulate capture of a value.  Usage:<br>
 * <blockquote><code>
 * &lt;sample property="..." method="assertXXX" [class="..."]&gt;<br>
 * &lt;sample property="..." method="(get|is|has)XXX" component="component_id"&gt;<br>
 * </code></blockquote>
 * The <b>sample</b> step stores the value found from the given method
 * <a href="{@docRoot}/abbot/tester/ComponentTester.html">ComponentTester</a>
 * class; the class tag is required for methods based on a class derived from
 * ComponentTester; the class tag indicates the Component class, not hte
 * Tester class (teh appropriate tester class will be derived automatically).
 * The second format indicates a property sample on the given component.
 * In both cases, the result of the invocation will be saved in the current
 * Resolver as a property with the given property name.
 */

public class Sample extends PropertyCall {
    private static final String USAGE = 
        "<sample property=... component=... method=.../>\n"
        + "<sample property=... method=... [class=...]/>";

    private String propertyName = null;

    public Sample(Resolver resolver, HashMap attributes) {
        super(resolver, attributes);
        propertyName = (String)attributes.get(TAG_PROPERTY);
    }

    /** Component property sample. */
    public Sample(Resolver resolver, String description,
                  String methodName, String[] args,
                  String id, String propName) {
        super(resolver, description, methodName, args, id);
        propertyName = propName;
    }

    /** Static method property sample. */
    public Sample(Resolver resolver, String description,
                  String className, String methodName, String[] args,
                  String propName) {
        super(resolver, description, className, methodName, args);
        propertyName = propName;
    }

    public HashMap getAttributes() {
        HashMap map = super.getAttributes();
        if (propertyName != null) {
            map.put(TAG_PROPERTY, propertyName);
        }
        return map;
    }

    protected String getDefaultDescription() {
        return getPropertyName() + "=" + super.getDefaultDescription();
    }

    public String getUsage() { return USAGE; }
    public String getXMLTag() { return TAG_SAMPLE; }

    public void setPropertyName(String name) {
        propertyName = name;
    }

    public String getPropertyName() {
        return propertyName;
    }

    /** Store the results of the invocation in the designated property. */
    protected Object invoke() throws Throwable {
        Object obj = super.invoke();
        if (propertyName != null) {
            // FIXME string-encode arrays
            String value = obj == null 
                ? "(null)" : obj.toString();
            getResolver().setProperty(propertyName, value);
        }
        return obj;
    }
}
