package abbot.script;

import java.awt.Component;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.WeakHashMap;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import abbot.ComponentFinder;
import abbot.Log;
import abbot.Resolver;
import abbot.tester.awt.ComponentTester;
import abbot.converter.Converter;
import abbot.converter.FileConverter;
import abbot.i18n.Strings;
import abbot.util.Properties;

/**
 * Provide a structure to encapsulate actions invoked on GUI components and
 * tests performed on those components.  Scripts need to be short and concise
 * (and therefore easy to read/write). Extensions don't have to be.<p> 
 * This takes a single filename as a constructor argument.<p>
 * Use
 * <a href="{@docRoot}/junit/extensions/abbot/ScriptTestCase.html">ScriptTestCase</a>
 * and
 * </a href="{@docRoot}/junit/extensions/abbot/ScriptTestSuite.html">ScriptTestSuite</a>
 * to generate a suite (auto-generate the collection).<p>  
 */

public class Script extends Sequence implements Resolver {

    private static final String USAGE = 
        "<AWTTestScript [desc=\"\"] [forked=\"true\"] [slow=\"true\"]"
        + " [vmargs=\"...\"]>...</AWTTestScript>\n";

    /** Robot delay for slow playback.  */
    private static int slowDelay = 200;
    private String filename;
    private File relativeDirectory = null;
    private boolean loaded = false;
    private boolean fork = false;
    private boolean slow = false;
    private int lastSaved = 0;
    private String vmargs = null;
    private HashMap properties = new HashMap();

    static {
        slowDelay = Properties.getProperty("abbot.script.slow_delay",
                                           0, 60000, slowDelay);
    }

    private static HashMap createDefaultHashMap(String filename) {
        HashMap map = new HashMap();
        map.put(TAG_FILENAME, filename);
        return map;
    }

    /** Create a new, empty script.  Used as a temporary resolver. */
    public Script() {
        this(null, new HashMap());
    }

    /** Create a script.  Actual reading of the script is deferred. */
    public Script(String filename) {
        this(null, createDefaultHashMap(filename));
    }

    public Script(Resolver parent, HashMap attributes) {
        super(parent, attributes);
        String filename = (String)attributes.get(TAG_FILENAME);
        File file = filename != null
            ? new File(filename)
            : getTempFile(parent != null ? parent.getDirectory() : null);
        setFile(file);
        if (parent != null) {
            setRelativeTo(parent.getDirectory());
        }
    }

    private File getTempFile(File dir) {
        File file;
        try {
            file = (dir != null
                    ? File.createTempFile("script", ".xml", dir)
                    : File.createTempFile("script", ".xml"));
        }
        catch(IOException io) {
            file = (dir != null
                    ? new File(dir, "tmpscript.xml")
                    : new File("tmpscript.xml"));
        }
        file.deleteOnExit();
        return file;
    }

    public String getName() {
        return filename;
    }

    public void setForked(boolean fork) {
        this.fork = fork;
    }

    public boolean isForked() { return fork; }

    public void setVMArgs(String args) {
        if (args != null && "".equals(args))
            args = null;
        vmargs = args;
    }
    public String getVMArgs() { return vmargs; }

    public boolean isSlowPlayback() { return slow; }

    public void setSlowPlayback(boolean slow) { this.slow = slow; }

    /** Return the file where this script is saved.  Will always be an
     * absolute path.
     */
    public File getFile() {
        File file = new File(filename);
        if (!file.isAbsolute()) {
            String path =
                getRelativeTo().getPath() + File.separator + filename;
            file = new File(path);
        }
        return file;
    }
    /** Warning: if the directory is changed, need to change associated files
        as well.
    */
    public void setFile(File outputFile) { 
        Log.debug("Output file set to " + outputFile);
        if (outputFile == null)
            throw new IllegalArgumentException("File must not be null");
        if (filename == null || !outputFile.equals(getFile())) {
            filename = outputFile.getPath();
            if (relativeDirectory != null)
                setRelativeTo(relativeDirectory);
            loaded = false;
        }
    }

    /** Recursively save any included scripts which may have been modified. */
    private void saveIncludedScripts(Sequence seq) throws IOException {
        if (seq instanceof Script) {
            ((Script)seq).save();
        }
        else {
            synchronized(seq.steps()) {
                Iterator iter = seq.steps().iterator();
                while (iter.hasNext()) {
                    Step step = (Step)iter.next();
                    if (step instanceof Sequence)
                        saveIncludedScripts((Sequence)step);
                }
            }
        }
    }

    /** Typical xml header, so we can know about how much file prefix to
     * skip.
     */
    private static final String XML_INFO =
        "<?xml version=\"1.0\" encoding=\"utf-8\"?>";

    private boolean exportAllData = false;
    // FIXME need to save changes to included scripts rather than
    // saving the scripts themselves
    public void save(Writer writer) throws IOException {
        exportAllData = true;
        Element el = toXML();
        exportAllData = false;
        el.setName(TAG_AWTTESTSCRIPT);
        // Filename doesn't need to be saved
        el.removeAttribute(TAG_FILENAME);

        Document doc = new Document(el);
        XMLOutputter outputter = new XMLOutputter("  ", true);
        outputter.output(doc, writer);
    }

    /** Only thing directly editable on a script is its file path. */
    public String toEditableString() {
        return getFilename();
    }

    /** Has this script changed since the last save. */
    public boolean isDirty() {
        return getHash() != lastSaved;
    }

    /** Write the script to file.  Note that this differs from the toXML for
        the script, which simply indicates the file on which it is based. */
    public void save() throws IOException {
        File file = getFile();
        Log.debug("Saving script to '" + file + "' " + hashCode());
        OutputStreamWriter writer =
            new OutputStreamWriter(new FileOutputStream(file), "UTF-8");
        save(new BufferedWriter(writer));
        writer.close();
        lastSaved = getHash();
    }

    /** Ensure that all referenced components are actually in the components
     * list.
     */
    private void verify() throws InvalidScriptException {
        Iterator iter = refs.values().iterator();
        while (iter.hasNext()) {
            ComponentReference ref = (ComponentReference)iter.next();
            if (ref.getParentID() != null
                && ref.getParentReference() == null) {
                String msg = Strings.get("ParentMissing", new Object[] { 
                    ref.getParentID() });
                throw new InvalidScriptException(msg);
            }
            if (ref.getWindowID() != null
                && ref.getWindowReference() == null) {
                String msg = Strings.get("WindowMissing", new Object[] { 
                    ref.getWindowID() });
                throw new InvalidScriptException(msg);
            }
        }
    }

    /** Make the path to the given child script relative to this one. */
    private void updateRelativePath(Step child) {
        // Make sure included scripts are located relative to this one
        if (child instanceof Script) {
            ((Script)child).setRelativeTo(getDirectory());
        }
    }

    protected void parseChild(Element el) throws InvalidScriptException {
        if (el.getName().equals(TAG_COMPONENT)) {
            addComponentReference(el);
        }
        else {
            synchronized(steps()) {
                super.parseChild(el);
                updateRelativePath((Step)steps().get(size()-1));
            }
        }
    }

    public void load(Reader reader)
        throws InvalidScriptException, IOException {
        clear();
        try {
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(reader);
            Element root = doc.getRootElement();
            HashMap map = createAttributeMap(root);
            parseStepAttributes(map);
            fork = Boolean.valueOf((String)map.get(TAG_FORKED)).booleanValue();
            slow = Boolean.valueOf((String)map.get(TAG_SLOW)).booleanValue();
            vmargs = (String)map.get(TAG_VMARGS);
            parseChildren(root);
        }
        catch(JDOMException e) {
            throw new InvalidScriptException(e.getMessage());
        }
        // Make sure we have all referenced components
        synchronized(refs) {
            verify();
        }
    }

    private void loadNoThrow() {
        try { load(); }
        catch(Throwable e) { setScriptError(e); }
    }

    public int size() {
        loadNoThrow();
        return super.size();
    }

    public java.util.List steps() {
        loadNoThrow();
        return super.steps();
    }

    public int indexOf(Step step) {
        loadNoThrow();
        return super.indexOf(step);
    }

    public Step getStep(int index) {
        loadNoThrow();
        return super.getStep(index);
    }

    public void addStep(int index, Step step) {
        loadNoThrow();
        super.addStep(index, step);
        updateRelativePath(step);
    }

    public void addStep(Step step) {
        loadNoThrow();
        super.addStep(step);
        updateRelativePath(step);
    }

    /** Replaces the step at the given index. */
    public void setStep(int index, Step step) {
        loadNoThrow();
        super.setStep(index, step);
        updateRelativePath(step);
    }

    public void removeStep(Step step) {
        loadNoThrow();
        super.removeStep(step);
    }

    public void removeStep(int index) {
        loadNoThrow();
        super.removeStep(index);
    }

    /** Parse the script file and build internal structures. */
    public void load() throws InvalidScriptException, IOException {
        if (!loaded) {
            File file = getFile();
            Log.debug("Loading script " + file
                      + " (" + hashCode() + ")");
            loaded = true;
            if (!file.exists()) {
                String msg = "The included script '" + getFilename()
                    + "' does not exist at the expected location '"
                    + file.getAbsolutePath() + "'";
                throw new InvalidScriptException(msg);
            }
            else if (file.length() != 0) {
                InputStreamReader reader =
                    new InputStreamReader(new FileInputStream(file), "UTF-8");
                try {
                    load(new BufferedReader(reader));
                }
                catch(InvalidScriptException e) {
                    loaded = false;
                    throw e;
                }
                catch(IOException io) {
                    loaded = false;
                    throw io;
                }
                finally {
                    reader.close();
                }
            }
            lastSaved = getHash();
        }
    }

    private int getHash() {
        try {
            exportAllData = true;
            return toXMLString(this).hashCode();
        }
        finally {
            exportAllData = false;
        }
    }

    public String getXMLTag() {
        return TAG_SCRIPT;
    }

    /** Save component references in addition to everything else. */
    public Element addContent(Element el) {
        // Only save content if writing to disk
        if (exportAllData) {
            synchronized (refs) {
                Iterator iter = refs.values().iterator();
                while (iter.hasNext()) {
                    ComponentReference cref = (ComponentReference)iter.next();
                    el.addContent(cref.toXML());
                }
            }
            // Now collect our child steps
            return super.addContent(el);
        }
        return el;
    }

    /** Return the (possibly relative) path to this script. */ 
    public String getFilename() { return filename; }

    /** Provide XML attributes for this Step.  This class adds its filename. */
    public HashMap getAttributes() {
        HashMap map = exportAllData ? super.getAttributes() : new HashMap();
        map.put(TAG_FILENAME, getFilename());

        if (exportAllData) {
            // default is no fork
            if (fork) {
                map.put(TAG_FORKED, "true");
                if (vmargs != null) {
                    map.put(TAG_VMARGS, vmargs);
                }
            }
            if (slow) {
                map.put(TAG_SLOW, "true");
            }
        }
        return map;
    }

    protected void runStep(StepRunner runner) throws Throwable {
        properties.clear();
        load();
        // Make all files relative to this script
        Converter fc = new FileConverter() {
            public String relativeTo() {
                Log.debug("All file references will be relative to "
                          + getDirectory().getAbsolutePath());
                return getDirectory().getAbsolutePath();
            }
        };
        Converter oldfc = ArgumentParser.setConverter(File.class, fc);
        int old = ComponentTester.getAutoDelay();
        if (slow) {
            ComponentTester.setAutoDelay(slowDelay);
        }
        try {
            super.runStep(runner);
        }
        finally {
            ComponentTester.setAutoDelay(old);
            ArgumentParser.setConverter(File.class, oldfc);
        }
    }

    /** Return a unique reference ID. */
    public String getUniqueID(ComponentReference ref) {
        // Use the component's name, if available
        String id = ref.getID();
        String cname = ref.getRefClassName();
        cname = cname.substring(cname.lastIndexOf(".") + 1);
        if (id == null) {
            // Don't ever use an empty string for the ID
            if (ref.getName() != null
                && !"".equals(ref.getName())) {
                id = ref.getName();
            }
            else if (ref.getTag() != null
                     && !"".equals(ref.getTag())) {
                id = ref.getTag();
            }
            else {
                id = cname + " Instance";
            }
        }

        String ext = "";
        int count = 1;
        while (refs.get(id + ext) != null) {
            ext = " " + count++;
        }
        return id + ext;
    }

    /** Set up a blank script, discarding any current state. */
    public void clear() {
        loadNoThrow();
        synchronized(refs) {
            refs.clear();
            components.clear();
        }
        super.clear();
    }

    public String getUsage() { return USAGE; }

    public String getDescription() {
        // If there's a description in the file itself, make sure we read it
        loadNoThrow();
        return super.getDescription();
    }

    protected String getDefaultDescription() {
        return toString();
    }

    public String toString() {
        String ext = fork ? " &" : "";

        return Strings.get("ScriptDesc", new Object[] { getFilename(), ext });
    }

    public boolean hasLaunch() {
        return size() > 0 && (((Step)steps().get(0)) instanceof Launch);
    }

    public boolean hasTerminate() {
        return size() > 0 
            && (((Step)steps().get(size()-1)) instanceof Terminate);
    }

    /** By default, all pathnames are relative to the current working
        directory.
    */
    public File getRelativeTo() {
        if (relativeDirectory == null)
            return new File(System.getProperty("user.dir"));
        return relativeDirectory;
    }

    /** Indicate that when invoking toXML, a path relative to the given one
     * should be shown.  Note that this is a runtime setting only and never
     * shows up in saved XML.
     */
    public void setRelativeTo(File dir) {
        Log.debug("Want relative dir " + dir);
        relativeDirectory = dir;
        if (dir != null) {
            // FIXME ideally, we'd want a more robust "make relative" here.
            // for now, simply check to see if the relpath is a prefix.
            // or if the file itself is relative
            File file = new File(filename);
            String relPath = dir.getPath();
            if (filename.startsWith(relPath)) {
                char ch = filename.charAt(relPath.length());
                if (ch == '/' || ch == '\\') {
                    filename = filename.substring(relPath.length() + 1);
                }
            }
        }
        Log.debug("Relative dir set to " + relativeDirectory + " for " + this);
    }

    /** Return whether the given file looks like a valid AWT script. */
    public static boolean isScript(File file) {
        if (file.length() == 0)
            return true;
        if (!file.exists() || !file.isFile() 
            || file.length() < TAG_AWTTESTSCRIPT.length() * 2 + 5)
            return false;
        FileInputStream is = null;
        try {
            int len = XML_INFO.length() + TAG_AWTTESTSCRIPT.length() + 15;
            is = new FileInputStream(file);
            byte[] buf = new byte[len];
            is.read(buf, 0, buf.length);
            String str = new String(buf, 0, buf.length);
            return str.indexOf(TAG_AWTTESTSCRIPT) != -1;
        }
        catch(Exception exc) {
            return false;
        }
        finally {
            if (is != null)
                try { is.close(); } catch(Exception exc) { }
        }
    }
    
    // NOTE: it may make sense to make the launch part of the script itself...
    public Launch getLaunchStep() {
        synchronized(steps()) {
            if (hasLaunch()) {
                return (Launch)steps().get(0);
            }
        }
        return null;
    }

    /** Maps a short mnemonic into a ComponentReference. */
    private HashMap refs = new HashMap();
    /** Maps actual components into ComponentReferences.  This speeds up
     * lookup when attempting to match a component to an existing reference.
     */
    private WeakHashMap components = new WeakHashMap();

    /** All relative files should be accessed relative to this directory,
        which is the directory where the script resides.
        It will always return an absolute path.
    */
    public File getDirectory() {
        return getFile().getParentFile();
    }

    public Collection getComponentReferences() {
        synchronized(refs) {
            return ((HashMap)refs.clone()).values();
        }
    }

    public void addComponentReference(ComponentReference ref) {
        synchronized(refs) {
            refs.put(ref.getID(), ref);
        }
    }

    /** Add a new component reference for the given component. */
    public ComponentReference addComponent(Component comp) {
        // See if we've already got it
        synchronized(refs) {
            ComponentReference ref =
                getComponentReference(comp, ComponentFinder.MATCH_EXACT);
            if (ref == null) {
                // Nope, create a new one
                /*Log.debug("Component " + ComponentTester.toString(comp)
                  + " not yet referenced, adding it");*/
                ref = new ComponentReference(this, comp);
                refs.put(ref.getID(), ref);
                components.put(comp, ref);
                if (ref.getID().equals("")) {
                    Log.warn("Reference id is empty: " + ref);
                }
            }
            else {
                /*Log.debug("Component " + ComponentTester.toString(comp)
                  + " already referenced, using " + ref);*/
            }
            return ref;
        }
    }

    /** Add a new component reference to the script.  For use only when
     * parsing a script.
     */
    ComponentReference addComponentReference(Element el)
        throws InvalidScriptException {
        synchronized(refs) {
            ComponentReference ref = new ComponentReference(this, el);
            refs.put(ref.getID(), ref);
            return ref;
        }
    }

    /** Return the reference for the given component, or null if none yet
     * exists. 
     */
    public ComponentReference getComponentReference(Component comp) {
        return getComponentReference(comp, ComponentFinder.MATCH_EXACT);
    }

    /** Return the reference for the given component, or null if none yet
     * exists. 
     */
    private ComponentReference getComponentReference(Component comp,
                                                     int requiredMatch) {
        synchronized(refs) {
            ComponentReference match = 
                (ComponentReference)components.get(comp);
            if (match == null) {
                match = getFinder().matchComponent(comp,
                                                   refs.values().iterator(),
                                                   requiredMatch);
                if (match != null) {
                    Log.debug("Cacheing component match");
                    components.put(comp, match);
                }
            }
            return match;
        }
    }

    /** Convert the given reference ID into a component reference.  If it's
     * not in the Script's list, throw an appropriate exception.
     */
    public ComponentReference getComponentReference(String name) {
        synchronized(refs) {
            Log.debug("Looking for reference with id '" + name + "' among "
                      + refs.size() + " references");
            return (ComponentReference)refs.get(name);
        }
    }

    public void setProperty(String name, String value) {
        properties.put(name, value);
    }

    public String getProperty(String name) {
        return (String)properties.get(name);
    }
}

