package abbot.script;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.SwingUtilities;

import abbot.AssertionFailedError;
import abbot.DefaultComponentFinder;
import abbot.ExitException;
import abbot.Log;
import abbot.NoExitSecurityManager;
import abbot.i18n.Strings;

public class StepRunner {
    private boolean stopOnFailure = true;
    private boolean stopOnError = true;
    private ArrayList listeners = new ArrayList();
    private HashMap errors = new HashMap();
    private Launch launchStep = null;
    /** Whether to stop running. */
    private boolean stop = false;

    /** Provides control and tracking of the execution of a step or series of
        steps.  By default the runner stops on the first encountered
        failure/error.
    */
    public StepRunner() {
        // Assume any existing windows are part of some running framework, and
        // ensure the script ignores them.
        DefaultComponentFinder.getFinder().discardAllComponents();
    }

    public void setStopOnFailure(boolean stop) {
        stopOnFailure = stop;
    }

    public void setStopOnError(boolean stop) {
        stopOnError = stop;
    }

    public boolean getStopOnFailure() { return stopOnFailure; }
    public boolean getStopOnError() { return stopOnError; }

    /** Stop execution of the script after the current step completes. */
    public void stop() { stop = true; }

    /** Return whether the runner has been stopped. */
    public boolean stopped() { return stop; }

    /** Run the given step, propagating any failures or errors. */
    public void run(Script script) throws Throwable {
        if (SwingUtilities.isEventDispatchThread()) {
            throw new Error(Strings.get("BadScriptInvoke"));
        }
        clearErrors();

        SecurityManager oldsm = System.getSecurityManager();
        String doInstall = System.getProperty("abbot.use_security_manager");
        boolean installed = false;
        if (oldsm == null && !"false".equals(doInstall)) {
            installed = true;
            // When the application tries to exit, throw control back to the
            // step runner to dispose of it
            Log.debug("Installing sm");
            System.setSecurityManager(new NoExitSecurityManager() {
                public void checkRead(String file) {
                    // avoid annoying drive a: bug on w32 VM
                }
                protected void terminate() {
                    StepRunner.this.terminate();
                }
            });
        }
        try {
            if (script.isForked()) {
                Log.debug("Forking " + script);
                StepRunner runner = new ForkedStepRunner();
                runner.listeners.addAll(listeners);
                try {
                    runner.runStep(script);
                }
                finally {
                    errors.putAll(runner.errors);
                }
            }
            else {
                runStep(script);
            }
        }
        catch(ExitException ee) {
            // application tried to exit
            Log.debug("App tried to exit");
            terminate();
        }
        finally {
            // restore the old one
            if (installed) {
                Log.debug("Uninstalling sm");
                System.setSecurityManager(oldsm);
            }
        }
    }

    protected void clearErrors() {
        stop = false;
        errors.clear();
    }

    private void runThreaded(final Step step) {
        Thread thread = new Thread("Background launch of " + step) {
            public void run() {
                try {
                    step.run();
                }
                catch(AssertionFailedError afe){
                    fireStepFailure(step, afe);
                    if (stopOnFailure)
                        StepRunner.this.stop();
                }
                catch(Throwable thr) {
                    fireStepError(step, thr);
                    if (stopOnError)
                        StepRunner.this.stop();
                }
            }
        };
        thread.setDaemon(true);
        thread.start();
    }

    /** Main run method, which stores any failures or exceptions for later
     * retrieval.  Any step will fire STEP_START on starting, and exactly one
     * of STEP_END, STEP_FAILURE, or STEP_ERROR upon termination.
     */
    // FIXME sequence used to fire progress events. do we still want them?
    void runStep(final Step step) throws Throwable {
        Log.debug("Running " + step);
        fireStepStart(step);
        // checking for stopped here allows a listener to stop execution on a
        // particular step.
        if (!stopped()) {
            try {
                if (step instanceof Sequence) {
                    ((Sequence)step).runStep(this);
                }
                else if (step instanceof Launch) {
                    launchStep = (Launch)step;
                    if (launchStep.isThreaded()) {
                        runThreaded(step);
                    }
                    else {
                        step.run();
                    }
                }
                else {
                    step.run();
                    if (step instanceof Terminate) {
                        terminate();
                    }
                }
            }
            catch(AssertionFailedError afe) {
                fireStepFailure(step, afe);
                if (stopOnFailure)
                    throw afe;
            }
            catch(Throwable thr) {
                fireStepError(step, thr);
                if (stopOnError)
                    throw thr;
            }
            fireStepEnd(step);
        }
    }

    /** Invoke the given script's launch step, if any. */
    public void launch(Script script) throws Throwable {
        if (SwingUtilities.isEventDispatchThread()) {
            throw new Error(Strings.get("BadScriptInvoke"));
        }
        clearErrors();

        Launch launch = script.getLaunchStep();
        if (launch != null) {
            Log.debug("Launching " + launch);
            runStep(launch);
        }
    }

    public void terminate() {
        // Remove all remaining windows and uninstall any custom class loader
        Log.debug("terminating code under test");
        DefaultComponentFinder.getFinder().disposeWindows();
        if (launchStep != null) {
            // While it's possible that a threaded launch may not have started
            // by this point, it's not likely.
            AppClassLoader dcl =
                (AppClassLoader)launchStep.getClassLoader();
            dcl.uninstall();
            launchStep = null;
        }
    }

    protected void setError(Step step, Throwable thr) {
        errors.put(step, thr);
    }

    public Throwable getError(Step step) {
        return (Throwable)errors.get(step);
    }

    public void addStepListener(StepListener sl) {
        synchronized(listeners) {
            listeners.add(sl);
        }
    }

    public void removeStepListener(StepListener sl) {
        synchronized(listeners) {
            listeners.remove(sl);
        }
    }
    
    /** If this is used to propagate a failure/error, be sure to invoke
     * setError on the step first.
     */
    protected void fireStepEvent(StepEvent event) {
        Iterator iter;
        synchronized(listeners) {
            iter = ((ArrayList)listeners.clone()).iterator();
        }
        while (iter.hasNext()) {
            StepListener sl = (StepListener)iter.next();
            sl.stateChanged(event);
        }
    }
    
    private void fireStepEvent(Step step, String type,
                                 int val, Throwable throwable) {
        synchronized(listeners) {
            if (listeners.size() != 0) {
                StepEvent event = new StepEvent(step, type, val, throwable);
                fireStepEvent(event);
            }
        }
    }
    
    protected void fireStepStart(Step step) {
        fireStepEvent(step, StepEvent.STEP_START, 0, null);
    }
    
    protected void fireStepProgress(Step step, int val) {
        fireStepEvent(step, StepEvent.STEP_PROGRESS, val, null);
    }

    protected void fireStepEnd(Step step) {
        fireStepEvent(step, StepEvent.STEP_END, 0, null);
    }

    protected void fireStepFailure(Step step, Throwable afe) {
        setError(step, afe);
        fireStepEvent(step, StepEvent.STEP_FAILURE, 0, afe);
    }

    protected void fireStepError(Step step, Throwable thr) {
        setError(step, thr);
        fireStepEvent(step, StepEvent.STEP_ERROR, 0, thr);
    }
}
