package abbot.tester.awt;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.IllegalComponentStateException;
import java.awt.Point;
import java.awt.event.AWTEventListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.util.Stack;

import abbot.Log;
import abbot.util.EventNormalizer;
import abbot.util.SingleThreadedEventListener;

/** Class to keep track of a given input state.  Includes mouse/pointer
 * position and keyboard modifier key state.<p>
 * Synchronization assumes that any given instance might be called from more
 * than one event dispatch thread.
 */

public class InputState {

    private static final int MOUSE_MASK = (MouseEvent.BUTTON1_MASK
                                           |MouseEvent.BUTTON2_MASK
                                           |MouseEvent.BUTTON3_MASK);
    
    /** Current mouse position, in component coordinates. */
    private Point mouseLocation = new Point(0, 0);
    /** Current mouse position, in screen coordinates. */
    private Point mouseLocationOnScreen = new Point(0, 0);
    /** Keep a stack of mouse-entered components.  Note that the pointer is
        still considered within a frame when the mouse enters a contained
        component.
    */
    private Stack componentStack = new Stack();
    private int buttonsDown = 0;
    private int modifiersDown = 0;
    private long lastEventTime = 0;
    private int clickCount = 0;
    private Component dragSource = null;
    private EventNormalizer normalizer = null;

    public InputState() {
        long mask = MouseEvent.MOUSE_MOTION_EVENT_MASK
            |MouseEvent.MOUSE_EVENT_MASK|KeyEvent.KEY_EVENT_MASK;
        AWTEventListener listener = new SingleThreadedEventListener() {
            public void processEvent(AWTEvent event) {
                update(event);
            }
        };
        normalizer = new EventNormalizer();
        normalizer.startListening(listener, mask);
    }

    public void dispose() {
        normalizer.stopListening();
        normalizer = null;
    }

    /** Explicitly update the internal state. */
    void update(AWTEvent event) {
        if (event instanceof MouseEvent) {
            updateState((MouseEvent)event);
        }
        else if (event instanceof KeyEvent) {
            updateState((KeyEvent)event);
        }
    }

    protected void updateState(KeyEvent ke) {
        // Ignore "old" events
        if (ke.getWhen() < lastEventTime)
            return;

        synchronized(this) {
            lastEventTime = ke.getWhen();
            modifiersDown = ke.getModifiers();
            // FIXME add state of individual keys
        }
    }

    protected void updateState(MouseEvent me){
        // Ignore "old" events
        if (me.getWhen() < lastEventTime) {
            Log.debug("Ignoring " + Robot.toString(me));
            return;
        }
        Log.debug("Updating based on " + Robot.toString(me));
        Point where = me.getPoint();
        // getComponentAt and getLocationOnScreen want the tree lock, so be
        // careful not to use any additional locks at the same time to avoid
        // deadlock.
        Component containingComponent = null;
        Point containingScreenLoc = null;
        Point eventScreenLoc = null;
        Point dragLoc = null;
        boolean screenLocationFound = true;
        // Determine the current mouse position in screen coordinates
        try {
            containingComponent = getComponentAt(me.getComponent(), where);
            containingScreenLoc = containingComponent != null
                ? containingComponent.getLocationOnScreen() : null;
            eventScreenLoc = me.getComponent().getLocationOnScreen();
            if (containingComponent == null) {
                if (me.getID() == MouseEvent.MOUSE_DRAGGED
                    && getMouseComponent() != null) {
                    containingComponent = getMouseComponent();
                    containingScreenLoc = 
                        containingComponent.getLocationOnScreen();
                }
                else {
                    // MOUSE_EXITED events report coords which aren't
                    // technically contained within the given component.
                    containingComponent = me.getComponent();
                    containingScreenLoc = eventScreenLoc;
                }
            }
            else if (containingComponent != me.getComponent()) {
                Point p1 = containingScreenLoc;
                Point p2 = eventScreenLoc;
                where.translate(p2.x - p1.x, p2.y - p1.y);
            }
        }
        catch(IllegalComponentStateException e) {
            // component might be hidden by the time we process this event
            screenLocationFound = false;
        }
        synchronized(this) {
            lastEventTime = me.getWhen();
            buttonsDown = me.getModifiers() & MOUSE_MASK;
            clickCount = me.getClickCount();
            if (me.getID() == MouseEvent.MOUSE_DRAGGED) {
                dragSource = me.getComponent();
                if (screenLocationFound
                    && containingComponent != dragSource) {
                    // where is relative to the drag source; make it relative
                    // to the real containing component.
                    where.translate(eventScreenLoc.x, eventScreenLoc.y);
                    where.translate(-containingScreenLoc.x,
                                    -containingScreenLoc.y);
                }
            }
            else if (me.getID() == MouseEvent.MOUSE_MOVED) {
                dragSource = null;
            }
            if (me.getID() == MouseEvent.MOUSE_ENTERED) {
                componentStack.push(containingComponent);
            }
            else if (me.getID() == MouseEvent.MOUSE_EXITED) {
                if (componentStack.empty())
                    Log.debug("Got " + Robot.toString(me)
                              + " but component not on stack");
                else
                    componentStack.pop();
            }
            else if (me.getID() == MouseEvent.MOUSE_RELEASED) {
                dragSource = null;
            }
            if (screenLocationFound) {
                if (componentStack.empty())
                    mouseLocation = null;
                else
                    mouseLocation = new Point(where);
                Log.debug("New location is " + mouseLocation);
                mouseLocationOnScreen.setLocation(containingScreenLoc);
                mouseLocationOnScreen.translate(where.x, where.y);
                Log.debug("New screen location is " + mouseLocationOnScreen);
            }
        }
    }

    /** Same as Component.getComponentAt, but recurses to the lowest-level
        component instead of only one level.  Point is in component
        coordinates.<br> 
        The default Component.getComponentAt can return invisible components
        (JRootPane has an invisible JPanel (glass pane?) which will otherwise
        swallow everything).
        NOTE: getComponentAt grabs the TreeLock, so this should *only* be
        invoked on the event dispatch thread, preferably with no other locks
        held. 
        NOTE: What about drags outside a component?
    */
    public static Component getComponentAt(Component parent, Point p) {
        Component comp = parent.getComponentAt(p);
        Component found = comp;
        // If the point is within a subcomponent but that component is not
        // showing, find one that is
        if (comp == null || comp == parent) {
            return found;
        }
        if (!comp.isShowing()) {
            Component[] subs = ((Container) parent).getComponents();
            for (int i = 0; i < subs.length; i++) {
                if (subs[i] == comp)
                    continue;
                Point p1 = new Point(p.x - subs[i].getX(), p.y - subs[i].getY());
                Component sub = getComponentAt(subs[i], p1);
                if (sub != null) {
                    //Log.debug("Found " + sub);
                    found = sub;
                    break;
                }
                else {
                    //Log.debug("Not in " + sub);
                }
            }
        }
        // If the point is in a subcomponent, check for subcomponents of that
        // one. 
        else {
            Point p1 = new Point(p.x - comp.getX(), p.y - comp.getY());
            found = getComponentAt(comp, p1);
        }

        return found;
    }

    public synchronized Component getMouseComponent() {
        Component comp = componentStack.empty()
            ? null : (Component)componentStack.peek();
        // Check to see if the component has gone away
        if (comp != null && !comp.isShowing()) {
            componentStack.pop();
            return getMouseComponent();
        }
        return comp;
    }

    public synchronized boolean isDragging() { return dragSource != null; }
    public synchronized Component getDragSource() { return dragSource; }
    public synchronized void setDragSource(Component c) {
        dragSource = c;
    }
    public synchronized int getClickCount() { return clickCount; }
    protected synchronized void setClickCount(int count) {
        clickCount = count;
    }
    public synchronized long getLastEventTime() { return lastEventTime; }
    protected synchronized void setLastEventTime(long t) { lastEventTime = t; }
    public synchronized int getModifiers() { return modifiersDown; }
    protected synchronized void setModifiers(int m) { modifiersDown = m; }
    public synchronized int getButtons() { return buttonsDown; }
    protected synchronized void setButtons(int b) { buttonsDown = b; }
    /** Returns the mouse location relative to the component that currently
        contains the pointer, or null if outside all components.
    */
    public synchronized Point getMouseLocation() {
        return mouseLocation != null ? new Point(mouseLocation) : null;
    }
    /** Returns the last known mouse location. */
    public synchronized Point getMouseLocationOnScreen() {
        return new Point(mouseLocationOnScreen);
    }
}

