package abbot.tester.awt;

import java.awt.AWTEvent;
import java.awt.Component;
import java.awt.EventQueue;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.AWTEventListener;
import java.awt.event.WindowEvent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.WeakHashMap;

/** Keep track of all known windows. */
public class WindowTracker {
    
    /** Keep track of all app contexts we can. */
    private WeakHashMap contexts;
    private ContextTracker contextTracker;
    // Only ever need one of these
    private static WindowTracker tracker = null;

    public static synchronized WindowTracker getInstance() {
        if (tracker == null) 
            tracker = new WindowTracker();
        return tracker;
    }

    private WindowTracker() {
        contextTracker = new ContextTracker();
        long mask = WindowEvent.WINDOW_EVENT_MASK;
        Toolkit toolkit = Toolkit.getDefaultToolkit();
        toolkit.addAWTEventListener(contextTracker, mask);
        contexts = new WeakHashMap();
        contexts.put(toolkit.getSystemEventQueue(), new WeakHashMap());
    }

    /** Has the window's WINDOW_OPENED event been sent yet?
        A window's "showing" flag may be set true before the WINDOW_OPENED
        event is generated. */
    public boolean isWindowShowing(Window w) {
        Iterator iter = contexts.keySet().iterator();
        if (w.isShowing()) {
            while (iter.hasNext()) {
                WeakHashMap whm = (WeakHashMap)contexts.get(iter.next());
                if (whm.containsKey(w))
                    return true;
            }
        }
        return false;
    }

    /** Return all available root Windows, unfiltered.  A root Window is one
     * that has a null parent.  Nominally this means a list similar to that
     * returned by Frame.getFrames(), but in the case of an Applet may return
     * a few Dialogs as well.
     */
    public List getRootWindows() {
        ArrayList list = new ArrayList();
        synchronized(contexts) {
            Iterator iter = contexts.keySet().iterator();
            while (iter.hasNext()) {
                EventQueue queue = (EventQueue)iter.next();
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                if (queue == toolkit.getSystemEventQueue()) {
                    Frame[] frames = Frame.getFrames();
                    //Log.debug(String.valueOf(frames.length) + " Frames");
                    for (int i=0;i < frames.length;i++) {
                        list.add(frames[i]);
                    }
                }
                else {
                    WeakHashMap map = (WeakHashMap)contexts.get(queue);
                    list.addAll(map.values());
                }
            }
        }
        //Log.debug(String.valueOf(list.size()) + " total Frames");
        return list;
    }

    /** Whenever we get a window that's on a new event dispatch thread, take
     * note of the thread, since it may correspond to a new event queue and
     * AppContext. 
     */
    // FIXME what if it has the same app context? can we check?
    private class ContextTracker implements AWTEventListener {
        public void eventDispatched(AWTEvent ev) {
            WindowEvent we = (WindowEvent)ev;
            Component comp = we.getComponent();
            if (comp.getParent() == null) {
                EventQueue queue = comp.getToolkit().getSystemEventQueue();
                if (ev.getID() == WindowEvent.WINDOW_OPENED) {
                    synchronized(contexts) {
                        WeakHashMap whm = (WeakHashMap)contexts.get(queue);
                        if (whm == null) {
                            contexts.put(queue, whm = new WeakHashMap());
                        }
                        whm.put(comp, comp);
                    }
                }
                else if (ev.getID() == WindowEvent.WINDOW_CLOSED) {
                    synchronized(contexts) {
                        WeakHashMap whm = (WeakHashMap)contexts.get(queue);
                        whm.remove(comp);
                    }
                }
            }
        }
    }
}
