package abbot.util;

import java.awt.EventQueue;

import javax.swing.SwingUtilities;

import abbot.Log;

/** Handler for uncaught exceptions on any event dispatch thread.
    Once this has been installed, the class must be accessible by any
    subsequently launched dispatch thread.<p>

    This handler is installed by setting the System property
    sun.awt.exception.handler.  See javadoc for java.awt.EventDispatchThread
    for details.  This is sort of a patch to Sun's implementation, which only
    checks the property once and caches the result ever after.  This
    implementation will always chain to the handler indicated by the current
    value of the property.
 */
public abstract class AbstractEventExceptionHandler {
    /** See javadoc for java.awt.EventDispatchThread. */
    public static final String PROP_NAME = "sun.awt.exception.handler";

    private static boolean installed = false;
    /** If there was an event exception handler set when we install one, keep
        a reference to it for possible later use.
    */
    protected static String oldHandler = null;

    /** Install a handler for event dispatch exceptions.  This is kind
        of a hack, but it's Sun's hack.
        See the javadoc for java.awt.EventDispatchThread for details.
        NOTE: we throw an exception immediately, which 
        ensures that our handler is installed, since otherwise 
        someone might set this property later.
        java.awt.EventDispatchThread doesn't actually load the handler
        specified by the property until an exception is caught by the
        event dispatch thread.  SwingSet2 in 1.4.1 installs its own.
        Note that a new instance is created for each exception thrown.

        @throws RuntimeException if the handler cannot be installed.
        @throws IllegalStateException if this method is invoked from an event
        dispatch thread. 
        @throws IllegalArgumentException if the given class is not derived
        from this one. 
    */
    public static void install(final Class handlerClass) {
        if (SwingUtilities.isEventDispatchThread()) {
            throw new IllegalStateException("Handler must not be installed from the event dispatch thread");
        }
        if (!AbstractEventExceptionHandler.class.
            isAssignableFrom(handlerClass)) {
            throw new IllegalArgumentException("Handler must be derived from AbstractEventExceptionHandler");
        }

        if (installed) {
            // If we've already installed an instance of
            // AbstractEventExceptionHandler, all we need to do is set the
            // property name.
            System.setProperty(PROP_NAME, handlerClass.getName());
        }
        else {
            boolean firstInstallation = !installed;
            
            // Even if it's been set to something else, we can override it if 
            // there hasn't been an event exception thrown yet.
            EventQueue.invokeLater(new Runnable() {
                public void run() {
                    oldHandler = System.getProperty(PROP_NAME);
                    // Set the property just before throwing the exception;
                    // OSX sets the property as part of the VM startup, so
                    // we have to override it here.
                    System.setProperty(PROP_NAME, handlerClass.getName());
                    throw new DummyException();
                }
            });
            // Does nothing but wait for the previous invocation to finish
            AWT.invokeAndWait(new Runnable() { public void run() { } });
        
            if (installed) {
                if (oldHandler != null) {
                    Log.warn("Replaced an existing event exception handler ("
                             + oldHandler + ")");
                }
            }
            else {
                throw new RuntimeException("A different handler for event "
                                           + "dispatch thread exceptions ("
                                           + oldHandler + ") has already been "
                                           + "installed, and there is no way "
                                           + " to override it");
            }
        }
    }

    /** Define this to handle the exception as needed. */
    protected abstract void exceptionCaught(Throwable thrown);

    /** Handle exceptions thrown on the event dispatch thread. */
    public void handle(Throwable thr) {
        String handler = System.getProperty(PROP_NAME);
        if (!handler.equals(getClass().getName())) {
            Log.debug("Another exception handler ("
                      + handler + ") has been installed, invoking it");
            try {
                ClassLoader cl =
                    Thread.currentThread().getContextClassLoader();
                Class c = Class.forName(handler, true, cl);
                c.getMethod("handle", new Class[] { Throwable.class }).
                    invoke(c.newInstance(), new Object[] { thr });
            }
            catch(Throwable e) {
                Log.warn(e);
            }
        }
        // The exception may be created by a different class loader
        // so compare by name only
        if (thr.getClass().getName().equals(DummyException.class.getName())) {
            // Install succeeded
            installed = true;
        }
        else {
            Log.debug("Handling exception on event dispatch thread: " + thr);
            Log.debug(thr);
            exceptionCaught(thr);
        }
    }

    private static class DummyException extends RuntimeException {
    }
}

