package abbot.util;

import abbot.Log;

/** Provides support for loading a class <i>before</i> checking the parent
 * class loader for it.  If delegation is off, defers to parent only if the
 * class is not found in this loader's path. 
 */
public class PreloadingClassLoader extends PathClassLoader {

    public PreloadingClassLoader(String path, ClassLoader parent) {
        super(path, parent);
    }

    /** Returns whether the given class should be given to the parent class
     * loader to try before this one does.  The default implementation always
     * returns false.
     */ 
    protected boolean shouldDelegate(String name) {
        return false;
    }

    /** Find the given class in the search path. */
    public Class findClass(String name) throws ClassNotFoundException {
        Log.debug("Looking up " + name);
        return super.findClass(name);
    }

    /** Load the given class, if possible, but don't delegate to any other
        class loader.
    */
    public synchronized Class loadClass(String name, boolean resolve)
        throws ClassNotFoundException {
        if (shouldDelegate(name)) {
            Log.debug("Delegating lookup for " + name);
            return super.loadClass(name, resolve);
        }
        else {
            Log.debug("Preloading lookup for " + name);
            Class c = findLoadedClass(name);
            if (c == null) {
                try {
                    c = findClass(name);
                }
                catch(SecurityException se) {
                    Log.debug(se);
                    // We're not allowed to find it, so give it to the parent
                    return super.loadClass(name, resolve);
                }
                catch(ClassNotFoundException cnf) {
                    // If we can't find it, maybe the parent can
                    return super.loadClass(name, resolve);
                }
                if (resolve) {
                    resolveClass(c);
                }
            }
            else {
                Log.debug("Class already loaded " + name);
            }
            return c;
        }
    }
}
