
package com.ibm.abbot.editor.recorder;

import java.util.*;
import java.lang.reflect.*;

import org.eclipse.swt.widgets.*;
import org.eclipse.swt.*;

import com.ibm.abbot.*;
import com.ibm.abbot.script.*;
import com.ibm.abbot.tester.swt.*;

import abbot.Log;
import abbot.BugReport;
import abbot.editor.recorder.RecordingFailedException;

/**
 * Provides recording of both low-level events and semantic events.
 */
public class EventRecorder {
	Display display;
	ArrayList steps;
	boolean captureMotion, verbose = false;
	Resolver resolver;
	SemanticRecorder semanticRecorder = null;
	HashMap semanticRecorders = new HashMap();
	private long lastStepTime = 0;
	private long lastEventTime = 0;
	private WidgetFinder finder;

	// all types of events an SWT widget can respond to
	int[] eventTypes ={	SWT.None,		SWT.KeyDown, 		SWT.KeyUp,
						SWT.MouseDown,	SWT.MouseUp,		SWT.MouseMove,
						SWT.MouseEnter,	SWT.MouseExit,		SWT.MouseDoubleClick,
						SWT.Paint,		SWT.Move,			SWT.Resize,
						SWT.Dispose,	SWT.Selection,		SWT.DefaultSelection,
						SWT.FocusIn,	SWT.FocusOut,		SWT.Expand,
						SWT.Collapse,	SWT.Iconify,		SWT.Deiconify,
						SWT.Close,		SWT.Show,			SWT.Hide,
						SWT.Modify,		SWT.Verify,			SWT.Activate,
						SWT.Deactivate,	SWT.Help,			SWT.DragDetect,
						SWT.Arm,		SWT.Traverse,		SWT.MouseHover,
						SWT.HardKeyDown,SWT.HardKeyUp,		SWT.MenuDetect };

	// the actual event listener that acts as the source for the recording
	Listener recorder = new Listener(){
		public void handleEvent(org.eclipse.swt.widgets.Event e){
			report("Recording event: "+e+"character="+e.character+" stateMask="+e.stateMask+" keyCode="+e.keyCode);
			recordEvent(e);
		}
	};

	/**
	 * Return a sequence containing all the semantic and basic events captured
	 * thus far.
	 */
	protected Step createStep() {
		return new Sequence(getResolver(), null, steps);
	}
	
	/** 
	 * Return a step or sequence of steps representing the steps created thus
	 * far, or null if none.
	 */
	  public Step getStep() {
		  return createStep();
	  }
	
	/** 
	 * Create an event recorder for the given resolver, optionally
	 * ignoring mouse motion.
	 */
	public EventRecorder(Resolver resolver, Display display, boolean captureMotion){
		this.display = display;
		steps = new ArrayList();
		this.captureMotion = captureMotion;
		this.resolver = resolver;
		this.finder = DefaultWidgetFinder.getFinder();
		lastStepTime=  System.currentTimeMillis();
	}
	
	/**
	 * Start recording.
	 */
	public void start(){
		steps.clear();
		initDisplay();
		lastEventTime = System.currentTimeMillis();
		Log.log("EventRecorder started");
	}
	
	/**
	 * Terminate recording.
	 */
	public void terminate()throws RecordingFailedException {
		clearDisplay();
		Log.log("EventRecorder terminated");
	}
	
	/**
	 * Add filters to the Display to listen to all SWT Event types.
	 */
	protected void initDisplay(){
		if(display!=null){
			for(int i=0; i<eventTypes.length;i++){
				final int index = i;
				Robot.syncExec(display,null,new Runnable(){
					public void run(){
						display.addFilter(eventTypes[index],recorder);
					}
				});
			}		
			// trying to 'jar' the recorder into starting		
//			display.syncExec(new Runnable(){
//				public void run(){
//					Shell active = display.getActiveShell();
//					if(active!=null){
//						WidgetTester.getTester(Widget.class).activate(active);	
//					}
//				}
//			});
		}
	}
	
	/**
	 * Remove filters from the Display.
	 */
	protected void clearDisplay(){
		if(display!=null){
			for(int i=0; i<eventTypes.length;i++){
				final int index = i;
				Robot.syncExec(display,null,new Runnable(){
					public void run(){
						display.removeFilter(eventTypes[index],recorder);
					}
				});
			}
		}
	}
	
	/** Handle an event.  This can either be ignored or contribute to the
	 * recording.
	 * For a given event, if no current semantic recorder is active,
	 * select one based on the event's component.  If the semantic recorder
	 * accepts the event, then it is used to consume each subsequent event,
	 * until its recordEvent method returns true, indicating that the semantic
	 * event has completed.
	 */ 
	protected void eventDispatched(Event event)
		throws RecordingFailedException {
	
		// Discard any key/button release events at the start of the recording.
		if (steps.size() == 0
			&& event.type==SWT.KeyUp) {
			Log.log("Ignoring initial release event");
			return;
		}
	
		SemanticRecorder newRecorder = null;
		
		if (semanticRecorder == null) {
			SemanticRecorder sr = getSemanticRecorder(event.widget.getClass());
			if (sr.accept(event)) {
				semanticRecorder = newRecorder = sr;
				//setStatus("Recording semantic event with " + sr);
			}
		}
	
		// If we're currently recording a semantic event, continue to do so
		if (semanticRecorder != null) {
			boolean consumed = semanticRecorder.record(event);
			boolean finished = semanticRecorder.isFinished();
			if (finished) {
				Log.debug("Semantic recorder is finished");
				saveSemanticEvent();
			}
			// If not consumed, need to check for semantic recorder (again)
			// (but avoid recursing indefinitely)
			if (!consumed && newRecorder == null) {
				Log.debug("Event was not consumed, parse it again");
				eventDispatched(event);
			}
		}
		else {
			captureRawEvent(event);
		}
	}
	
	/** Capture the given event as a raw event. */
	private void captureRawEvent(Event event) {
		Log.debug("Raw event recorded: "+event);
		// TODO process raw event for recording
	}
	
	/** Insert an arbitrary script step into the currently recorded stream. */
	public void insertStep(Step step) {
		steps.add(step);
		if ((step instanceof Assert)
			&& ((Assert)step).getMethodName().equals("assertFrameShowing")) {
			long timeout = ((Assert)step).getTimeout();
			long delta = System.currentTimeMillis() - lastStepTime;
			if (delta > timeout)
				timeout += delta;
			((Assert)step).setTimeout(timeout);
		}
		lastStepTime = getLastEventTime();
	}
	
	public long getLastEventTime() {
		return lastEventTime;
	}
	
	/** Process the given event. 
		@throws RecordingFailedException if an error was encountered and
		recording should discontinue.
	*/
	public void recordEvent(Event event)
		throws RecordingFailedException {
		Log.debug("Recorder: " + event);
		
		// TODO Fix the deadlock situation that occurs when recordEvent is called
		// inside the recording event-handler.  
		if (finder.isFiltered(event.widget))
			return;
		lastEventTime = System.currentTimeMillis();
		eventDispatched(event);
	}	
		
	/** Return whether an event was generated.  Assumes a SemanticRecorder is
		active.
		@throws RecordingFailedException if an error was encountered.
	*/
	private boolean saveSemanticEvent() throws RecordingFailedException {
		Log.log("Storing event from current semantic recorder");
		try {
			Step step = semanticRecorder.getStep();
			if (step != null) {
				insertStep(step);
				//setStatus("Added " + step);
			}
			else {
				//setStatus("No semantic event found, events skipped");
			}
			semanticRecorder = null;
			return step != null;
		}
		catch(BugReport bug) {
			throw new RecordingFailedException(bug);
		}
		catch(Exception e) {
			throw new RecordingFailedException(new BugReport("Unexpected exception while recording", e));
		}
	}
	
	/** Return the semantic recorder for the given component class. */
	private SemanticRecorder getSemanticRecorder(Class cls) {
		if (!(Widget.class.isAssignableFrom(cls))) {
			throw new IllegalArgumentException("Class must derive from "
											   + "Widget ("+cls+")");
		}
		SemanticRecorder sr = (SemanticRecorder)semanticRecorders.get(cls);
		if (sr == null) {
			Class ccls = Robot.getCanonicalClass(cls);
			if (ccls != cls) {
				sr = getSemanticRecorder(ccls);
				// Additionally cache the mapping from the non-canonical class
				semanticRecorders.put(cls, sr);
				return sr;
			}
			String cname = Robot.simpleClassName(cls);
			try {
				cname = "com.ibm.abbot.editor.recorder." + cname + "Recorder";
				Class recorderClass = Class.forName(cname);
				Constructor ctor = recorderClass.getConstructor(new Class[] {
					Resolver.class, 
				});
				sr = (SemanticRecorder)ctor.newInstance(new Object[] { 
					getResolver()
				});
				//sr.addActionListener(getListener());
			}
			catch(InvocationTargetException e) {
				Log.warn(e);
			}
			catch(NoSuchMethodException e) {
				sr = getSemanticRecorder(cls.getSuperclass());
			}
			catch(InstantiationException e) {
				sr = getSemanticRecorder(cls.getSuperclass());
			}
			catch(IllegalAccessException iae) {
				sr = getSemanticRecorder(cls.getSuperclass());
			}
			catch(ClassNotFoundException cnf) {
				sr = getSemanticRecorder(cls.getSuperclass());
			}
			// Cache the results for future reference
			semanticRecorders.put(cls, sr);
		}
		return sr;
	}
	
	/** Return the Resolver to be used by this recorder. */
	protected Resolver getResolver() { return resolver; }

	public void setVerbose(boolean verbose){
		this.verbose = verbose;
	}
	
	public void report(String msg){
		if(verbose)
			System.out.println(msg);
	}
	
	
	protected void processEvent(org.eclipse.swt.widgets.Event e){
		
	}
}
