
package com.ibm.abbot.tester.swt;

import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.*;

/**
 * Provides widget-specific actions, assertions, and getter methods for
 * widgets of type List.
 */
public class ListTester extends ScrollableTester {
	/* Widget properties that are obtainable by member getter functions */
	protected int focusIndex, itemCount, itemHeight, selectionCount, selectionIndex, 
		topIndex;
	protected String item;
	protected String[] items, selection;

	/**
	 * Sets the above properties to their current values for the given widget. 
	 * NOTE: This should be called in a block of code synchronized on this
	 * tester.
	 */
	protected synchronized void getProperties(final List list){
		super.getProperties(list);
		Robot.syncExec(list.getDisplay(),this,new Runnable(){
			public void run(){
				focusIndex = list.getFocusIndex();
				itemCount = list.getItemCount();
				itemHeight = list.getItemHeight();
				selectionCount = list.getSelectionCount();
				selectionIndex = list.getSelectionIndex();
				topIndex = list.getTopIndex();
				items = list.getItems();
				selection = list.getSelection();
			}
		});			
	}

	/**
	 * These getter methods return a particular property of the given widget.
	 * @see the corresponding member function in class Widget   
	 */ 
	/* Begin getters */	
	public synchronized String getItem(final List list, final int index){
		Robot.syncExec(list.getDisplay(),this,new Runnable(){
			public void run(){
				item = list.getItem(index);
			}
		});
		return item;
	}

	public synchronized int getFocusIndex(List list) {
		getProperties(list);
		return focusIndex;
	}

	public synchronized int getItemCount(List list) {
		getProperties(list);
		return itemCount;
	}

	public synchronized int getItemHeight(List list) {
		getProperties(list);
		return itemHeight;
	}

	public synchronized String[] getItems(List list) {
		getProperties(list);
		return items;
	}

	public synchronized String[] getSelection(List list) {
		getProperties(list);
		return selection;
	}

	public synchronized int getSelectionCount(List list) {
		getProperties(list);
		return selectionCount;
	}

	public synchronized int getSelectionIndex(List list) {
		getProperties(list);
		return selectionIndex;
	}

	public synchronized int getTopIndex(List list) {
		getProperties(list);
		return topIndex;
	}
	/* End getters */

	/** Move the mouse pointer over the item at the given index **/
	protected void mouseMoveIndex(final List list, final int index){
		actionFocus(list);
		Robot.syncExec(list.getDisplay(),null,new Runnable(){
			public void run(){
				list.setTopIndex(index);
			}
		});
		int topIndex = getTopIndex(list);
		int borderWidth = list.getBorderWidth();
		int itemHeight = list.getItemHeight();
		Rectangle bounds = getGlobalBounds(list);
		Rectangle clientArea = list.getClientArea();
		mouseMove(bounds.x+clientArea.width/2,bounds.y+borderWidth+itemHeight*(index-topIndex)+itemHeight/2);		
		actionWaitForIdle(list.getDisplay());
	}
	
	/** Click the first occurance given item **/
	public void actionClickItem(List list, String item){
		actionClickItem(list,item,SWT.BUTTON1);
	}

	/** Click the item at the given index **/
	public void actionClickIndex(List list, int index){
		actionClickIndex(list,index,SWT.BUTTON1);
	}
	
	/** Click the first occurance of given item based on the given accelerator **/
	public void actionClickItem(List list, String item, int accelerator){
		String[] items = getItems(list);
		int index = -1;
		for(int i=0; i<items.length;i++){
			if(item.equals(items[i])&&index==-1)
				index = i;
		}
		if(index==-1){
			throw new ActionFailedException("List item \""+item+"\" not found");
		}
		actionClickIndex(list,index,accelerator);	
	}
	
	/** Click the item at the given index based on the given accelerator **/
	public void actionClickIndex(List list, int index, int accelerator){
		actionFocus(list);
		mouseMoveIndex(list,index);
		actionWaitForIdle(list.getDisplay());
		mousePress(accelerator);
		mouseRelease(accelerator);
		actionWaitForIdle(list.getDisplay());		
	}	
	
	/** 
	 * Select the first occurance of an item from the given list if it wasn't already 
	 * selected, and unselect it otherwise. 
	 */
	public void actionSelectItem(List list, String item){
		String[] items = getItems(list);
		int index = -1;		
		for(int i=0; i<items.length;i++){
			if(item.equals(items[i])&&index==-1)
				index = i;			
		}
		if(index==-1){
			throw new ActionFailedException("List item \""+item+"\" not found");
		}
		actionSelectIndex(list,index);	
	}
	
	/** 
	 * Selects the item at the given index, or unselects it if it was 
	 * already selected.
	 */
	public void actionSelectIndex(final List list, final int index){
		actionFocus(list);
		mouseMoveIndex(list,index);
		Robot.syncExec(list.getDisplay(),null,new Runnable(){
			public void run(){
				int[] selectionIndices = list.getSelectionIndices();
				boolean selected = false;
				int indexIntoSelected = -1;
				for(int i=0; i<selectionIndices.length;i++){
					if(selectionIndices[i]==index){
						selected = true;
						indexIntoSelected = i;
					}
				}
				if(selected){
					int[] setSelected = new int[selectionIndices.length-1];
					int x = 0;
					for(int i=0; i<setSelected.length;i++){
						if(selectionIndices[i]!=index){
							setSelected[x] = selectionIndices[i];
							x++;
						}						
					}
					list.setSelection(setSelected);
				}
				else{//!selected
					int[] setSelected = new int[selectionIndices.length+1];
					for(int i=0; i<selectionIndices.length;i++){
						setSelected[i] = selectionIndices[i];			
					}
					setSelected[setSelected.length-1]=index;
					list.setSelection(setSelected);
				}			
			}
		});
		actionWaitForIdle(list.getDisplay());
	}
}
