package com.ibm.swt;

import java.awt.AWTException;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;

import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.*;

import abbot.tester.awt.KeyStrokeMap;
import java.util.Arrays;

//---- for testing ----
import org.eclipse.swt.layout.*;
/**
 * Provides the functionality of <code>java.awt.Robot</code> for use 
 * in SWT environments.  Thus this class should be regarded as "equivalent to"
 * <code>java.awt.Robot</code>, not <code>abbot.tester.awt.Robot</code>.
 * 
 * @author Kevin Dale
 * @version $Id: Robot.java,v 1.2 2003/08/20 22:29:22 ktdale Exp $
 * 	
 **/
public class Robot /*extends java.awt.Robot*/{
	
	/** the # of modifiers- like SWT.CTRL, SWT.ALT, etc- that are currently defined  */
	private static final int MODIFIER_COUNT= 7;
	
	/** mask used to identify if any mouse buttons were clicked */
	private static final int BUTTON_MASK = (SWT.BUTTON1|SWT.BUTTON2|SWT.BUTTON3);
	
	/** the display associated with this robot */
	private Display displayProperty;
	
	/** the java.awt.Robot that does all the leg-work for us */
	private java.awt.Robot robot;
	
	/**
	 * Constructs a <code>Robot</code> object in the coordinate system of the primary screen.
	 * 
	 * @throws org.eclipse.swt.SWTException
	 */
	public Robot()throws SWTException{
		//displayProperty = Display.getDefault();
		try{
			robot = new java.awt.Robot();
		}
		catch(AWTException awte){
			throw new SWTException("(Translated AWTException) "+awte.getMessage());
		}
	}
	
	/**
	 * CURRENTLY THIS IS FUNCTIONALLY THE SAME AS THE NO-PARAM CONSTRUCTOR
	 * 
	 * Creates a <code>Robot</code> for the given <code>Display</code>.
	 * 
	 * @param display the <code>Display</code> associated with this robot
	 * 
	 * @throws org.eclipse.swt.SWTException
	 */
	public Robot(Display display) throws SWTException {
		this.displayProperty = display;
		try{
			robot = new java.awt.Robot();
		}
		catch(AWTException awte){
			throw new SWTException("(Translated AWTException) "+awte.getMessage());
		}
	}
		
	/**
	 * Creates an image containing pixels read from the screen.  
	 * <p>
	 * NOTE: Application code must explicitly invoke the <code>Image.dispose()</code> 
 	 * method to release the operating system resources managed by each instance
 	 * when those instances are no longer required.
 	 * </p>
	 * 
	 * @param rect <code>Rectangle</code> to capture in screen coordinates
	 * 
	 * @return the captured <code>Image</code>
	 */		
	public synchronized Image createScreenCapture(Rectangle rect){
		PaletteData pData = new PaletteData(255<<16,	// red mask
											255<<8,		// green mask
											255);		// blue mask
											
		ImageData iData = new ImageData(rect.width,rect.height,24,pData);
		
		Color color;
		for(int x=0; x<rect.width; x++){
			for(int y=0; y<rect.height; y++){
				color = getPixelColor(rect.x+x,rect.y+y);
				iData.setPixel(x,y,color.getRed()<<16 |
								   color.getGreen()<<8 |
								   color.getBlue());
			}	
		}
		
		Image image = new Image(Display.getDefault(),iData);
		return image;
	}
	
	/**
	 * Returns the color of a pixel at the given screen coordinates.
	 * <p>
	 * NOTE: Application code must explicitly invoke the <code>Color.dispose()</code> 
 	 * method to release the operating system resources managed by each instance
 	 * when those instances are no longer required.
 	 * </p>
 	 * 
 	 * @param x X-position of pixel
 	 * @param y Y-position of pixel
 	 * 
 	 * @return color of the specified pixel
	 */
	public synchronized Color getPixelColor(int x, int y){
		java.awt.Color awtColor = robot.getPixelColor(x,y);
		return new Color(Display.getDefault(), awtColor.getRed(),
								  awtColor.getGreen(),
								  awtColor.getBlue());
	}
	
	/**
	 * Sleep for the specified amount of time.
	 * 
	 * @param delay the number of milliseconds to sleep
	 */
	public synchronized void delay(int ms){
		robot.delay(ms);
	}
	
	/**
	 * Returns the number of milliseconds this <code>Robot</code> sleeps after 
	 * generating an event.
	 * 
	 */
	public synchronized int getAutoDelay(){
		return robot.getAutoDelay();
	}
	
	/**
	 * Returns whether this <code>Robot</code> automatically invokes 
	 * <code>waitForIdle</code> after generating an event.
	 * 
	 * @return whether <code>waitForIdle</code> is automatically called
	 */
	public synchronized boolean isAutoWaitForIdle(){
		return robot.isAutoWaitForIdle();
	}
	
	/**
	 * Rotates the scroll wheel on wheel-equipped mice. 
	 * 
	 * @param wheelAmt number of "notches" to move the mouse wheel 
	 * Negative values indicate movement up/away from the user, positive values 
	 * indicate movement down/towards the user.
	 * 
	 */
	/* NOT IN JDK 1.3.1
	public synchronized void mouseWheel(int wheelAmt){
		robot.mouseWheel(wheelAmt);
	}
	*/
	
	/**
	 * Sets the number of milliseconds this <code>Robot</code> sleeps after 
	 * generating an event.
	 * 
	 * @param ms time to sleep in milliseconds 
	 */
	public synchronized void setAutoDelay(int ms){
		robot.setAutoDelay(ms);
	}
	
	/**
	 * Sets whether this <code>Robot</code> automatically invokes <code>waitForIdle</code> 
     * after generating an event.
     * 
     * @param isOn Whether <code>waitForIdle</code> is automatically invoked
	 */	
	public synchronized void setAutoWaitForIdle(boolean isOn){
		robot.setAutoWaitForIdle(isOn); 
	}
	
	/**
	 * Returns a string representation of this <code>Robot</code>.
	 * 
	 * @returns The <code>String</code> representation
	 */	
	public synchronized String toString(){
		String params = "autoDelay = "+getAutoDelay()+", "+"autoWaitForIdle = "+isAutoWaitForIdle();
		return getClass().getName() + "[ " + params + " ]";
	}
	
	/**
	 * Dispatches events from the OS's event queue until all events
	 * in the queue, prior to calling this method, have been handled. 
	 */
	public synchronized void waitForIdle() {
		if(displayProperty.getThread()==Thread.currentThread()){
			while(displayProperty.readAndDispatch()) ;		
		}
		else{
			displayProperty.syncExec(new Runnable() {
				public void run() {
					while(displayProperty.readAndDispatch());
				}
			});
		}
	}
	
	public /*synchronized*/ void waitForIdle(final Display display){
		display.syncExec(new Runnable() {
			public void run() {
				while(display.readAndDispatch());
			}
		});		
	}
	
	/**
	 * Presses all keys in a given accelerator.  To type a character or digit,
	 * just use the unicode value.  For example, <code> accelerator = 'k', 
	 * accelerator = 'K', accelerator = '5' </code>.  Note that this is case-
	 * sensitive, so the <code>SWT.SHIFT</code> key is implied if the character 
	 * is uppercase or otherwise requires the <code>SWT.SHIFT</code> key.  	 
	 * 
	 * For non-character keys, use the keycodes defined in org.eclipse.swt.SWT.
	 * For example, to type F1, <code> accelerator = SWT.F1; </code>
	 * 
	 * Note that an accelerator can contain multiple modifier-key masks- like 
	 * <code>accelerator = SWT.CTRL | SWT.ALT | SWT.SHIFT;</code>- but at most 
	 * one character or keycode.  Also, mouse-button masks are ignored here.
	 * 
	 * @param accelerator SWT accelerator containing the keys to be pressed
	 * 
	 * @see org.eclipse.swt.SWT for more info on accelerators
	 */
	public synchronized void keyPress(int accelerator){
		int[] keys = SWTCodeToAWT(accelerator);
		boolean shift = false;
		for(int i=0; i<keys.length;i++){
			//System.out.print(keys[i]+",");
			if(keys[i]!=0){// make sure that this entry is not empty
				if(keys[i]==KeyEvent.VK_SHIFT && shift) ;// skip it; shift has already been pressed
				else try{ 
					if(keys[i]==KeyEvent.VK_PERIOD)
						robot.keyPress(KeyEvent.VK_DECIMAL);
					else if(keys[i]==KeyEvent.VK_MINUS)
						robot.keyPress(KeyEvent.VK_SUBTRACT);	
					else
						robot.keyPress(keys[i]);
				}
				catch(IllegalArgumentException iae){
					System.err.println("IllegalArgumentException: keystroke :"+keys[i]
										+"\nAccelerator:"+accelerator
										+"\nCast as a char: "+(char)accelerator);
										
				}				
				if(keys[i]==KeyEvent.VK_SHIFT){
					shift = true;
				}
			}
		}
		//System.out.println();
	}

	/**
	 * Releases all keys in a given accelerator. 
	 * 
	 * @param accelerator SWT accelerator containing the keys to be released
	 * 
	 * @see org.eclipse.swt.SWT for more info on accelerators
	 * @see keyPress(int accelerator)
	 */
	public synchronized void keyRelease(int accelerator){
		int[] keys = SWTCodeToAWT(accelerator);
		boolean shift = false;
		for(int i=0; i<keys.length;i++){
			if(keys[i]!=0){// make sure that this entry is not empty
				if(keys[i]==KeyEvent.VK_SHIFT && shift) ;// skip it; shift has already been pressed
				else if(keys[i]==KeyEvent.VK_PERIOD)
					robot.keyRelease(KeyEvent.VK_DECIMAL);
				else if(keys[i]==KeyEvent.VK_MINUS)
					robot.keyRelease(KeyEvent.VK_SUBTRACT);	
				else
					robot.keyRelease(keys[i]);
				
				if(keys[i]==KeyEvent.VK_SHIFT){
					shift = true;
				}
			}
		}
	}
	
	/**
	 * Moves mouse pointer to given screen coordinates.
	 * 
	 * @param x X-position
	 * @param y Y-position
	 */
	public synchronized void mouseMove(int x, int y){
		robot.mouseMove(x,y);
	}
	
	/**
	 * Presses all mouse buttons contained in a given accelerator, which 
	 * can include any/all of the following: <code> SWT.BUTTON1, SWT.BUTTON2, SWT.BUTTON3</code>.
	 * 
	 * Note that this method will not generate any keystrokes, only mouse
	 * button presses.
	 * 
	 * @param accelerator SWT accelerator containing the mouse buttons to be pressed
	 * 
	 * @see org.eclipse.swt.SWT for more info on accelerators
	 * @see keyPress(int accelerator)	 
	 */
	public synchronized void mousePress(int accelerator){
		accelerator &= BUTTON_MASK;
		if((accelerator&SWT.BUTTON1)==SWT.BUTTON1)
			robot.mousePress(InputEvent.BUTTON1_MASK);
		if((accelerator&SWT.BUTTON2)==SWT.BUTTON2)
			robot.mousePress(InputEvent.BUTTON2_MASK);
		if((accelerator&SWT.BUTTON3)==SWT.BUTTON3)
			robot.mousePress(InputEvent.BUTTON3_MASK);
	}
	
	/**
	 * Releases all mouse buttons contained in a given accelerator.
	 * <p>
	 * Note that this method will not release any keys, only mouse buttons.
	 * </p> 
	 * @param accelerator SWT accelerator containing the mouse buttons to be released
	 * 
	 * @see org.eclipse.swt.SWT for more infor on accelerators
	 * @see mousePress(int accelerator)
	 * @see keyPress(int accelerator)
	 */
	public synchronized void mouseRelease(int accelerator){
		accelerator &= BUTTON_MASK;
		if((accelerator&SWT.BUTTON1)==SWT.BUTTON1)
			robot.mouseRelease(InputEvent.BUTTON1_MASK);
		if((accelerator&SWT.BUTTON2)==SWT.BUTTON2)
			robot.mouseRelease(InputEvent.BUTTON2_MASK);
		if((accelerator&SWT.BUTTON3)==SWT.BUTTON3)
			robot.mouseRelease(InputEvent.BUTTON3_MASK);
	}
	
	/* Converts a single accelerator code into multiple virtual keycodes.  
	 * This ignores mouse buttons.
	 */
	private int[] SWTCodeToAWT(int code){
		//System.out.println(Integer.toBinaryString(code));
		int[] res = new int[MODIFIER_COUNT+2];// one extra for the keycode/character(last pos) and VK_SHIFT may occupy more than one pos
		Arrays.fill(res,0);
		int idx = 0;
		if((SWT.MODIFIER_MASK & code)!=0){//check for all modifier keys
			if((code&SWT.ALT)==SWT.ALT)		res[idx++]=KeyEvent.VK_ALT;
			if((code&SWT.SHIFT)==SWT.SHIFT)	res[idx++]=KeyEvent.VK_SHIFT;
			if((code&SWT.CTRL)==SWT.CTRL)	res[idx++]=KeyEvent.VK_CONTROL;
			//if((code&SWT.COMMAND)==SWT.COMMAND)res[idx++]=KeyEvent.VK_UNDEFINED;		
			
			// mouse buttons
			//if((code&SWT.BUTTON1)==SWT.BUTTON1)	res[idx++]=InputEvent.BUTTON1_MASK;
			//if((code&SWT.BUTTON2)==SWT.BUTTON2)	res[idx++]=InputEvent.BUTTON2_MASK;
			//if((code&SWT.BUTTON3)==SWT.BUTTON3)	res[idx++]=InputEvent.BUTTON3_MASK;
		
		}
		int keyCode = code & SWT.KEY_MASK;
		if((SWT.KEYCODE_BIT & keyCode)!=0){// code contains a keycode- set the last array element
			switch(keyCode){
				case SWT.ARROW_UP:			res[MODIFIER_COUNT+1]= KeyEvent.VK_UP; break;
				case SWT.ARROW_DOWN:		res[MODIFIER_COUNT+1]= KeyEvent.VK_DOWN; break;
				case SWT.ARROW_LEFT:		res[MODIFIER_COUNT+1]= KeyEvent.VK_LEFT; break;
				case SWT.ARROW_RIGHT:		res[MODIFIER_COUNT+1]= KeyEvent.VK_RIGHT; break;
				case SWT.PAGE_UP:			res[MODIFIER_COUNT+1]= KeyEvent.VK_PAGE_UP; break;
				case SWT.PAGE_DOWN:			res[MODIFIER_COUNT+1]= KeyEvent.VK_PAGE_DOWN; break;
				case SWT.HOME:				res[MODIFIER_COUNT+1]= KeyEvent.VK_HOME; break;
				case SWT.END:				res[MODIFIER_COUNT+1]= KeyEvent.VK_END; break;
				case SWT.INSERT:			res[MODIFIER_COUNT+1]= KeyEvent.VK_INSERT; break;

				case SWT.F1:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F1; break;
				case SWT.F2:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F2; break;
				case SWT.F3:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F3; break;
				case SWT.F4:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F4; break;
				case SWT.F5:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F5; break;
				case SWT.F6:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F6; break;
				case SWT.F7:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F7; break;
				case SWT.F8:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F8; break;
				case SWT.F9:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F9; break;
				case SWT.F10:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F10; break;
				case SWT.F11:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F11; break;
				case SWT.F12:				res[MODIFIER_COUNT+1]= KeyEvent.VK_F12; break;
				default:					res[MODIFIER_COUNT+1]= KeyEvent.VK_UNDEFINED; break;
			}			
		}
		else{// code contains a unicode character or digit (or ESC or DEL)						
			if(keyCode==SWT.DEL)
				res[MODIFIER_COUNT+1] = KeyEvent.VK_DELETE;
			else if(keyCode==SWT.ESC)
				res[MODIFIER_COUNT+1] = KeyEvent.VK_ESCAPE;
			else{
				if(Character.isUpperCase((char)keyCode))
					res[MODIFIER_COUNT] = KeyEvent.VK_SHIFT;
				javax.swing.KeyStroke ks = KeyStrokeMap.getKeyStroke((char)keyCode);
				if(ks==null){
					//System.err.println("No key mapping for '"+ (char)keyCode+"'");	
					res[MODIFIER_COUNT+1] = KeyEvent.VK_UNDEFINED;
				}
				else{
					res[MODIFIER_COUNT+1] = ks.getKeyCode();
				}
			}
		}
		return res;
	}	
	
	/**
	 * Get the <code>Display</code> object with which this robot is
	 * synchronized.
	 * 
	 * @return the <code>Display</code> associated with this <code>Robot</code>
	 */
	public Display getDisplay(){
		return displayProperty;
	}

	/**
	 * Set the <code>Display</code> object with which this robot is
	 * synchronized.
	 * 
	 * @param display the <code>Display</code> to associate with this <code>Robot</code>
	 */	
	public void setDisplay(Display display){
		this.displayProperty = display;
	}	
	
	/* test area */
	public static void main(String[] argv){
		try{
			Robot robot = new Robot();
			//robot.delay(3000);
			
			Display display = new Display();
			Shell shell = new Shell(display, SWT.SHELL_TRIM);
			shell.setLayout(new RowLayout(SWT.VERTICAL));
			
			
			//Canvas canvas = new Canvas(shell,SWT.NONE);
			Image image = robot.createScreenCapture(new Rectangle(0,0,200,200));
			Image image2 = new Image(display,"C:/WINNT/Web/wuicon2.gif");
			
			
			shell.open();
			System.out.println("SHELL OPENED");
			
			GC gc = new GC(shell);
			//gc.drawImage(image2,200,200);
			gc.drawImage(image,0,0);
			
			
			
			
			while(!shell.isDisposed()){
				if(!display.readAndDispatch())
					display.sleep();
			}
			
			
			/*
			
			//robot.setAutoDelay(300);
			int accelerator;
			
			// test uppercase letters
			for(char c = 'A'; c<='Z';c++){
				
				robot.keyPress((int)c);
				robot.keyRelease((int)c);	
				robot.delay(50);				
			}
			
			// test lowercase letters
			for(char c = 'a';c<='z';c++){
				robot.keyPress((int)c);
				robot.keyRelease((int)c);	
				robot.delay(50);		
			}
			
			// test other chars and shift-covered chars
			char[] specialChars = {' ','`','1','2','3','4','5','6','7','8','9','0','-','=','[',']','\\',';','\'',',','.','/',
									   '~','!','@','#','$','%','^','&','*','(',')','_','+','{','}','|',':', '"', '<','>','?'};
			for(int i=0; i<specialChars.length;i++){
				accelerator = (int)specialChars[i];
				robot.keyPress(accelerator);
				robot.keyRelease(accelerator);
				robot.delay(50);	
			}
						
			// test special chars that are also defined in SWT
			int [] keycodes = {SWT.TAB, SWT. BS, SWT.CR, SWT.LF};
			for(int i=0; i<1;i++){
				robot.keyPress(keycodes[i]);
				robot.keyRelease(keycodes[i]);
				robot.delay(50);	
			}
			
			
			// test keycodes defined in SWT
			int [] keycodes2 = {SWT.ARROW_UP,SWT.ARROW_DOWN,SWT.ARROW_LEFT,SWT.ARROW_RIGHT,SWT.PAGE_UP,SWT.PAGE_DOWN,SWT.HOME,SWT.END,SWT.INSERT};
			
			for(int i=0; i<keycodes2.length-1;i++){
				robot.keyPress(keycodes2[i]|SWT.KEYCODE_BIT);
				robot.keyRelease(keycodes2[i]|SWT.KEYCODE_BIT);
				robot.delay(50);
			}
			
			//test SWT.INSERT
			robot.keyPress(keycodes2[keycodes2.length-1]|SWT.KEYCODE_BIT);
			robot.keyRelease(keycodes2[keycodes2.length-1]|SWT.KEYCODE_BIT);
			
			
			// test SWT.DEL and SWT.ESC
			//bot.keyPress(SWT.DEL);
			//bot.keyRelease(SWT.DEL);
			robot.keyPress(SWT.ESC);
			robot.keyRelease(SWT.ESC);
			
			// test mouse clicks
			robot.mousePress(SWT.BUTTON3);
			robot.mouseRelease(SWT.BUTTON3);
		*/
		}
		catch(Exception e){
			e.printStackTrace();
		}
		
	}
	

}
