package abbot;

import java.awt.Button;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog;
import java.awt.Frame;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import junit.extensions.abbot.*;
import junit.extensions.abbot.Timer;
import junit.framework.TestCase;
import abbot.script.ComponentReference;

/** 
 * Verify we can generate and look up components in a wide variety of GUI
 * configurations.  Note that this tests the ComponentReference ctor as well,
 * since the two classes work closely together.
 */
public class DefaultComponentFinderTest extends ComponentTestFixture {

    private DefaultComponentFinder finder =
        (DefaultComponentFinder)DefaultComponentFinder.getFinder();

    /**
     * Test to see if the label and index can identify a component
     */
    public void testSiblings() {
        JTextField tf1 = new JTextField();
        JTextField tf2 = new JTextField();
        tf1.setColumns(10);
        tf2.setColumns(10);
        JPanel pane = new JPanel();
        pane.add(tf1);
        pane.add(tf2);

        showFrame(pane);

        ComponentReference cr1 = resolver.addComponent(tf1);
        ComponentReference cr2 = resolver.addComponent(tf2);
        assertTrue("Should get two unique references", cr1 != cr2);

        int w1 = finder.getComponentMatchWeight(tf1, cr1);
        int w2 = finder.getComponentMatchWeight(tf2, cr1);
        
        assertTrue("Component 2 (" + w2 + ") matched at least as well as "
                   + "Component 1 (" + w1 + ")", w1 > w2);

        w1 = finder.getComponentMatchWeight(tf1, cr2);
        w2 = finder.getComponentMatchWeight(tf2, cr2);
        
        assertTrue("Component 1 (" + w1 + ") matched at least as well as "
                   + "Component 2 (" + w2 + ")", w2 > w1);

    }


    /**
     * Verify two identical sibling components can be distinguished.
     */
    public void testIdenticalSiblings() {
        JButton button1 = new JButton("Button");
        JButton button2 = new JButton("Button");
        JPanel pane = new JPanel();
        pane.add(button1);
        pane.add(button2);

        showFrame(pane);

        ComponentReference cr1 = resolver.addComponent(button1);
        ComponentReference cr2 = resolver.addComponent(button2);
        assertTrue("Should get two unique references", cr1 != cr2);

        int w1 = finder.getComponentMatchWeight(button1, cr1);
        int w2 = finder.getComponentMatchWeight(button2, cr1);
        
        assertTrue("Button 2 (" + w2 + ") matched at least as well as "
                   + "Button 1 (" + w1 + ")", w1 > w2);

        w1 = finder.getComponentMatchWeight(button1, cr2);
        w2 = finder.getComponentMatchWeight(button2, cr2);

        assertTrue("Button 1 (" + w1 + ") matched at least as well as "
                   + "Button 2 (" + w2 + ")", w2 > w1);
    }

    
    /**
     * Verify two identical cousin (sibling parents) components can be
     * distinguished. 
     */
    public void testIdenticalCousins() {
        JButton button1 = new JButton("Button");
        JButton button2 = new JButton("Button");
        JPanel  panel1 = new JPanel();
        JPanel  panel2 = new JPanel();
        panel1.add(button1);
        panel2.add(button2);

        JPanel pane = new JPanel();
        pane.add(panel1);
        pane.add(panel2);

        showFrame(pane);

        ComponentReference cr1 = resolver.addComponent(button1);
        ComponentReference cr2 = resolver.addComponent(button2);
        assertTrue("Should get two unique references", cr1 != cr2);

        int w1 = finder.getComponentMatchWeight(button1, cr1);
        int w2 = finder.getComponentMatchWeight(button2, cr1);

        assertTrue("Button 2 (" + w2 + ") matched at least as well as "
                   + "Button 1 (" + w1 +")", w1 > w2);

        w1 = finder.getComponentMatchWeight(button1, cr2);
        w2 = finder.getComponentMatchWeight(button2, cr2);

        assertTrue("Button 1 (" + w1 + ") matched at least as well as "
                   + "Button 2 (" + w2 +")", w2 > w1);

    }

    /**
     * Test to see if two buttons which are identical except for their name
     * and index can be found
     */
    public void testTwoNamedButtons() {
        JButton button1 = new JButton("Button");
        JButton button2 = new JButton("Button");
        JPanel  panel1 = new JPanel();
        JPanel  panel2 = new JPanel();
        
        button1.setName("Button1");
        button2.setName("Button2");

        panel1.add(button1);
        panel2.add(button2);
        JPanel pane = new JPanel();
        pane.add(panel1);
        pane.add(panel2);

        showFrame(pane);

        ComponentReference cr1 = resolver.addComponent(button1);
        ComponentReference cr2 = resolver.addComponent(button2);
        assertTrue("Should get two unique references", cr1 != cr2);

        int w1 = finder.getComponentMatchWeight(button1, cr1);
        int w2 = finder.getComponentMatchWeight(button2, cr1);

        assertTrue("Button 2 matches at least as well as Button 1",
                   w1 > w2);
        w1 = finder.getComponentMatchWeight(button1, cr2);
        w2 = finder.getComponentMatchWeight(button2, cr2);

        assertTrue("Button 1 matches at least as well as Button 2",
                   w2 > w1);
    }

    /**
     * Test to ensure that index can be used on otherwise indistingushable 
     * components
     */
    public void testIdenticalLabelGroup() {
        
        final int NUM_LABELS = 10;

        ArrayList labels = new ArrayList();
        ArrayList refs = new ArrayList();

        JPanel pane = new JPanel();
        showFrame(pane);
        for (int i=0; i < NUM_LABELS; ++i){
            JLabel label = new JLabel("Empty");
            labels.add(label);
            pane.add(label);
            refs.add(resolver.addComponent(label));
        }
        
        for (int i=0;i < NUM_LABELS - 1; ++i) {
            for (int next=i+1;next < NUM_LABELS;next++) {
                Component c1 = (Component)labels.get(i);
                Component c2 = (Component)labels.get(next);
                ComponentReference cr1 = (ComponentReference)refs.get(i);
                ComponentReference cr2 = (ComponentReference)refs.get(next);

                int w1 = finder.getComponentMatchWeight(c1, cr1);
                int w2 = finder.getComponentMatchWeight(c2, cr1);
                assertTrue("label " + next + " (" + w2 + ") matches at least "
                           + "as well as label " + i + " (" + w1 + ")",
                           w1 > w2);

                w1 = finder.getComponentMatchWeight(c1, cr2);
                w2 = finder.getComponentMatchWeight(c2, cr2);
                assertTrue("label " + i + " (" + w1 + ") matches at least "
                           + "as well as label " + next + " (" + w2 + ")",
                           w2 > w1);
            }
        }
    }

    
    /**
     * Test to see if a JOptionPane and be located
     */
    public void testCanFindJOptionPane() {

        JOptionPane pane = new JOptionPane("A message",
                                           JOptionPane.INFORMATION_MESSAGE);
        // start the dialog on another thread so that it's visible when we try
        // and do a findComponent on its JOptionPane
        JFrame frame = new JFrame(getName());
        final JDialog dialog =
            pane.createDialog(frame, "JOptionPane Find Test");
        SwingUtilities.invokeLater(new Runnable() {
             public void run() { dialog.show(); }
        });
        
        ComponentReference cr1 = resolver.addComponent(pane);
        // Expect no exceptions
        assertNotNull("No component found", finder.findComponent(cr1));
    }

    /**
     * Test to ensure that JTextFields can be found.
     */
    public void testCanFindJTextField() {
        JTextField textField  = new JTextField();
        Frame frame = showFrame(textField);
        ComponentReference cr1 = resolver.addComponent(textField);
        
        assertNotNull("Text field not found", finder.findComponent(cr1));
    }

    /** Ensure we can find sub-dialogs. */
    public void testFindDialogs() {
        final String title = "Modal Dialog";
        final String title2 = "Non-modal Sub-dialog";
        JFrame frame = new JFrame(getName());
        JDialog d1 = new JDialog(frame, title, true);
        d1.getContentPane().add(new JLabel("Modal dialog"));
        JDialog d2 = new JDialog(d1, title2, false);
        d2.getContentPane().add(new JLabel("Non-modal subdialog"));

        showWindow(frame);
        showWindow(d1);
        showWindow(d2);
        assertEquals("First dialog should have been found",
                     d1, finder.findWindow(title));
        assertEquals("Second dialog should have been found",
                     d2, finder.findWindow(title2));
    }

    /** 
     * Ensure we can find dialogs/windows with a null parent. 
     */
    public void testNullParentDialog() {
        final JDialog d1 = new JDialog((Frame)null, "dialog 1");
        d1.getContentPane().add(new JLabel(getName()));
        SwingUtilities.invokeLater(new Runnable() {
            public void run() { d1.pack();d1.show(); }
        });
        d1.hide();
        d1.dispose();
        finder.disposeWindows();

        final JDialog d2 = new JDialog((Frame)null, "dialog 2");
        SwingUtilities.invokeLater(new Runnable() {
            public void run() { d2.pack();d2.show(); }
        });
        Timer timer = new Timer();
        while(!d2.isShowing()) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("Timed out waiting for dialog "
                                           + "to show");
        }
        assertEquals("Should have been able to find dialog",
                     d2, finder.findWindowByTitle(d2.getTitle()));
        d2.hide();
        d2.dispose();
    }

    /**
     * Test that unamed components have the correct default value for 
     * their name
     */
    public void testGetComponentName() {
        Component comp = new Button("Button");
        assertNull("Unnamed AWT Component's distilled name should be null",
                   finder.getComponentName(comp));
        comp.setName("my AWT button");
        assertEquals("Named AWT Component's name is used",
                     comp.getName(), finder.getComponentName(comp));

        comp = new JFrame("JFrame");
        assertNull("Unnamed JFrame's distilled name should be null",
                   finder.getComponentName(comp));
     }



    /** Ensure that find window works with JOptionPane-generated
     * components. */
    public void testDuplicateJOptionPanes() {
        JOptionPane pane = new JOptionPane("A message", JOptionPane.
                                           INFORMATION_MESSAGE);
        JFrame frame = new JFrame(getName());
        Dialog dialog = pane.createDialog(frame, "Dialog");
        showWindow(dialog);
        dialog.hide();
        Timer timer = new Timer();
        while (dialog.isShowing()) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("Timed out waiting for dialog to hide");
        }
        JOptionPane pane2 = new JOptionPane("A message", JOptionPane.
                                            INFORMATION_MESSAGE);
        Dialog d2 = pane2.createDialog(frame, "Dialog 2");
        showWindow(d2);
        Component comp = finder.findWindow(d2.getTitle());
        assertEquals("Wrong JOptionPane found", d2, comp);
    }

    private JButton findButton(Component c) {
        if (c instanceof JButton)
            return (JButton)c;
        if (c instanceof Container) {
            Component[] children = ((Container)c).getComponents();
            for (int i=0;i < children.length;i++) {
                JButton b = findButton(children[i]);
                if (b != null)
                    return b;
            }
        }
        return null;
    }

    /** Component reference matching one dialog should match an identical
     * subsequent one.
     */
    public void testSubsequentDialogContentsLookup() throws Throwable {
        ComponentReference ref =
            new ComponentReference("OK Button", javax.swing.JButton.class,
                                   null, "OK");
        Dialog d1 = null;
        Dialog d2 = null;
        JFrame frame = new JFrame(getName());
        JOptionPane pane = new JOptionPane("Dialog", JOptionPane.
                                           INFORMATION_MESSAGE);
        JOptionPane pane2 = new JOptionPane("Dialog", JOptionPane.
                                            INFORMATION_MESSAGE);
        d1 = pane.createDialog(frame, "Dialog");
        showWindow(d1);
        JButton button = findButton(d1);
        assertNotNull("Button shouldn't be null", button);
        assertEquals("Wrong component found",
                     button, finder.findComponent(ref));
        d1.hide();

        d2 = pane2.createDialog(frame, "Dialog");
        showWindow(d2);
        button = findButton(d2);
        assertNotNull("Button 2 shouldn't be null", button);
        assertEquals("Wrong component found",
                     button, finder.findComponent(ref));
    }

    public void testFindDialogWithHiddenFrame() throws Throwable {
        JFrame frame = new JFrame(getName());
        final Dialog d1 = new JDialog(frame, getName() + " Dialog");
        SwingUtilities.invokeLater(new Runnable() {
            public void run() { d1.pack(); d1.show(); }
        });
        Timer timer = new Timer();
        while (!d1.isShowing()) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("Timed out waiting for dialog to show");
        }
        ComponentReference ref = new ComponentReference(resolver, d1);
        assertEquals("Dialog not found", d1, finder.findComponent(ref));
    }

    public DefaultComponentFinderTest(String name) {
        super(name);
    }

    private Resolver resolver;
    protected void setUp() {
        resolver = new abbot.script.Script();
    }

    public static void main(String[] args) {
        TestHelper.runTests(args, DefaultComponentFinderTest.class);
    }
}
