package abbot.editor.recorder;

import java.awt.*;
import java.awt.event.*;

import abbot.*;
import abbot.script.*;
import abbot.tester.awt.*;
import abbot.editor.recorder.*;

/**
 * All SemanticRecorder tests should derive from this class.
 * Provides a framework similar to EventRecorder which creates appropriate
 * semantic recorders based on incoming events.<p>
 * Will continue recording past the first semantic recorder used, preserving
 * the first one until an invocation of assert[No]Step, which resets the
 * preserved recorder, replacing it with the current one when appropriate.
 * This allows a more or less continuous stream of actions alternating with
 * assert[No]Step calls.<p>
 * NOTE: it would probably be simpler to not do this and require explicit
 * startRecording steps, although there is some benefit to checking recorders
 * back to back (it more closely simulates the EventRecorder behavior).
 */
// FIXME still needs some refactoring
// Needs clearer design on when to stop recording and examination of trailing
// events that are consumed/not consumed
public abstract class AbstractSemanticRecorderTest
    extends AbstractRecorderTest {

    protected volatile SemanticRecorder currentRecorder = null;
    // There may be some cases where the initial events are recorded as
    // desired but subsequent events occur which might interfere with the
    // original recorder.  In this case, keep tracking the additional events
    // but preserve the first recorder.
    protected volatile SemanticRecorder savedRecorder = null;

    /** Create a new test case with the given name. */
    public AbstractSemanticRecorderTest(String name) {
        super(name);
    }

    /** Each Recorder subclass test should implement this method to return the
        SemanticRecorder to be tested. */
    protected abstract SemanticRecorder createRecorder(Resolver r);

    protected void startRecording() {
        // Don't pick up any lingering events
        getRobot().waitForIdle();
        super.startRecording();
    }

    protected Recorder getRecorder() {
        savedRecorder = currentRecorder = null;
        return new Recorder(getResolver()) {
            protected void eventDispatched(AWTEvent event) {
                AbstractSemanticRecorderTest.this.eventDispatched(event);
            }
            protected Step createStep() {
                return savedRecorder.getStep();
            }
            public long getEventMask() {
                return EventRecorder.RECORDING_EVENT_MASK;
            }
            public void terminate() { }
        };
    }

    // This dispatching is the canonical usage for a semantic recorder; cf
    // EventRecorder's lookup and dispatch
    // FIXME maybe make a semantic controller object that both this test and
    // event recorder use?
    private synchronized void eventDispatched(AWTEvent event) {
        Log.log("SR event: " + getRobot().toString(event));

        if (currentRecorder == null) {
            // ordinarily, you'd get an appropriate recorder based on the
            // event type
            SemanticRecorder cr = createRecorder(getResolver());
            if (cr.accept(event)) {
                currentRecorder = cr;
                Log.log("New recorder, event accepted");
            }
        }

        // Don't bother passing on events if the recorder has captured its
        // target event
        if (currentRecorder != null && !currentRecorder.isFinished()) {
            boolean consumed;
            boolean finished;
            Log.debug("Recording with " + currentRecorder);
            synchronized(currentRecorder) {
                consumed = currentRecorder.record(event);
                if (currentRecorder.isFinished()) {
                    Log.log("Recorder is finished");
                    // FIXME is this the best way to handle multiple recorders
                    if (savedRecorder == null
                        || savedRecorder.getStep() == null) {
                        Log.log("Preserving recorder");
                        savedRecorder = currentRecorder;
                    }
                    else {
                        Log.debug("Already have a saved recorder, ignored");
                    }
                    currentRecorder = null;
                }
            }
            if (!consumed) {
                Log.log("Re-sending event");
                eventDispatched(event);
            }
        }
    }

    protected void assertNoStep() {
        getRobot().waitForIdle();
        String bugInfo = bugInfo(watcher.listEvents());
        watcher.clear();
        if (savedRecorder != null) {
            assertEquals("Recorder should not have recorded a step " + bugInfo,
                         null, savedRecorder.getStep());
            savedRecorder = null;
        }
    }

    /** Make sure the step created matches the given pattern.  The recorder
     * may not nececessarily have detected the end of the pattern.
     */
    protected void assertStep(String pattern) {
        assertStep(pattern, false);
    }

    /** Make sure the recorder created step matches the given pattern,
     * optionally requiring the recorder to have detected the end of the
     * semantic event.
     */
    // FIXME this depends on the description that is generated; it should
    // check the actual step structure instead
    protected void assertStep(String pattern, boolean mustBeFinished) {
        // Make sure all events have been processed
        getRobot().waitForIdle();
        String bugInfo = bugInfo(watcher.listEvents());
        Log.log("Verifying recorder state (" + pattern + ")");
        try {
            synchronized(this) {
                if (savedRecorder == null || savedRecorder.getStep() == null) {
                    savedRecorder = currentRecorder;
                    currentRecorder = null;
                }
            }
            assertTrue("No recorder instantiated " + bugInfo,
                       savedRecorder != null);
            synchronized(savedRecorder) {
                if (mustBeFinished) {
                    assertTrue("No current recorder " + bugInfo,
                               savedRecorder != null);
                    assertTrue("Semantic recorder should have detected the "
                               + "end of this event sequence " + bugInfo, 
                               savedRecorder.isFinished());
                }
                super.assertStep(pattern, savedRecorder.getStep());
                savedRecorder = null;
            }
        }
        catch(RuntimeException thr) {
            Log.log("** FAILED ***");
            throw thr;
        }
    }

    protected boolean hasRecorder() {
        return currentRecorder != null || savedRecorder != null;
    }

    protected Step getStep() {
        return savedRecorder != null ? savedRecorder.getStep() : null;
    }
}
