package abbot.editor.recorder;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.table.JTableHeader;

import junit.extensions.abbot.*;
import junit.extensions.abbot.Timer;
import abbot.*;
import abbot.script.*;
import abbot.tester.awt.*;


/**
 * Unit test to verify proper capture of basic user semantic events.  A
 * ComponentTester is used to generate the appropriate events, which is
 * significantly easier than reproducing the steps manually.  These tests
 * should be run after the ComponentTester tests.
 */
public class ComponentRecorderTest 
    extends AbstractSemanticRecorderTest {

    private ComponentTester tester =
        ComponentTester.getTester(Component.class);

    /** Create a new test case with the given name. */
    public ComponentRecorderTest(String name) {
        super(name);
    }

    protected SemanticRecorder createRecorder(Resolver r) {
        return new ComponentRecorder(r);
    }

    // Open events to capture (those known to generate unique sequences of
    // events):
    // First window (open)
    // Subsequent window (open')
    // Window (close)

    public void testNamedWindowShow() {
        Frame frame = showFrame(new JLabel(getName()));
        JWindow win = new JWindow(frame);
        win.getContentPane().add(new JLabel(getName()));
        String NAME = "MyWindow";
        win.setName(NAME);
        startRecording();
        showWindow(win);
        assertStep("Wait for FrameShowing\\(" + NAME + "\\)");
        assertNoStep();

        win.dispose();
        assertStep("Wait for !FrameShowing\\(" + NAME + "\\)");
    }

    public void testFrameShowDispose() {
        startRecording();
        Frame frame = showFrame(new JLabel(getName()));
        // FIXME 1.4 produces additional RESIZE and SHOWING events prior to
        // WINDOW_OPEN -- need to catch these

        // Must catch an initial window open.  First open produces
        // WINDOW_OPENED.   
        assertStep("Wait for FrameShowing\\(" + getName() + "\\)");

        // Dispose produces WINDOW_CLOSED, without a COMPONENT_HIDDEN
        // 1.4.1 adds a HIERARCHY_CHANGED !SHOWING prior to CLOSED
        frame.dispose(); 
        assertStep("Wait for !FrameShowing\\(" + getName() + "\\)");

        // Shouldn't record multiple WINDOW_CLOSED events (pre-1.4)
        frame.dispose();
        assertTrue("Multiple CLOSE events generated", !hasRecorder());
    }

    public void testFrameShowHideShow() {
        Frame frame = showFrame(new JLabel(getName()));
        hideWindow(frame);
        startRecording();
        showWindow(frame);
        assertStep("Wait for FrameShowing\\(" + getName());

        hideWindow(frame);
        assertStep("Wait for !FrameShowing\\(" + getName());

        showWindow(frame);
        assertStep("Wait for FrameShowing\\(" + getName());
    }

    public void testDialogShowDispose() {
        // Showing a dialog produces a WINDOW_OPEN, but no SHOWING
        Frame frame = showFrame(new JLabel(getName()));
        startRecording();
        final JDialog dialog = new JDialog(frame, "Dialog");
        showWindow(dialog);
        assertStep("Wait for FrameShowing");
        
        dialog.dispose();
        assertStep("Wait for !FrameShowing");
        
        // Shouldn't record multiple WINDOW_CLOSED events (pre-1.4)
        dialog.dispose();
        assertTrue("Multiple CLOSE events generated", !hasRecorder());
    }

    public void testDialogShowHideShow() {
        // Showing a dialog produces a WINDOW_OPEN, but no SHOWING
        Frame frame = showFrame(new JLabel(getName()));
        startRecording();
        final JDialog dialog = new JDialog(frame, "Dialog");
        showWindow(dialog);
        assertStep("Wait for FrameShowing");
        
        hideWindow(dialog);
        assertStep("Wait for !FrameShowing");
        
        showWindow(dialog);
        assertStep("Wait for FrameShowing");
    }

    private boolean itemOne = false;
    private boolean itemTwo = false;
    private boolean subItemOne = false;
    private boolean subItemTwo = false;
    private JPanel initPopup() {
        final JPopupMenu popup = new JPopupMenu();
        popup.add(new JMenuItem(new AbstractAction("Item One") {
            public void actionPerformed(ActionEvent ev) {
                itemOne = true;
            }
        }));
        popup.add(new JMenuItem(new AbstractAction("Item Two") {
            public void actionPerformed(ActionEvent ev) {
                itemTwo = true;
            }
        }));
        JMenu submenu = new JMenu("Submenu");
        submenu.add(new JMenuItem(new AbstractAction("Subitem One") {
            public void actionPerformed(ActionEvent ev) {
                subItemOne = true;
            }
        }));
        submenu.add(new JMenuItem(new AbstractAction("Subitem Two") {
            public void actionPerformed(ActionEvent ev) {
                subItemTwo = true;
            }
        }));
        popup.add(submenu);
        JPanel pane = new JPanel() {
            public Dimension getPreferredSize() {
                return new Dimension(100, 100);
            }
        };
        pane.addMouseListener(new MouseAdapter() {
            private void showPopup(MouseEvent me) {
                popup.pack();
                popup.show((Component)me.getSource(), 
                           me.getX(), me.getY());
            }
            public void mousePressed(MouseEvent me) {
                if (me.isPopupTrigger()) {
                    showPopup(me);
                }
            }
            public void mouseReleased(MouseEvent me) {
                if (me.isPopupTrigger()) {
                    showPopup(me);
                }
            }
        });
        return pane;
    }

    public void testRecordPopup() {
        JPanel pane = initPopup();
        Frame frame = showFrame(pane);

        startRecording();

        tester.actionSelectPopupMenuItem(pane, "Item One");
        assertStep("SelectPopupMenuItem");

        tester.actionSelectPopupMenuItem(pane, "Subitem Two");
        assertStep("SelectPopupMenuItem");
    }

    /** If no popup exists, a regular click should be recorded instead, using
     * the appropriate mask.
     */
    public void testRecordNoPopup() {
        JLabel label = new JLabel("No Popup Here");
        showFrame(label);
        startRecording();
        tester.actionClick(label,
                           label.getWidth()/2,
                           label.getHeight()/2,
                           tester.POPUP_MODIFIER);
        assertStep("Click\\(.*,POPUP_MASK\\)");
    }

    public void testRecordAWTMenu() {
        MenuBar mb = new MenuBar();
        Menu menu1 = new Menu("File"); mb.add(menu1);
        Menu menu2 = new Menu("Edit"); mb.add(menu2);
        MenuItem mi = new MenuItem("Copy");
        menu2.add(mi);
        JFrame frame = (JFrame)showFrame(new JLabel(getName()));
        frame.setMenuBar(mb);

        startRecording();
        tester.actionSelectAWTMenuItemByLabel(frame, mi.getLabel());
        assertStep("AWTMenuItem");
    }

    public void testRecordMenu() {
        JMenuBar menubar = new JMenuBar();
        JMenu menu1 = new JMenu("File"); menubar.add(menu1);
        JMenu menu2 = new JMenu("Edit"); menubar.add(menu2);
        JMenuItem mi = new JMenuItem("Copy");
        menu2.add(mi);
        JFrame frame = new JFrame(getName());
        frame.setJMenuBar(menubar);
        showWindow(frame);

        startRecording();
        tester.actionSelectMenuItem(mi);
        assertStep("MenuItem");
    }

    public void testRecordClick() {
        Frame frame = showFrame(new JLabel(getName()));
        startRecording();
        tester.actionClick(frame);
        assertStep("Click");
    }

    // Fails on OSX 1.4.1
    public void xtestRecordClickWithModifiers() {
        Frame frame = showFrame(new JLabel(getName()));
        startRecording();
        String mods = "BUTTON1_MASK|CTRL_MASK";
        tester.actionClick(frame,
                           frame.getWidth()/2,
                           frame.getHeight()/2, mods);
        assertStep("Click\\(\\)");
    }

    /** If the step we're recording results in the target component being
     * hidden (or in general if the component gets hidden before the step is
     * finished recording), the step should still be recorded properly.
     */
    // NOTE: this doesn't test the general case, which is still not
    // supported. 
    public void testRecordWhenComponentIsHidden() {
        JButton close = new JButton("close");

        JFrame frame = new JFrame(getName());
        final JDialog dialog =
            new JDialog(frame, getName() + " Dialog");
        dialog.getContentPane().add(close);
        close.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ev) {
                dialog.dispose();
            }
        });
        JButton push = new JButton("push");
        push.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ev) {
                dialog.pack(); dialog.show();
            }
        });

        showFrame(push);
        tester.actionClick(push);
        startRecording();
        tester.actionClick(close);
        assertStep("Click");
    }

    private void showTextField() {
        JTextField tf = new JTextField();
        tf.setColumns(10);
        showFrame(tf);
        // make sure text field has focus
        tester.actionFocus(tf);
    }

    public void testRecordKeyStroke() {
        showTextField();
        startRecording();
        tester.actionKeyString("a");
        assertStep("KeyString\\(a\\)");
    }

    public void testRecordShiftedKeyStroke() {
        showTextField();
        startRecording();
        tester.actionKeyString("A");
        assertStep("KeyString\\(A\\)");
    }

    public void testRecordSpecialKeys() {
        showTextField();
        startRecording();
        String[] keys = {
            "VK_ENTER", "VK_TAB", "VK_BACK_SPACE", "VK_DELETE", "VK_ESCAPE"
        };
        for (int i=0;i < keys.length;i++) {
            tester.actionKeyStroke(keys[i]);
            // NOTE: not every platform generates KEY_TYPED events; if not,
            // then there is no problem. (w2k 1.3.1 gets no DELETE)
            if (hasRecorder())
              assertStep("KeyStroke\\(" + keys[i] + "\\)");
        }
    }

    /** No event should be recorded. */
    public void testRecordAltKey() {
        showTextField();
        startRecording();
        tester.actionKeyPress("VK_ALT");
        tester.actionKeyPress("VK_A");
        tester.actionKeyRelease("VK_A");
        tester.actionKeyRelease("VK_ALT");
        assertNoStep();
    }

    /** No event should be recorded. */
    // FIXME OSX has fixed this
    public void xtestRecordAltGraphKey() {
        // OSX throws an exception, apparently they don't support the key
        // Linux doesn't register it as a modifier
        // Windows says it's an invalid argument
        if (!Platform.isOSX() && !Platform.isWindows()) {
            showTextField();
            startRecording();
            tester.actionKeyPress("VK_ALT_GRAPH");
            tester.actionKeyPress("VK_A");
            tester.actionKeyRelease("VK_A");
            tester.actionKeyRelease("VK_ALT_GRAPH");
            assertNoStep();
        }
    }

    /** No event should be recorded. */
    public void testRecordControlKey() {
        showTextField();
        startRecording();
        tester.actionKeyPress("VK_CONTROL");
        tester.actionKeyPress("VK_A");
        tester.actionKeyRelease("VK_A");
        tester.actionKeyRelease("VK_CONTROL");
        assertNoStep();
    }

    public void testRecordDragDrop() {
        // Use a table with draggable headers as a default drag & drop test
        // Drag the first column header over the second and verify they
        // switched positions
        Object[][] values = {
            { "green", "ugly" },
            { "red", "beautiful" },
        };
        Object[] names = { "Color", "Type" };
        JTable table = new JTable(values, names);
        Frame frame = showFrame(new JScrollPane(table));
        JTableHeader header = table.getTableHeader();
        startRecording();
        int width = header.getSize().width;
        tester.actionDrag(header, 1, 1);
        assertStep("Drag");
        tester.actionDrop(header, width * 3 / 4, 1);
        assertStep("Drop");
        assertEquals("Column header not moved", 
                     1, table.convertColumnIndexToView(0));
    }
    
    public void testRecordMacInputMethod() {
        if (Platform.isMacintosh()) {
            showTextField();
            startRecording();
            tester.actionKeyStroke("VK_E", "ALT_MASK");
            tester.actionKeyStroke("VK_E");
            assertStep("KeyString\\(\u00e9\\)");
        }
    }

    // FIXME only do this where InputMethod is supported -- how do we detect
    // that? 
    public void xtestRecordInputMethod() {
        showTextField();
        startRecording();
        try {
            tester.actionKeyStroke("VK_HALF_WIDTH");
        }
        catch(ArrayIndexOutOfBoundsException aio) {
            if (Platform.isMacintosh())
                return;
            throw aio;
        }
        catch(IllegalArgumentException iae) {
            if (Platform.isLinux()) 
                return;
            throw iae;
        }
        tester.actionKeyStroke("VK_K");
        tester.actionKeyStroke("VK_A");
        tester.actionKeyStroke("VK_N");
        tester.actionKeyStroke("VK_J");
        tester.actionKeyStroke("VK_I");
        tester.actionKeyStroke("VK_FULL_WIDTH");
        assertStep("Some text"); // FIXME
        Step step = getStep();
        assertTrue("Should have recorded a sequence",
                   step instanceof Sequence);
        assertEquals("Wrong number of keystrokes", 7, ((Sequence)step).size());
    }

    public void testRecordMultipleClick() {
        JLabel label = new JLabel(getName());
        showFrame(label);
        startRecording();
        tester.actionClick(label, label.getWidth()/2,
                           label.getHeight()/2,
                           "BUTTON1_MASK", 2);
        Step step = getStep();
        assertStep("Click\\(.*,2\\)", false);
    }

    private class ToolTipWatcher implements AWTEventListener {
        public volatile boolean opened = false;
        public void eventDispatched(AWTEvent ev) {
            if (ev.getID() == WindowEvent.WINDOW_OPENED) {
                opened = true;
                Toolkit.getDefaultToolkit().
                    removeAWTEventListener(ToolTipWatcher.this);
            }
        }
    }

    /** Tooltip displays should not be recorded. */
    public void testRecordToolTips() {
        JLabel label = new JLabel(getName());
        ToolTipWatcher tw = new ToolTipWatcher();
        label.setToolTipText("A really long label that is sure to create a window that exceeds the original frame bounds");
        showFrame(label);
        Toolkit.getDefaultToolkit().
            addAWTEventListener(tw, WindowEvent.WINDOW_EVENT_MASK);
        startRecording();
        tester.mouseMove(label, 0, 0);
        tester.mouseMove(label);
        Timer timer = new Timer();
        while (!tw.opened) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("No tooltip appeared");
        }
        assertNoStep();
    }

    public static void main(String[] args) {
        RepeatHelper.runTests(args, ComponentRecorderTest.class);
    }
}
