package abbot.editor.recorder;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import junit.extensions.abbot.RepeatHelper;
import abbot.Log;
import abbot.Platform;
import abbot.script.*;
import abbot.tester.awt.*;
import abbot.tester.awt.Robot;

/**
 * Unit test to verify proper capture of multiple user semantic events when
 * parsed from a continuous event stream.
 * FIXME add support to feed the event recorder an event sequence and have it
 * do its cleanup
 * FIXME add support to feed the event recorder an event sequence to record
 * Still want to do the robot generation, 'cuz that'll catch errors on new
 * platforms.  Feeding specific event sequences avoids having to throw up an
 * IF at all.
 */
public class EventRecorderTest extends AbstractRecorderTest {

    private EventRecorder recorder = null;
    private ComponentTester tester =
        ComponentTester.getTester(Component.class);
    private Step step = null;

    // Provide this so we can test as if we were a platform that doesn't
    // generate them
    private boolean suppressKeyTypedEvents = false;
        
    protected Recorder getRecorder() {
        step = null;
        // Add a little extra to optionally suppress KEY_TYPED events
        recorder = new EventRecorder(new Script(), false) {
            protected void eventDispatched(AWTEvent event) {
                if (!suppressKeyTypedEvents
                    || event.getID() != KeyEvent.KEY_TYPED) {
                    Log.debug("ERT: " + Robot.toString(event));
                    super.eventDispatched(event);
                }
            }
        };
        recorder.addActionListener(new ActionListener() {
            /** Will be invoked when the recorder posts status messages. */
            public void actionPerformed(ActionEvent ev) {
                Log.debug(ev.getActionCommand());
            }
        });
        return recorder;
    }

    private void waitForStep() {
        tester.waitForIdle();
        step = recorder.getStep();
    }

    private Step getStep() {
        if (step == null) 
            step = recorder.getStep();
        return step;
    }

    public void testClickWithDialog() {
        // clicking a button that results in a change of focus to that button
        // shouldn't leave a trailing MOUSE_RELEASED after the Click is parsed
        // are there other possible intervening events?
        // 
        // In addition, make sure the resulting dialog only results in a
        // single "wait for frame showing"
        JButton button = new JButton("Push me");
        final Frame frame = showFrame(button);

        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent ev) {
                JDialog dialog = new JDialog(frame, "Dialog", true);
                dialog.getContentPane().add(new JButton("OK"));
                dialog.pack();
                dialog.show();
            }
        });
        startRecording();
        tester.actionClick(button);
        stopRecording();
        
        assertTrue("Event recorder should return a Sequence", 
                   getStep() instanceof Sequence);
        Sequence seq = (Sequence)getStep();
        assertStepCount(2, seq);
        assertStep("Click", seq.getStep(0));
        assertStep("Wait for FrameShowing", seq.getStep(1));
    }

    /** Display a text field, already focused. */
    private JTextField showTextField() {
        JTextField tf = new JTextField();
        tf.setColumns(10);
        showFrame(tf);
        tester.actionFocus(tf);
        return tf;
    }

    /** Ensure that a single key press results in a single keystroke event. */
    // Sometimes fails on linux with pointer focus
    public void testRecordSingleKey() {
        // Seems we get dual key_typed events for non-text components; make
        // sure we save only one.  If a text field is used, only one key_typed
        // event is received.
        JButton button = new JButton("Focus here");
        JTextField tf = new JTextField("Focus here next");
        JPanel pane = new JPanel();
        pane.add(button);
        pane.add(tf);
        showFrame(pane);

        // First send to a button
        tester.actionFocus(button);
        startRecording();
        tester.actionKeyStroke("VK_A");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   getStep() instanceof Sequence);
        Sequence seq = (Sequence)getStep();
        assertStepCount(1, seq);
        assertStep("KeyString\\(a\\)", seq.getStep(0));

        // Now send to a text field
        tester.actionFocus(tf);
        startRecording();
        tester.actionKeyStroke("VK_B");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   getStep() instanceof Sequence);
        seq = (Sequence)getStep();
        assertStepCount(1, seq);
        assertStep("KeyString\\(b\\)", seq.getStep(0));
    }

    /** Should capture numeric keypad events if numlock is down.
        Test twice so that we can catch both states of the num lock key.
        Restore the num lock state when we're done.
     */
    public void testRecordNumPad() {
        JTextField tf = showTextField();
        // Don't want any special key mappings; this prevents *any* keys from
        // appearing, but still allows the recorder to run as expected.
        tf.setInputMap(JComponent.WHEN_FOCUSED, new InputMap());

        boolean numLockPressed = false;
        int NUMKEYS = 10;
        int PAGEKEYS = 9;
        try {
            for (int j=0;j < 2;j++) {
                if (j == 1) {
                    // Press the num_lock key to test the alternate version
                    tester.actionKeyStroke("VK_NUM_LOCK");
                    numLockPressed = true;
                }
                startRecording();
                for (int i=0;i < NUMKEYS;i++) {
                    tester.actionKeyStroke("VK_NUMPAD" + i);
                }
                stopRecording();
                waitForStep();
                assertTrue("Event recorder should return a Sequence",
                           getStep() instanceof Sequence);
                Sequence seq = (Sequence)getStep();
                assertTrue("No steps recorded", seq.size() > 0);
                if (seq.getStep(0).toString().
                    startsWith("KeyStroke(VK_INSERT")) {
                    assertStepCount(PAGEKEYS, seq);
                    String[] keys = {
                        "VK_INSERT",
                        "VK_END",
                        "VK_KP_DOWN",
                        "VK_PAGE_DOWN",
                        "VK_KP_LEFT",
                        "VK_KP_RIGHT",
                        "VK_HOME",
                        "VK_KP_UP",
                        "VK_PAGE_UP",
                    };
                    for (int i=0;i < PAGEKEYS;i++) {
                        assertStep("KeyStroke\\(" + keys[i] + "\\)",
                                   seq.getStep(i));
                    }
                }
                else {
                    assertStepCount(NUMKEYS, seq);
                    for (int i=0;i < NUMKEYS;i++) {
                        assertStep("KeyStroke\\(VK_NUMPAD" + i + "\\)",
                                   seq.getStep(i));
                    }
                }
            }
        }
        finally {
            if (numLockPressed) {
                tester.actionKeyStroke("VK_NUM_LOCK");
            }
        }
    }

    /** Should save no individual key press/release events. */
    public void testRecordKeyString() {
        showTextField();
        startRecording();
        tester.actionKeyStroke("VK_T");
        tester.actionKeyStroke("VK_E");
        tester.actionKeyStroke("VK_X");
        tester.actionKeyStroke("VK_T");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)getStep();
        assertStepCount(1, seq);
        assertStep("KeyString\\(text\\)", seq.getStep(0));
    }

    /** Should save no modifiers or individual key press/release. */
    public void testRecordKeyStringShifted() {
        showTextField();
        startRecording();
        tester.actionKeyPress("VK_SHIFT");
        tester.actionKeyStroke("VK_T");
        tester.actionKeyStroke("VK_E");
        tester.actionKeyStroke("VK_X");
        tester.actionKeyStroke("VK_T");
        tester.actionKeyRelease("VK_SHIFT");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(1, seq);
        assertStep("KeyString\\(TEXT\\)", seq.getStep(0));
    }

    /** Should save no modifiers, since the modifier gets incorporated into
     * each key stroke.
     */
    public void testRecordKeyStringPartiallyShifted() {
        showTextField();
        startRecording();
        tester.actionKeyPress("VK_SHIFT");
        tester.actionKeyStroke("VK_T");
        tester.actionKeyRelease("VK_SHIFT");
        tester.actionKeyStroke("VK_E");
        tester.actionKeyStroke("VK_X");
        tester.actionKeyStroke("VK_T");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(1, seq);
        assertStep("KeyString\\(Text\\)", seq.getStep(0));
    }

    /** A modifier down/up should be captured that way. */
    public void testRecordModifier() {
        showTextField();
        startRecording();
        tester.actionKeyPress("VK_SHIFT");
        tester.actionKeyRelease("VK_SHIFT");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(2, seq);
        assertStep("KeyEvent.KEY_PRESSED \\(VK_SHIFT\\) on \\$\\{JTextField Instance\\}", 
                     seq.getStep(0));
        assertStep("KeyEvent.KEY_RELEASED \\(VK_SHIFT\\) on \\$\\{JTextField Instance\\}", 
                     seq.getStep(1));
    }

    public void testRecordMetaKey() {
        showTextField();
        startRecording();
        // w32 doesn't support the "meta" key modifier
        String mask = Platform.isWindows() ? "ALT_MASK" : "META_MASK";
        tester.actionKeyStroke("VK_P", mask);
        stopRecording();
        waitForStep();
        Sequence seq = (Sequence)step;
        assertStepCount(1, seq);
        // X11 may map meta to alt...
        assertStep("KeyStroke\\(VK_P,(ALT|META)_MASK\\)", seq.getStep(0));
    }

    public void testRecordWithNoKeyTypedEvents() {
        suppressKeyTypedEvents = true;
        showTextField();
        startRecording();
        tester.actionKeyStroke("VK_T");
        tester.actionKeyStroke("VK_E");
        tester.actionKeyStroke("VK_X");
        tester.actionKeyStroke("VK_T");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(4, seq);
        assertStep("KeyStroke\\(VK_T\\)", seq.getStep(0));
        assertStep("KeyStroke\\(VK_E\\)", seq.getStep(1));
        assertStep("KeyStroke\\(VK_X\\)", seq.getStep(2));
        assertStep("KeyStroke\\(VK_T\\)", seq.getStep(3));
        suppressKeyTypedEvents = false;
    }

    /** Test an event stream from rapidly typed keys. */
    public void testRecordJumbledKeyEvents() {
        showTextField();
        startRecording();
        // Actual event sequence seen when typing "SHIFt" rapidly
        tester.actionKeyPress("VK_SHIFT");
        tester.actionKeyPress("VK_S");
        tester.actionKeyRelease("VK_S");
        tester.actionKeyPress("VK_H");
        tester.actionKeyPress("VK_I");
        tester.actionKeyRelease("VK_H");
        tester.actionKeyPress("VK_F");
        tester.actionKeyRelease("VK_I");
        tester.actionKeyRelease("VK_F");
        tester.actionKeyRelease("VK_SHIFT");
        tester.actionKeyStroke("VK_T");
        stopRecording();
        waitForStep();
        // Should capture exactly 5 steps, one for each letter typed
        assertTrue("Event recorder should return a Sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(1, seq);
        assertStep("KeyString\\(SHIFt\\)", seq.getStep(0));
    }

    /** Tabs are used for focus traversal. */
    public void testRecordFocusTraversal() {
        JTextField comp1, comp2, comp3, comp4;
        JPanel pane = new JPanel();
        pane.add(comp1 = new JTextField("Field 1"));
        pane.add(comp2 = new JTextField("Field 2"));
        pane.add(comp3 = new JTextField("Field 3"));
        pane.add(comp4 = new JTextField("Field 4"));
        showFrame(pane);
        tester.actionFocus(comp1);
        startRecording();
        tester.actionKeyStroke("VK_TAB");
        tester.actionKeyStroke("VK_TAB");
        tester.actionKeyStroke("VK_TAB", "SHIFT_MASK");
        long start = System.currentTimeMillis();
        while (!comp2.hasFocus()) {
            if (System.currentTimeMillis() - start > 5000) {
                throw new Error("Timed out waiting for proper focus on "
                                + Robot.toString(comp2));
            }
        }
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(3, seq);
        toString(seq);
        assertStep("KeyStroke\\(VK_TAB\\)1*", seq.getStep(0));
        assertStep("KeyStroke\\(VK_TAB\\)2*", seq.getStep(1));
        assertStep("KeyStroke\\(VK_TAB,SHIFT_MASK\\)", seq.getStep(2));
        assertEquals("No text expected", "Field 1", comp1.getText());
        assertEquals("No text expected", "Field 2", comp2.getText());
        assertEquals("No text expected", "Field 3", comp3.getText());
        assertEquals("No text expected", "Field 4", comp4.getText());
    }

    /** Keystrokes with modifiers other than shift should be saved as
     * keystrokes rather than keystrings.
     */
    public void testRecordKeyStrokeInput() {
        showTextField();
        startRecording();
        tester.key(KeyEvent.VK_E, KeyEvent.CTRL_MASK);
        tester.waitForIdle();
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(1, seq);
        assertStep("KeyStroke\\(VK_E,CTRL_MASK\\)", seq.getStep(0));
    }

    public void testMacModifierStripping() {
        // Ensure the input method modifiers are stripped
        if (Platform.isMacintosh()) {
            showTextField();
            startRecording();
            tester.keyPress(KeyEvent.VK_ALT);
            tester.keyPress(KeyEvent.VK_E);
            tester.keyRelease(KeyEvent.VK_E);
            tester.keyRelease(KeyEvent.VK_ALT);
            tester.keyPress(KeyEvent.VK_E);
            tester.keyRelease(KeyEvent.VK_E);
            tester.waitForIdle();
            stopRecording();
            waitForStep();
            assertTrue("Event recorder should return a sequence",
                       step instanceof Sequence);
            Sequence seq = (Sequence)step;
            assertStepCount(1, seq);
            assertStep("KeyString\\(\u00e9\\)", seq.getStep(0));
        }
    }

    public void testMouseThenKey() {
        JButton button = new JButton("Press");
        showFrame(button);
        startRecording();
        tester.actionClick(button);
        tester.actionKeyStroke("VK_A");
        stopRecording();
        waitForStep();
        assertTrue("Event recorder should return a sequence",
                   step instanceof Sequence);
        Sequence seq = (Sequence)step;
        assertStepCount(2, seq);
        assertStep("Click\\(.*\\)", seq.getStep(0));
        assertStep("KeyString\\(a\\)", seq.getStep(1));
    }

    /** Create a new test case with the given name. */
    public EventRecorderTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        RepeatHelper.runTests(args, EventRecorderTest.class);
    }
}
