package abbot.script;

import javax.swing.SwingUtilities;

import junit.extensions.abbot.TestHelper;
import junit.framework.AssertionFailedError;
import junit.framework.TestCase;
import abbot.util.PreloadingClassLoader;

/** 
 * Verify AppClassLoader operation.
 */

// Additional tests:
// A class loaded by both system and app loader should differ
// load from different class path
// load resource from different class path
// load from manifest class path in jar in class path
public class AppClassLoaderTest extends TestCase {

    /** Ensure the bootstrap classes are always loaded by the bootstrap
     * loader, except in cases where we specifically want preloading.
     */
    public void testExclusion() throws Throwable {
        // applet viewer should get the custom loader
        AppClassLoader dcl = new AppClassLoader(null, false);

        Class appletClass = Class.forName("sun.applet.AppletViewer", true, dcl);
        assertTrue("AppletViewer class loading should *not* be delegated",
                   appletClass.getClassLoader() instanceof PreloadingClassLoader);

        Class orbClass = Class.forName("org.omg.CORBA.ORB", true, dcl);
        assertTrue("org.omg class loading should always be delegated",
                   !(orbClass.getClassLoader() instanceof PreloadingClassLoader));

        Class sliderClass = Class.forName("javax.swing.JSlider", true, dcl);
        assertTrue("javax.swing class loading should always be delegated",
                   !(sliderClass.getClassLoader() instanceof PreloadingClassLoader));
    }

    private ClassLoader eventClassLoader = null;
    public void testContextInstallUninstall() throws Throwable {
        final AppClassLoader dcl = new AppClassLoader(null, false);
        ClassLoader original = getClass().getClassLoader();
        ClassLoader origContext =
            Thread.currentThread().getContextClassLoader();
        try {
            dcl.install();
            eventClassLoader = null;
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    eventClassLoader =
                        Thread.currentThread().getContextClassLoader();
                }
            });
            assertEquals("Context not installed on event thread",
                         dcl, eventClassLoader);
            dcl.uninstall();
            eventClassLoader = null;
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    eventClassLoader =
                        Thread.currentThread().getContextClassLoader();
                }
            });
            assertTrue("Context not uninstalled from event thread",
                       eventClassLoader != dcl);
        }
        finally {
            // Make sure it's uninstalled in case something bad happened
            dcl.uninstall();
        }
    }

    public void testLoadAppletViewer() throws Throwable {
        String pathName = "lib/example.jar";
        String className = "sun.applet.AppletViewer";
        ClassLoader cl1 = new AppClassLoader(pathName, false);
        ClassLoader cl2 = new AppClassLoader(pathName, false);
        try {
            Class cls1 = Class.forName(className, true, cl1);
            Class cls2 = Class.forName(className, true, cl2);
            assertTrue("Class loader should define the class package",
                       cls1.getPackage() != null);
            assertEquals("Class loader defined the wrong package",
                         "sun.applet", cls1.getPackage().getName());

            Class cls3 = Class.forName(className);
            assertTrue("Core classes should be different when "
                       + "loaded dynamically", !cls1.equals(cls2));
            assertTrue("Class loader should never delegate AppletViewer class",
                       !cls3.equals(cls1) && !cls3.equals(cls2));
        }
        catch(ClassNotFoundException cnf) {
            fail("App class loader failed to load " + className
                 + " from path " + pathName);
        }
    }

    public void testBootstrapChecking() {
        AppClassLoader dcl = new AppClassLoader(null, true);
        String classes[] = { 
            "sun.applet.AppletViewer",
            "java.awt.Component",
            "javax.swing.JButton",
            "org.omg.CORBA.ORB",
        };
        for (int i=0;i < classes.length;i++) {
            assertTrue(classes[i] + " should be a bootstrap class",
                       dcl.isBootstrapClass(classes[i]));
        }
    }

    /** Load a custom tester and a tester and ensure the class hierarchy is
     * appropriate.
     */
    public void testExtensionTesterLoading() throws Throwable {
        AppClassLoader dcl =
            new AppClassLoader("lib/example.jar", false);
        Class customTesterClass = 
            Class.forName("abbot.tester.awt.extensions.ArrowButtonTester",
                          true, dcl);
        assertTrue("Wrong class loader for extension tester",
                   customTesterClass.getClassLoader() instanceof PreloadingClassLoader);

        assertEquals("Base tester class loader should match that of "
                     + "the test environment",
                     getClass().getClassLoader(),
                     customTesterClass.getSuperclass().getClassLoader());

        assertEquals("Base tester class should be that of the test "
                     + "environment", 
                     abbot.tester.awt.ComponentTester.class,
                     customTesterClass.getSuperclass());
    }

    public void testDirectBaseTesterLoading() throws Throwable {
        AppClassLoader dcl =
            new AppClassLoader("lib/abbot.jar", false);
        Class testerClass = 
            Class.forName("abbot.tester.ComponentTester", true, dcl);
        assertEquals("Custom class loader should be able to load a basic tester directly",
                     dcl, testerClass.getClassLoader());

        // A new reloading class loader should still get its own copy
        AppClassLoader dcl2 =
            new AppClassLoader("lib/abbot.jar", false);
        Class testerClass2 = 
            Class.forName("abbot.tester.ComponentTester", true, dcl2);
        assertEquals("Basic tester should be reloadable",
                     dcl2, testerClass2.getClassLoader());
    }

    /** Extensions not in the app loader class path but in the framework
     * class path should still be resolved correctly.
     */
    public void testExtensionsNotInClassPath() throws Throwable {
        AppClassLoader dcl =
            new AppClassLoader("lib/example.jar", false);
        try {
            Class customTesterClass = 
                Class.forName("abbot.tester.extensions.DummyTester",
                              true, dcl);
            assertTrue("Wrong class loader for custom tester",
                       customTesterClass.getClassLoader() instanceof PreloadingClassLoader);
            // The parent class of the custom class should not come from the
            // app class loader
            assertEquals("Tester superclass should match the test environment",
                         getClass().getClassLoader(),
                         customTesterClass.getSuperclass().getClassLoader());
        }
        catch(ClassNotFoundException cnf) {
            String msg = "The class loader should revert to java.class.path "
                + "if the extension is not found in the class loader "
                + "classpath.  The extension was not found.";
            throw new AssertionFailedError(msg);
        }
    }

    public AppClassLoaderTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        TestHelper.runTests(args, AppClassLoaderTest.class);
    }
}

