package abbot.script;

import junit.extensions.abbot.TestHelper;
import junit.framework.TestCase;
import abbot.converter.Converter;

/** Test ArgumentParser functions. */

public class ArgumentParserTest extends TestCase {

    public void testParseSingleArg() {
        String arg = "one";
        String[] args = ArgumentParser.parseArgumentList("one");
        assertEquals("Wrong number of arguments", 1, args.length);
        assertEquals("Badly parsed", arg, args[0]);
    }

    public void testParseMultipleArgs() {
        String[] expected = { "one", "two", "three" };
        String[] args = ArgumentParser.parseArgumentList("one,two,three");
        assertEquals("Wrong number of arguments",
                     expected.length, args.length);
        for (int i=0;i < args.length;i++) {
            assertEquals("Badly parsed", expected[i], args[i]);
        }
    }

    public void testParseArrayArgs() {
        String[] expected = { "one", "two", "[three,three]", "four" };
        String[] args = ArgumentParser.
            parseArgumentList("one,two,[three,three],four");
        assertEquals("Wrong number of arguments",
                     expected.length, args.length);
        for (int i=0;i < args.length;i++) {
            assertEquals("Badly parsed", expected[i], args[i]);
        }
    }

    public void testSubstitution() {
        abbot.Resolver resolver = new Script();
        String PROP_NAME = "my.prop";
        String PROP_VALUE = "value";
        resolver.setProperty(PROP_NAME, PROP_VALUE);
        String[][] args = {
            { "${my.prop}", PROP_VALUE },
            { "X${my.prop}", "X" + PROP_VALUE },
            { "${my.prop}X", PROP_VALUE + "X" },
            { "${my.prop", "${my.prop" },
            { "${my.prop}${my.prop}", PROP_VALUE + PROP_VALUE },
            { "${no.prop}${my.prop}", "${no.prop}" + PROP_VALUE },
        };
        for (int i=0;i < args.length;i++) {
            assertEquals("Bad substitution", args[i][1],
                         ArgumentParser.substitute(resolver, args[i][0]));
        }
    }

    public void testEvalWithSubstitution() {
        abbot.Resolver resolver = new Script();
        resolver.setProperty("my.prop", "value");
        assertEquals("Bad substitution", "value",
                     ArgumentParser.eval(resolver, "${my.prop}",
                                         String.class));
    }

    public void testEscapes() {
        String[] expected = {
            // comma should get unescaped in final arg
            "one,one", 
            // brackets should be left escaped
            "\\[two\\]", 
            // comma within array should be left escaped
            "[three\\,three]",
            // shouldn't get confused by trailing escaped brackets
            "four,\\]"
        };
        String[] args = ArgumentParser.
            parseArgumentList("one\\,one,\\[two\\],[three\\,three],four\\,\\]");
        assertEquals("Wrong number of arguments",
                     expected.length, args.length);
        for (int i=0;i < args.length;i++) {
            assertEquals("Badly parsed", expected[i], args[i]);
        }
    }

    public void testParseEscapes() {
        String with = "\\[commentary\\]";
        String without = "[commentary]";
        String stripped = ArgumentParser.unescapeBrackets(with);
        assertEquals("Escaped brackets not properly parsed",
                     without, stripped);
    }

    public void testReplace() {
        String s1 = "A string with some stuff to replace";
        String s2 = "A XXring with some XXuff to replace";
        assertEquals("Bad replacement", s2,
                     ArgumentParser.replace(s1, "st", "XX"));
    }

    public void testLoadConverter() {
        Converter cvt = ArgumentParser.getConverter(java.io.File.class);
        assertEquals("Wrong class loaded",
                     abbot.converter.FileConverter.class, cvt.getClass());
    }

    /** Create a new test case with the given name. */
    public ArgumentParserTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        TestHelper.runTests(args, ArgumentParserTest.class);
    }
}
