package abbot.script;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;

import junit.extensions.abbot.TestHelper;
import junit.framework.AssertionFailedError;
import junit.framework.TestCase;
import abbot.Log;
import abbot.util.ProcessOutputHandler;

/** 
 * Verify the sequence works as advertised.
 */
 // FIXME verify forked version has same classpath!
public class ForkedStepRunnerTest extends TestCase {

    private final static String LS = System.getProperty("line.separator");
    private final static String OUTPUT_TEXT = "This is the output stream";
    private final static String ERROR_TEXT = "This is the error stream";

    public void testFork() throws Throwable {
        ForkedStepRunner fs = new ForkedStepRunner();
        String[] args = {
            getClass().getName(), "fork",
        };
        final StringBuffer sb = new StringBuffer();
        final StringBuffer sbe = new StringBuffer();
        Process p;
        try {
            p = fs.fork(null, args);
        }
        catch(IOException io) {
            throw new AssertionFailedError("VM fork failed");
        }
        try {
            Thread thread = new ProcessOutputHandler(getName(), p) {
                protected void handleInput(byte[] buf, int len) {
                    sb.append(new String(buf, 0, len));
                }
                protected void handleError(byte[] buf, int len) {
                    sbe.append(new String(buf, 0, len));
                }
            };
            thread.start();
            p.waitFor();
            assertEquals("Wrong exit value", 0, p.exitValue());
            // FIXME add a timeout
            thread.join();
            assertEquals("Wrong output from subprocess",
                         OUTPUT_TEXT + LS, sb.toString());
            assertEquals("Wrong error from subprocess",
                         ERROR_TEXT + LS, sbe.toString());
        }
        finally {
            p.destroy();
        }
    }

    /** Steps are identified across processes by their encoded locations
     * within the script.
     */
    public void testStepEncoding() throws Throwable {
        String src = "<AWTTestScript>"
            + "  <sequence><!--0-->" /* A */
            + "    <sequence><!--0,1-->"
            + "      <terminate/><!--0,1,1-->" /* B */
            + "    </sequence>"
            + "    <sequence><!--0,2-->"
            + "    </sequence>"
            + "  </sequence>"
            + "  <sequence><!--1-->"
            + "    <sequence><!--1,1-->"
            + "      <terminate/><!--1,1,1-->"
            + "      <terminate/><!--1,1,2-->" /* C */
            + "    </sequence>"
            + "    <sequence><!--1,2-->" /* D */
            + "    </sequence>"
            + "  </sequence>"
            + "</AWTTestScript>";
        Script script = loadScript(src);
        // A
        Step step = script.getStep(0);
        String code = ForkedStepRunner.encodeStep(script, step);
        assertEquals("Wrong encoding", "0", code);
        assertEquals("Wrong decoding",
                     step, ForkedStepRunner.decodeStep(script, code));
        // B
        step = ((Sequence)((Sequence)script.getStep(0)).getStep(1)).getStep(1);
        code = ForkedStepRunner.encodeStep(script, step);
        assertEquals("Wrong nested encoding", "0,1,1", code);
        assertEquals("Wrong nested decoding",
                     step, ForkedStepRunner.decodeStep(script, code));
        // C
        step = ((Sequence)((Sequence)script.getStep(1)).getStep(1)).getStep(2);
        code = ForkedStepRunner.encodeStep(script, step);
        assertEquals("Wrong nested encoding", "1,1,2", code);
        assertEquals("Wrong nested decoding",
                     step, ForkedStepRunner.decodeStep(script, code));
        // D
        step = ((Sequence)script.getStep(1)).getStep(2);
        code = ForkedStepRunner.encodeStep(script, step);
        assertEquals("Wrong nested encoding", "1,2", code);
        assertEquals("Wrong nested decoding",
                     step, ForkedStepRunner.decodeStep(script, code));

        step = script;
        code = ForkedStepRunner.encodeStep(script, step);
        assertEquals("Wrong nested encoding", "-1", code);
        assertEquals("Wrong nested decoding",
                     step, ForkedStepRunner.decodeStep(script, code));

    }

    /** Run a script to completion. */
    public void testForkedStepRunner() throws Throwable {
        // Empty script that does nothing
        String src = "<AWTTestScript></AWTTestScript>";
        Script script = loadScript(src);
        final ArrayList events = new ArrayList();
        ForkedStepRunner fs = new ForkedStepRunner();
        fs.addStepListener(new StepListener() {
            public void stateChanged(StepEvent event) {
                events.add(event);
            }
        });
        fs.run(script);
        assertEquals("Wrong number of events", 2, events.size());
        assertEquals("Wrong step source, event 0",
                     script, ((StepEvent)events.get(0)).getStep());
        assertEquals("Wrong step id, event 0",
                     StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Wrong step source, event 1",
                     script, ((StepEvent)events.get(1)).getStep());
        assertEquals("Wrong step id, event 1",
                     StepEvent.STEP_END,
                     ((StepEvent)events.get(1)).getType());
    }

    /** Should be able to catch a failure from a forked script exactly like
     * you would from a regular one.
     */
    public void testForkedFailure() throws Throwable {
        String src = "<AWTTestScript>"
            + "<assert method=\"assertFrameShowing\" args=\"no such window\"/>"
            + "</AWTTestScript>";
        ForkedStepRunner fs = new ForkedStepRunner();
        Script script = loadScript(src);
        fs.run(script);
        Throwable thr = fs.getError(script);
        if (thr == null)
            fail("The failure was not propagated");
        assertEquals("Wrong failure type (" + thr.toString() + ") ", 
                     ForkedStepRunner.ForkedFailure.class,
                     thr.getClass());
    }

    /** Should be able to catch an error from a forked script exactly like
     * you would from a regular one.
     */
    public void testForkedError() throws Throwable {
        String src = "<AWTTestScript>"
            + "<launch class=\"nonsense class\" method=\"main\" args=\"[]\"/>"
            + "</AWTTestScript>";
        ForkedStepRunner fs = new ForkedStepRunner();
        Script script = loadScript(src);
        fs.run(script);
        Throwable thr = fs.getError(script);
        if (thr == null)
            fail("The error was not propagated");
        assertEquals("Wrong error type",
                     ForkedStepRunner.ForkedError.class,
                     thr.getClass());
        String str =
            "abbot.script.InvalidScriptException: Class 'nonsense class' not found (forked)";
        assertEquals("Wrong error", str, thr.toString());
    }

    private Script loadScript(String src) throws Throwable {
        StringReader reader = new StringReader(src);
        Script script = new Script();
        script.load(reader);
        return script;
    }

    public ForkedStepRunnerTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        args = Log.init(args);
        // special case for self-test
        if (args.length == 1 && args[0].equals("fork")) {
            System.out.println(OUTPUT_TEXT);
            System.err.println(ERROR_TEXT);
            System.exit(0);
        }
        TestHelper.runTests(args, ForkedStepRunnerTest.class);
    }
}
