package abbot.script;

import java.io.*;
import javax.swing.*;

import junit.extensions.abbot.*;
import junit.framework.TestCase;

import org.jdom.Element;

/** 
 * Verify the script works as advertised.
 */
public class ScriptTest extends ComponentTestFixture {

    private Script script = new Script();

    public void testFiles() throws Throwable {
        File tmp = File.createTempFile(getName(), ".xml");
        tmp.deleteOnExit();
        script.setFile(tmp);
        assertTrue("Script file should always be absolute",
                   script.getFile().isAbsolute());
        assertTrue("Script directory should always be absolute",
                   script.getDirectory().isAbsolute());
        assertEquals("Script filename should be absolute",
                     tmp.getPath(), script.getFilename());

        script.setFile(new File(tmp.getName()));
        script.setRelativeTo(tmp.getParentFile());
        assertTrue("Script file should always be absolute",
                   script.getFile().isAbsolute());
        assertTrue("Script directory should always be absolute",
                   script.getDirectory().isAbsolute());
        assertEquals("Script filename should be relative",
                     tmp.getName(), script.getFilename());
    }

    public void testDirty() throws Throwable {
        script.load();
        assertTrue("Script should not be dirty immediately after load",
                   !script.isDirty());
        script.addStep(new Script());
        assertTrue("Script should be dirty after modifications",
                   script.isDirty());
        script.removeStep(0);
        assertTrue("Script should not be dirty", !script.isDirty());
    }

    static final String ID = "button";
    static final String TAG = "\u65E5\u008Aencoded";
    static final String CLASS = "javax.swing.JButton";
    private static final String unicodeData = 
        "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
        + "<AWTTestScript>\r\n"
        + "  <component id=\""
        + ID + "\" class=\"" + CLASS + "\" tag=\"" + TAG + "\" />\r\n"
        + "  <terminate />\r\n"
        + "</AWTTestScript>\r\n\r\n";
    public void testWrite() throws Throwable {
        StringWriter writer = new StringWriter();
        Element el = new Element(script.TAG_COMPONENT).
            setAttribute(script.TAG_ID, ID).
            setAttribute(script.TAG_TAG, TAG).
            setAttribute(script.TAG_CLASS, CLASS);
        ComponentReference ref = script.addComponentReference(el);
        assertEquals("Incorrect tag encoding", TAG, ref.getTag());
        script.addStep(new Terminate(script, (String)null));
        script.save(writer);

        // Note that this is all Unicode, NOT UTF-8
        assertEquals("Incorrect XML written", unicodeData, writer.toString());
    }

    public void testRead() throws Throwable {
        try {
            String encoding = "UTF-8";
            ByteArrayInputStream ba =
                new ByteArrayInputStream(unicodeData.getBytes(encoding));
            InputStreamReader reader = new InputStreamReader(ba, encoding);

            script.load(reader);
            assertEquals("Wrong number of references",
                         1, script.getComponentReferences().size());
            assertEquals("Wrong number of steps", 1, script.size());
            ComponentReference ref = (ComponentReference)
                script.getComponentReferences().iterator().next();
            // Ensure all attributes are saved with unicode
            assertEquals("Wrong tag", TAG, ref.getTag());
        }
        catch(InvalidScriptException ise) {
            fail(ise.getMessage());
        }
    }

    /** Ensure that a nested script maintains its own context. */
    public void testNestedReference() throws Throwable {
        String encoding = "UTF-8";
        ByteArrayInputStream ba =
            new ByteArrayInputStream(unicodeData.getBytes(encoding));
        InputStreamReader reader = new InputStreamReader(ba, encoding);
        
        Script nested = new Script();
        nested.load(reader);
        script.addStep(nested);
        assertEquals("Wrong resolver", script, script.getResolver());
        assertEquals("Wrong nested resolver", nested, nested.getResolver());
        assertEquals("Wrong resolver for nested step",
                     nested, ((Step)nested.steps().get(0)).getResolver());

        String nestedScriptData = 
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
            + "<AWTTestScript>\r\n"
            + "  <script filename=\"nested.xml\" >\r\n"
            + "  </script>\r\n"
            + "</AWTTestScript>\r\n\r\n";
        ba = new ByteArrayInputStream(nestedScriptData.getBytes(encoding));
        reader = new InputStreamReader(ba, encoding);
        script.load(reader);
        nested = (Script)script.steps().get(0);
        assertEquals("Wrong resolver for nested script",
                     nested, nested.getResolver());
    }

    /** A nested script should have *only* its filename attribute saved. */
    public void testSaveNestedScript() throws Throwable {
        String fullNestedScript = 
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
            + "<AWTTestScript>\r\n"
            + "  <script filename=\"nested.xml\" >\r\n"
            + "    <comment desc=\"Filler\" />\r\n"
            + "  </script>\r\n"
            + "</AWTTestScript>\r\n\r\n";
        String savedNestedScript = 
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
            + "<AWTTestScript>\r\n"
            + "  <script filename=\"nested.xml\" />\r\n"
            + "</AWTTestScript>\r\n\r\n";
        String encoding = "UTF-8";
        ByteArrayInputStream ba =
            new ByteArrayInputStream(fullNestedScript.getBytes(encoding));
        InputStreamReader reader = new InputStreamReader(ba, "UTF-8");
        script.load(reader);
        StringWriter writer = new StringWriter();
        script.save(writer);
        assertEquals("Incorrectly saved",
                     savedNestedScript, writer.toString());
        Script nested = (Script)script.steps().get(0);
        assertEquals("Wrong relative directory", 
                     script.getDirectory(), nested.getRelativeTo());
    }

    /** If a nested file does not exist, an exception should be thrown. */
    public void testNestedFileNotFound() throws Throwable {
        Script nested = new Script();
        nested.setFile(new File("somewhere/relative/file.xml"));
        script.addStep(nested);
        try {
            script.run();
            fail("Exception should have been thrown");
        }
        catch(InvalidScriptException ise) {
        }
    }

    public void testFilePaths() {
        String relativePath = "relative.xml";
        String relativeDir = "relative";
        String absolutePath = new File(relativePath).getAbsolutePath();
        String absoluteDir = new File(relativeDir).getAbsolutePath();
        script.setFile(new File(relativePath));
        assertEquals("Wrong relative filename",
                     relativePath, script.getFilename());
        assertEquals("Wrong relative dir",
                     new File(System.getProperty("user.dir")),
                     script.getRelativeTo());
        assertEquals("Wrong file",
                     new File(script.getRelativeTo(), relativePath),
                     script.getFile());

        script.setRelativeTo(new File(absoluteDir));
        assertEquals("Wrong relative filename (1)",
                     relativePath, script.getFilename());
        assertEquals("Wrong relative dir (1)",
                     new File(absoluteDir), script.getRelativeTo());
        assertEquals("Wrong file (1)",
                     new File(absoluteDir, relativePath), script.getFile());

        script.setRelativeTo(new File(relativeDir));
        assertEquals("Wrong relative filename (2)",
                     relativePath, script.getFilename());
        assertEquals("Wrong relative dir (2)",
                     new File(relativeDir), script.getRelativeTo());
        assertEquals("Wrong file (2)",
                     new File(relativeDir, relativePath), script.getFile());

        script.setFile(new File(absolutePath));
        assertEquals("Wrong relative filename (3)",
                     absolutePath, script.getFilename());
        assertEquals("Wrong relative dir (3)",
                     new File(relativeDir), script.getRelativeTo());
        assertEquals("Wrong file (3)",
                     new File(absolutePath), script.getFile());
        
        script.setFile(new File(absoluteDir, relativePath));
        script.setRelativeTo(new File(absoluteDir));
        assertEquals("Wrong relative filename (4)",
                     relativePath, script.getFilename());
        assertEquals("Wrong relative dir (4)",
                     new File(absoluteDir), script.getRelativeTo());
        assertEquals("Wrong file (4)",
                     new File(absoluteDir, relativePath), script.getFile());
    }

    /** Any given component, if added multiple times, should result in the
     * same component reference.
     */
    public void testReferenceReuse() {
        JTextField tf = new JTextField();
        showFrame(tf);

        ComponentReference ref = script.addComponent(tf);
        assertTrue("New reference does not match the component"
                   + " from which it was created",
                   getFinder().componentsMatch(tf, ref,
                                               getFinder().MATCH_EXACT));

        ComponentReference ref2 = script.addComponent(tf);
        assertEquals("Same component should always result in the same ref",
                     ref, ref2);

        assertEquals("Reverse lookup of reference by component failed",
                     ref, script.getComponentReference(tf));
    }

    /** Any number of unique components should all result in unique
     * references.
     */
    public void testUniqueReferences() {
        JTextField tf1 = new JTextField();
        JTextField tf2 = new JTextField();
        JButton b1 = new JButton("Button");
        JButton b2 = new JButton("Button");
        JPanel pane = new JPanel();
        pane.add(tf1);
        pane.add(tf2);
        pane.add(b1);
        pane.add(b2);
        showFrame(pane);

        ComponentReference cr1 = script.addComponent(tf1);
        ComponentReference cr2 = script.addComponent(tf2);
        assertTrue("Should get two unique text field references", cr1 != cr2);

        ComponentReference cr3 = script.addComponent(b1);
        ComponentReference cr4 = script.addComponent(b2);
        assertTrue("Should get two unique button references", cr3 != cr4);
    }

    public ScriptTest(String name) {
        super(name);
    }

    public static void main(String[] args) {
        TestHelper.runTests(args, ScriptTest.class);
    }
}
