package abbot.script;

import java.util.ArrayList;

import junit.extensions.abbot.TestHelper;
import abbot.AssertionFailedError;

public class StepRunnerTest extends junit.framework.TestCase {

    private StepRunner runner;
    private ArrayList events;

    private class DefaultStep extends Step {
        public DefaultStep(String desc) {
            super(new Script(), desc);
        }
        public void runStep() throws Throwable { }
        public String getXMLTag() { return ""; }
        public String getUsage() { return ""; }
        protected String getDefaultDescription() { return ""; }
    }

    private class FailingStep extends DefaultStep {
        public FailingStep(String desc) {
            super(desc);
        }
        public void runStep() throws Throwable {
            throw new AssertionFailedError(getDescription());
        }
    }

    private class ErrorStep extends DefaultStep {
        public ErrorStep(String desc) {
            super(desc);
        }
        public void runStep() throws Throwable {
            throw new Error(getDescription());
        }
    }

    public void testRunStep() throws Throwable {
        runner.runStep(new DefaultStep("simple step"));
        assertEquals("Wrong number of step events generated",
                     2, events.size());
        assertEquals("Missing step start",
                     StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Missing step end",
                     StepEvent.STEP_END,
                     ((StepEvent)events.get(1)).getType());
    }

    public void testRunSequence() throws Throwable {
        Sequence seq = new Sequence(new Script(), "sequence of 2", null);
        Step st1 = new DefaultStep("step 1");
        seq.addStep(st1);
        Step st2 = new DefaultStep("step 2");
        seq.addStep(st2);
        runner.runStep(seq);
        assertEquals("Wrong number of step events generated",
                     6, events.size());
        assertEquals("Missing sequence start",
                     StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Expected sequence start, wrong source",
                     seq, ((StepEvent)events.get(0)).getStep());
        assertEquals("Missing step 1 start",
                     StepEvent.STEP_START,
                     ((StepEvent)events.get(1)).getType());
        assertEquals("Wrong source",
                     st1, ((StepEvent)events.get(1)).getStep());
        assertEquals("Missing step 1 end",
                     StepEvent.STEP_END,
                     ((StepEvent)events.get(2)).getType());
        assertEquals("Wrong source",
                     st2, ((StepEvent)events.get(3)).getStep());
        assertEquals("Missing step 2 start",
                     StepEvent.STEP_START,
                     ((StepEvent)events.get(3)).getType());
        assertEquals("Missing step 2 end",
                     StepEvent.STEP_END,
                     ((StepEvent)events.get(4)).getType());
        assertEquals("Missing sequence end",
                     StepEvent.STEP_END,
                     ((StepEvent)events.get(5)).getType());
    }

    public void testRunFailedStep() throws Throwable {
        Step step = new FailingStep("fail");
        Throwable failure = null;
        try {
            runner.runStep(step);
        }
        catch(AssertionFailedError afe) {
            failure = afe;
        }
        assertTrue("step failure should have been caught",
                   runner.getError(step) != null);
        assertEquals("wrong failure", failure, runner.getError(step));
        assertEquals("Wrong number of events", 2, events.size());
        assertEquals("Missing step start", StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Missing step failure", StepEvent.STEP_FAILURE,
                     ((StepEvent)events.get(1)).getType());
    }

    public void testRunFailedSequence() throws Throwable {
        Sequence seq = new Sequence(new Script(), "sequence of 1", null);
        Step step = new FailingStep("fail");
        seq.addStep(step);
        Throwable failure = null;
        try {
            runner.runStep(seq);
        }
        catch(AssertionFailedError afe) {
            failure = afe;
        }
        assertTrue("step failure should have been caught",
                   runner.getError(step) != null);
        assertEquals("wrong failure", failure, runner.getError(step));
        assertEquals("Wrong number of events", 4, events.size());
        assertEquals("Missing sequence start", StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Wrong event source", seq, 
                     ((StepEvent)events.get(0)).getStep());
        assertEquals("Missing step start", StepEvent.STEP_START,
                     ((StepEvent)events.get(1)).getType());
        assertEquals("Wrong event source", step, 
                     ((StepEvent)events.get(1)).getStep());
        assertEquals("Missing step failure", StepEvent.STEP_FAILURE,
                     ((StepEvent)events.get(2)).getType());
        assertEquals("Wrong event source", step, 
                     ((StepEvent)events.get(2)).getStep());
        assertEquals("Missing step failure", StepEvent.STEP_FAILURE,
                     ((StepEvent)events.get(3)).getType());
        assertEquals("Wrong event source", seq, 
                     ((StepEvent)events.get(3)).getStep());
    }

    public void testRunErrorStep() {
        Step step = new ErrorStep("error");
        Throwable error = null;
        try {
            runner.runStep(step);
        }
        catch(Throwable thr) {
            error = thr;
        }
        assertTrue("step error should have been caught",
                   runner.getError(step) != null);
        assertEquals("wrong error", error, runner.getError(step));
        assertEquals("Wrong number of events", 2, events.size());
        assertEquals("Missing step start", StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Missing step failure", StepEvent.STEP_ERROR,
                     ((StepEvent)events.get(1)).getType());
    }

    public void testRunErrorSequence() throws Throwable {
        Sequence seq = new Sequence(new Script(), "sequence of 1", null);
        Step step = new ErrorStep("error");
        seq.addStep(step);
        Throwable failure = null;
        try {
            runner.runStep(seq);
        }
        catch(Throwable thr) {
            failure = thr;
        }
        assertTrue("step error should have been caught",
                   runner.getError(step) != null);
        assertEquals("wrong failure", failure, runner.getError(step));
        assertEquals("Wrong number of events", 4, events.size());
        assertEquals("Missing sequence start", StepEvent.STEP_START,
                     ((StepEvent)events.get(0)).getType());
        assertEquals("Wrong event source", seq, 
                     ((StepEvent)events.get(0)).getStep());
        assertEquals("Missing step start", StepEvent.STEP_START,
                     ((StepEvent)events.get(1)).getType());
        assertEquals("Wrong event source", step, 
                     ((StepEvent)events.get(1)).getStep());
        assertEquals("Missing step failure", StepEvent.STEP_ERROR,
                     ((StepEvent)events.get(2)).getType());
        assertEquals("Wrong event source", step, 
                     ((StepEvent)events.get(2)).getStep());
        assertEquals("Missing step failure", StepEvent.STEP_ERROR,
                     ((StepEvent)events.get(3)).getType());
        assertEquals("Wrong event source", seq, 
                     ((StepEvent)events.get(3)).getStep());
    }

    /** If the script is set to run to completion, make sure it throws an
     * error at the end.
     */
    public void testThrowMultipleError() throws Throwable {
        Sequence seq = new Sequence(new Script(), "sequence of 2", null);
        Step error1 = new ErrorStep("error 1");
        seq.addStep(error1);
        Step error2 = new ErrorStep("error 2");
        seq.addStep(error2);
        Throwable error = null;
        runner.setStopOnError(false);
        try {
            runner.runStep(seq);
        }
        catch(Throwable thr) {
            StepEvent lastEvent = (StepEvent)events.get(events.size()-1);
            assertEquals("No error on event", thr, lastEvent.getError());
            assertEquals("Wrong event source", seq, lastEvent.getStep());
        }
    }

    protected void setUp() {
        events = new ArrayList();
        runner = new StepRunner();
        runner.addStepListener(new StepListener() {
            public void stateChanged(StepEvent ev) {
                events.add(ev);
            }
        });
    }

    public StepRunnerTest(String name) { super(name); }

    public static void main(String[] args) {
        TestHelper.runTests(args, StepRunnerTest.class);
    }
}
