package abbot.tester.awt;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.event.*;

import junit.extensions.abbot.*;
import junit.extensions.abbot.Timer;
import junit.framework.*;
import abbot.*;

/** Unit test to verify the base ComponentTester class.<p>
    
    <ul>
    <li>Test all exported actions.
    </ul>
 */
// OSX uses a button/modifier combo to indicate MB2/3...ick
// w32 uses a button/modifier combo to indicate MB3...ick
public class ComponentTesterTest extends ComponentTestFixture {

    private ComponentTester tester;

    protected void setUp() {
        tester = new ComponentTester();
    }
    protected void tearDown() {
        tester = null;
    }


    /** Create a new test case with the given name. */
    public ComponentTesterTest(String name) {
        super(name);
    }

    private void checkDialog(Frame parent) {
        // Now try it with a dialog
        String title = getName() + " (Dialog under " 
            + Robot.toString(parent) + ")";
        final JDialog d = new JDialog(parent, title);
        d.getContentPane().add(new JLabel("I'm a dialog, and I'm OK"));
        // Need to do this on a separate thread, since the dialog show will
        // block until the dialog closes
        new Thread("Dialog shower") {
            public void run() { d.pack(); d.show(); }
        }.start();
        
        Timer timer = new Timer();
        while (!d.isShowing()) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("Dialog failed to show after 5s");
            tester.sleep();
        }
        assertTrue("Dialog '" + title + "' wasn't showing or wasn't detected",
                   tester.assertFrameShowing(title));
        d.hide();
        d.dispose();
    }

    private class ActionFlag extends AbstractAction {
        public boolean gotAction = false;
        public ActionFlag(String label) { super(label); }
        public void actionPerformed(ActionEvent ev) {
            gotAction = true;
        }
    }

    public void testSelectPopupMenuItem() {
        JPanel pane = new JPanel();
        ActionFlag item1 = new ActionFlag("Item One");
        ActionFlag item2 = new ActionFlag("Item Two");
        ActionFlag sub1 = new ActionFlag("Subitem One");
        ActionFlag sub2 = new ActionFlag("Subitem Two");
        JPopupMenu popup = new JPopupMenu();
        popup.add(new JMenuItem(item1));
        popup.add(new JMenuItem(item2));
        JMenu submenu = new JMenu("Submenu");
        submenu.add(new JMenuItem(sub1));
        submenu.add(new JMenuItem(sub2));
        popup.add(submenu);

        installPopup(pane, popup);
        showFrame(pane, new Dimension(200, 200));

        // This implicitly tests "actionShowPopupMenu"
        tester.actionSelectPopupMenuItem(pane, "Item One");
        assertTrue("First popup menu item not selected", item1.gotAction);
        tester.actionSelectPopupMenuItem(pane, "Item Two");
        assertTrue("Second popup menu item not selected", item2.gotAction);
        tester.actionSelectPopupMenuItem(pane, "Subitem One");
        assertTrue("First popup submenu item not selected", sub1.gotAction);
        tester.actionSelectPopupMenuItem(pane, "Subitem Two");
        assertTrue("Second popup submenu item not selected", sub2.gotAction);
    }

    public void testFrameShowingAssertion() {
        Frame frame = showFrame(new JPanel(), new Dimension(200, 200));
        String title = frame.getTitle();
        // helper does a wait for idle, so the frame should be showing at this
        // point. 
        assertTrue("Frame '" + title + "' wasn't showing or wasn't detected",
                   tester.assertFrameShowing(title));

        // Try a dialog with a non-null parent
        checkDialog(frame);
        // Now try one with a null parent
        checkDialog(null);
        getFinder().disposeWindows();
        // And make sure we can find it the next time it's shown
        checkDialog(null);
    }

    public void testFrameShowingAssertionWithDialog() {
        final String title = "Not me";
        final String title2 = "Wait for me";
        Frame frame = showFrame(new Label(getName()), new Dimension(200, 200));
        JDialog d1 = new JDialog(frame, title, true);
        d1.getContentPane().add(new JLabel("Modal dialog"));
        JDialog d2 = new JDialog(d1, title2, false);
        d2.getContentPane().add(new JLabel("Non-modal subdialog"));
        try {
            showWindow(d1);
            showWindow(d2);
            assertTrue("Second dialog should have been detected",
                       tester.assertFrameShowing(title2));
        }
        finally { 
            d2.dispose();
            d1.dispose();
        }
    }

    public void testFrameShowing() {
        Frame frame = showFrame(new JLabel(getName()));
        assertEquals("Frame not showing", 
                     true, tester.assertFrameShowing(frame.getTitle()));
        frame.dispose();
        assertEquals("Frame still showing",
                     false, tester.assertFrameShowing(frame.getTitle()));
    }

    /** Check translation between event IDs and integer values. */
    public void testEventID() {
        assertEquals("mouse press", MouseEvent.MOUSE_PRESSED,
                     tester.getEventID(MouseEvent.class,
                                       "MOUSE_PRESSED"));
        assertEquals("mouse release", MouseEvent.MOUSE_RELEASED,
                     tester.getEventID(MouseEvent.class,
                                       "MOUSE_RELEASED"));
        assertEquals("mouse click", MouseEvent.MOUSE_CLICKED,
                     tester.getEventID(MouseEvent.class,
                                       "MOUSE_CLICKED"));
    }

    /** Verify forward and back parsing of modifiers. */
    public void testModifiers() {
        // NOTE: As of Java 1.3, Meta and Control also map to button 2/3.
        // Currently the symbolic button masks are preferred to the symbolic
        // key masks. 
        assertEquals("shift", InputEvent.SHIFT_MASK,
                     tester.getModifiers("SHIFT_MASK"));
        assertEquals("button 1", InputEvent.BUTTON1_MASK,
                     tester.getModifiers("BUTTON1_MASK"));
        assertEquals("button 2", InputEvent.BUTTON2_MASK,
                     tester.getModifiers("BUTTON2_MASK"));
        assertEquals("button 3", InputEvent.BUTTON3_MASK,
                     tester.getModifiers("BUTTON3_MASK"));
        // combinations
        assertEquals("shift + button 1", 
                     InputEvent.SHIFT_MASK | InputEvent.BUTTON1_MASK,
                     tester.getModifiers("SHIFT_MASK | BUTTON1_MASK"));
    }

    /** Check virtual keycode parsing. */
    public void testKeyCodeParsing() {
        assertEquals("key 'A'", KeyEvent.VK_A,
                     tester.getKeyCode("VK_A"));
        assertEquals("key '0'", KeyEvent.VK_0,
                     tester.getKeyCode("VK_0"));
    }

    private class KeyWatcher extends KeyAdapter {
        public boolean gotPress = false;
        public boolean gotRelease = false;
        public boolean gotTyped = false;
        public int keyCode = KeyEvent.VK_UNDEFINED;
        public int modifiers = 0;
        public void keyPressed(KeyEvent ke) {
            gotPress = true;
            keyCode = ke.getKeyCode();
            modifiers = ke.getModifiers();
        }
        public void keyReleased(KeyEvent ke) {
            gotRelease = true;
        }
        public void keyTyped(KeyEvent ke) {
            gotTyped = true;
        }
    }

    public void testKeyStroke() {
        JTextField tf = new JTextField();
        KeyWatcher kw = new KeyWatcher();
        tf.addKeyListener(kw);
        tf.setColumns(10);
        showFrame(tf);
        tester.actionFocus(tf);
        int code = KeyEvent.VK_A;
        int mods = KeyEvent.SHIFT_MASK;
        tester.actionKeyStroke("VK_A", "SHIFT_MASK");
        assertTrue("Never received key press", kw.gotPress);
        assertTrue("Never received release", kw.gotRelease);
        assertTrue("Never received type", kw.gotTyped);
        assertEquals("Wrong key code", code, kw.keyCode);
        assertEquals("Wrong modifiers", tester.getKeyModifiers(mods), 
                     tester.getKeyModifiers(kw.modifiers));
    }

    private class MouseWatcher extends MouseAdapter {
        public volatile int clickCount = 0;
        public void mouseClicked(MouseEvent me) {
            clickCount = me.getClickCount();
        }
    };
    /** Ensure the advertised multiple click method works. */
    public void testDoubleClick() {
        MouseWatcher mw = new MouseWatcher();
        JTextField tf = new JTextField("SelectMe");
        tf.addMouseListener(mw);
        showFrame(tf);
        tester.actionFocus(tf);
        tester.actionClick(tf, tf.getWidth()/2, tf.getHeight()/2,
                           "BUTTON1_MASK", 2);
        assertEquals("Wrong number of clicks", 2, mw.clickCount);
    }
    
    /** Ensure that we can send clicks close enough together to register as a
        double click. */
    public void testTwoClicks() {
        MouseWatcher mw = new MouseWatcher();
        JTextField tf = new JTextField("SelectMe");
        tf.addMouseListener(mw);
        showFrame(tf);
        tester.actionFocus(tf);
        tester.actionClick(tf);
        tester.actionClick(tf);
        assertEquals("Wrong number of clicks", 2, mw.clickCount);
    }

    public void testDeriveTag() {
        Button button = new Button("AWT Button");
        assertEquals("Tag should be the Button's label",
                     button.getLabel(), tester.deriveTag(button));
        JButton jbutton = new JButton("JButton");
        assertEquals("Tag should be the JButton's text",
                     jbutton.getText(), tester.deriveTag(jbutton));
        JFrame frame = new JFrame("JFrame");
        assertEquals("Tag should be the JFrame's title",
                     frame.getTitle(), tester.deriveTag(frame));
    }

    public void testStripHTML() {
        String html1 = "<html>simple</html>";
        String strip1 = "simple";
        String html2 = "<html><font color=\"red\" size=\"-1\">complex</font></html>";
        String strip2 = "complex";
        assertEquals("Simple html was not stripped",
                     strip1, ComponentTester.stripHTML(html1));
        assertEquals("Complex html was not stripped",
                     strip2, ComponentTester.stripHTML(html2));
    }

    private boolean gotDrag = false;
    public void testDragDrop() {
        JLabel label = new JLabel(getName());
        showFrame(label);
        gotDrag = false;
        MouseInputAdapter ma = new MouseInputAdapter(){
            public void mouseDragged(MouseEvent me){ gotDrag = true; }
        };
        label.addMouseListener(ma);
        label.addMouseMotionListener(ma);
        tester.actionDrag(label, 1, 1);
        tester.actionDrop(label, label.getWidth()/2,
                          label.getHeight()/2);
        assertTrue("Should have seen a drag event", gotDrag);
    }

    /** Assumes "classes" has the class in question. */
    public void testGetTester() throws Throwable {
        String cname = "test.fake.ComponentClassForLoading";
        ClassLoader cl = 
            new abbot.util.PreloadingClassLoader("classes", getClass().getClassLoader());
        Class cls = Class.forName(cname, true, cl);
        assertEquals("Wrong class loader for test class",
                     cl, cls.getClassLoader());
        ComponentTester customTester = ComponentTester.getTester(cls);
        ComponentTester tester =
            ComponentTester.getTester(java.awt.Component.class);
        assertEquals("Wrong tester loaded", tester, customTester);
        assertTrue("Class loaders should be different",
                   !cls.getClassLoader().equals(tester.getClass().getClassLoader()));
    }

    public static void main(String[] args) {
        RepeatHelper.runTests(args, ComponentTesterTest.class);
    }
}
