package test;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;

import javax.swing.*;
import javax.swing.border.EmptyBorder;

import junit.extensions.abbot.Timer;
import abbot.*;
import abbot.script.Script;
import abbot.tester.awt.*;


/** Simple helper class to provide commonly used utilities for testing
 * various Abbot components.
 * @deprecated This functionality has mostly been moved into
 * junit.extensions.abbot.ComponentTestFixture. 
 */
public class AbbotHelper implements AWTEventListener {
    /** Use the same finder for all tests. */
    public static ComponentFinder finder = 
        DefaultComponentFinder.getFinder();

    /** All event capturing objects we've created so far. */
    private ArrayList captures = new ArrayList();
    public Resolver resolver;
    public ComponentTester tester;
    public EventWatcher watcher;

    /** Must listen to raw event queue; event adapters/listeners don't get
     * events reliably.
     */
    public abstract class EventCapture implements AWTEventListener {
        protected Component component;
        public EventCapture(Component comp) {
            component = comp;
        }
        public void eventDispatched(AWTEvent event) {
            Log.debug("Got " + tester.toString(event));
            if (event.getSource() == component)
                handleEvent(event);
        }
        public abstract void handleEvent(AWTEvent event);
        public abstract long getMask();
        public void startCapture() {
            Toolkit.getDefaultToolkit().addAWTEventListener(this, getMask());
        }
        public void stopCapture() {
            Toolkit.getDefaultToolkit().removeAWTEventListener(this);
        }
    }

    public class MouseCapture extends EventCapture {
        public MouseEvent mousePressEvent = null;
        public MouseEvent mouseReleaseEvent = null;
        public MouseEvent mouseClickEvent = null;
        public MouseCapture(Component c) { super(c); }
        public void handleEvent(AWTEvent ev) {
            int id = ev.getID();
            if (id == MouseEvent.MOUSE_PRESSED)
                mousePressEvent = (MouseEvent)ev;
            else if (id == MouseEvent.MOUSE_RELEASED)
                mouseReleaseEvent = (MouseEvent)ev;
            else if (id == MouseEvent.MOUSE_CLICKED)
                mouseClickEvent = (MouseEvent)ev;
        }
        public void clear() {
            mousePressEvent = mouseReleaseEvent = mouseClickEvent = null;
        }
        public long getMask() { return AWTEvent.MOUSE_EVENT_MASK; }
    }

    public class KeyCapture extends EventCapture {
        public KeyEvent keyPressEvent = null;
        public KeyEvent keyReleaseEvent = null;
        public KeyEvent keyTypeEvent = null;
        public KeyCapture(Component c) { super(c); }
        public void handleEvent(AWTEvent ev) {
            int id = ev.getID();
            if (id == KeyEvent.KEY_PRESSED)
                keyPressEvent = (KeyEvent)ev;
            else if (id == KeyEvent.KEY_RELEASED)
                keyReleaseEvent = (KeyEvent)ev;
            else if (id == KeyEvent.KEY_TYPED)
                keyTypeEvent = (KeyEvent)ev;
        }
        public void clear() {
            keyPressEvent = keyReleaseEvent = keyTypeEvent = null;
        }
        public long getMask() { return AWTEvent.KEY_EVENT_MASK; }
    }
    public JFrame frame;
    private boolean frameActive = false;
    public JPanel pane;
    public static final int TIMEOUT = 5000;

    public boolean isFrameActive() {
        tester.actionWaitForIdle();
        return frameActive;
    }

    public void setFrameActive(boolean active) {
        frameActive = active;
    }

    public void hideWindow(Window win) {
        setVisible(win, false);
    }

    public void showWindow(Window win) {
        setVisible(win, true);
    }

    private void setVisible(final Window win, final boolean visible) {

        tester.invokeAndWait(new Runnable() {
            public void run() {
                if (visible) {
                    win.pack();
                    win.show();
                }
                else {
                    win.hide();
                }
            }
        });
        waitForShowing(win, visible);
    }

    public void waitForShowing(Window win, boolean visible) {
        long now = System.currentTimeMillis();
        while ((visible && (!win.isVisible() || !win.isShowing()))
               || (!visible && (win.isVisible() || win.isShowing()))) {
            if (System.currentTimeMillis() - now > 5000)
                throw new RuntimeException("Window failed to "
                                           + (visible ? "show" : "hide"));
            tester.sleep();
        }
        // Make sure the frame always starts out activated
        if (visible) {
            tester.activate(win);
        }
        //watcher.startListening();
    }

    public Dialog showSimpleDialog(String msg) {
        JOptionPane pane =
            new JOptionPane(msg, JOptionPane.INFORMATION_MESSAGE);
        final Dialog dialog = pane.createDialog(frame, "Dialog");
        SwingUtilities.invokeLater(new Runnable() {
            public void run() { dialog.pack(); dialog.show(); }
        });
        Timer timer = new Timer();
        while (!dialog.isShowing()) {
            if (timer.elapsed() > 5000)
                throw new RuntimeException("Timed out waiting for dialog to show");
        }
        return dialog;
    }

    public void showFrame(Component comp) {
        pane.add(comp);
        showWindow(frame);
    }

    /** Display the main frame. */
    public void showFrame() {
        showWindow(frame);
    }

    /** Mostly for debugging.  Store all received events into the log file. */
    public void logEvents(boolean enable) {
        Toolkit toolkit = Toolkit.getDefaultToolkit();
        if (enable) {
            Log.log("Event log started for " + frame.getTitle());
            toolkit.addAWTEventListener(this, -1);
        }
        else {
            toolkit.removeAWTEventListener(this);
        }
    }

    /** Generate a JFrame with some useful initial state. */
    public JFrame createFrame(String name) {
        JFrame frame = new JFrame(name);
        frame.setName(name);
        frame.setLocation(35, 35);
        JPanel pane = (JPanel)frame.getContentPane();
        pane.setName(name + " Content Pane");
        pane.setBorder(new EmptyBorder(50, 50, 50, 50));
        pane.setLayout(new FlowLayout());
        return frame;
    }

    public void setUp(String name) {
        resolver = new Script();

        frame = createFrame(name);
        frame.addWindowListener(new WindowAdapter() {
            public void windowActivated(WindowEvent e) {
                frameActive = true;
            }
            public void windowDeactivated(WindowEvent e) {
                frameActive = false;
            }
        });
        pane = (JPanel)frame.getContentPane();
        tester = ComponentTester.getTester(pane);
        watcher = new EventWatcher();
        tester.resetPointer();
    }

    public KeyCapture addKeyListener(Component comp) {
        KeyCapture kc = new KeyCapture(comp);
        kc.startCapture();
        captures.add(kc);
        return kc;
    }

    public MouseCapture addMouseListener(Component comp) {
        MouseCapture mc = new MouseCapture(comp);
        mc.startCapture();
        captures.add(mc);
        return mc;
    }

    public void tearDown() {
        watcher.stopListening();
        logEvents(false);
        frame.dispose();
        frame = null;
        pane = null;
        for (int i=0;i < captures.size();i++) {
            ((EventCapture)captures.get(i)).stopCapture();
        }
        captures.clear();
        // Get rid of ALL windows that are still hanging around
        finder.disposeWindows();
    }

    public void eventDispatched(AWTEvent event) {
        Log.log(ComponentTester.toString(event));
    }

    private JButton findButton(Component comp, String text) {
        if (comp instanceof JButton) {
            if (((JButton)comp).getText().equals(text)) 
                return (JButton)comp;
        }
        else if (comp instanceof Container) {
            Component[] children = ((Container)comp).getComponents();
            for (int i=0;i < children.length;i++) {
                JButton b = findButton(children[i], text);
                if (b != null)
                    return b;
            }
        }
        return null;
    }

    public JButton findButton(String text) {
        Frame[] frames = Frame.getFrames();
        for (int i=0;i < frames.length;i++) {
            if (frames[i].isShowing()) {
                JButton button = findButton(frames[i], text);
                if (button != null)
                    return button;
            }
            Window[] subs = frames[i].getOwnedWindows();
            for (int j=0;j < subs.length;j++) {
                if (subs[j].isShowing()) {
                    JButton button = findButton(subs[j], text);
                    if (button != null)
                        return button;
                }
            }
        }
        return null;
    }
}
